
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../common/core.h"
#include "../common/socket.h"
#include "../common/mmo.h"
#include "../common/crypt.h"

#include "char.h"
#include "logif.h"
#include "char_db.h"

static int logif_table_recv[5] = {2, 19, 40, 39, 51};

int check_connect_login() {
	if (login_fd <= 0 || session[login_fd] == NULL) {
		printf("Attempt to connect to login-server...\n");
		login_fd = make_connection(login_ip, login_port);
		if(login_fd <= 0) return 0;
		session[login_fd]->func_parse = logif_parse;
		realloc_fifo(login_fd, FIFOSIZE_SERVER, FIFOSIZE_SERVER);
		WFIFOB(login_fd,0) = 0xAA;
		WFIFOW(login_fd,1) = SWAP16(66);
		WFIFOB(login_fd,3) = 0xFF;
		WFIFOB(login_fd,4) = RAND_INC;
		memcpy(WFIFOP(login_fd,5), login_id, 32);
		memcpy(WFIFOP(login_fd,37), login_pw, 32);
		crypt(WFIFOP(login_fd,0));
		WFIFOSET(login_fd, 69);
	}
	return 0;
}

int logif_parse(int fd) {
	int res;
	unsigned int id;
	if(session[fd]->eof) {
		mmo_setallonline(0);
		printf("Can't connect to Login Server.\n");
		add_log("Can't connect to Login Server.\n");
		login_fd = 0;
		
		do_close(fd);
		
		return 0;
	}

	if(RFIFOB(fd,0) == 0xAA) {
		int len = SWAP16(RFIFOW(fd, 1)) + 3;
		if(len <= RFIFOREST(fd))
			RFIFOSKIP(fd, len);
		return 0;
	}

	if(RFIFOREST(fd) < logif_table_recv[RFIFOB(fd, 0)])
		return 0;

	switch (RFIFOB(fd,0)) {
	case 0x00:
		if (RFIFOB(fd,1)) {
			printf("CFG_ERR: Username or password to connect Login Server is invalid!\n");
			add_log("CFG_ERR: Username or password to connect Login Server is invalid!\n");
		} else {
			printf("Connected to Login Server.\n");
			add_log("Connected to Login Server.\n");
		}
		break;
	case 0x01:
		res=char_db_isnameused(RFIFOP(fd, 3));
		WFIFOB(fd, 0) = 0x01;
		WFIFOW(fd, 1) = RFIFOW(fd, 1);
		WFIFOB(fd, 3) = res;
		WFIFOSET(fd, 4);
		break;
	case 0x02:
		res=char_db_newchar(RFIFOP(fd, 3), RFIFOP(fd, 19), RFIFOB(fd, 38), RFIFOB(fd, 36), RFIFOB(fd, 37), RFIFOB(fd, 35), RFIFOB(fd,39));
		WFIFOB(fd, 0) = 0x02;
		WFIFOW(fd, 1) = RFIFOW(fd, 1);
		WFIFOB(fd, 3) = res;
		WFIFOSET(fd, 4);
		break;
	case 0x03:
		
		res=char_db_mapfifofromlogin(RFIFOP(fd, 3),RFIFOP(fd, 19),&id);
		if(res<0) {
			
			WFIFOB(fd, 0) = 0x03;
			WFIFOW(fd, 1) = RFIFOW(fd, 1);
			WFIFOB(fd, 3) = abs(res);
			WFIFOSET(fd, 26);
			break;
		}
		if(logindata_add(id, res, RFIFOP(fd, 3))) {
			//character is online, force disconnected
			WFIFOB(fd, 0) = 0x03;
			WFIFOW(fd, 1) = RFIFOW(fd, 1);
			WFIFOB(fd, 3) = 0x06;
			WFIFOSET(fd, 26);
			WFIFOB(map_fifo[res].fd, 0) = 0x04;
			WFIFOL(map_fifo[res].fd, 1) = id;
			WFIFOSET(map_fifo[res].fd, 5);
			break;
		}
		
		WFIFOB(map_fifo[res].fd, 0) = 0x02;
		WFIFOW(map_fifo[res].fd, 1) = RFIFOW(fd, 1);
		WFIFOL(map_fifo[res].fd, 3) = id;
		memset(WFIFOP(map_fifo[res].fd, 7), 0, 16);
		// copy name
		memcpy(WFIFOP(map_fifo[res].fd, 7), RFIFOP(fd, 3), 16);
		WFIFOL(map_fifo[res].fd, 33) = RFIFOL(fd, 35);
		WFIFOSET(map_fifo[res].fd, 37);
		break;
	case 0x04:
		res=char_db_setpass(RFIFOP(fd, 3), RFIFOP(fd, 19), RFIFOP(fd, 35));
		WFIFOB(fd, 0) = 0x04;
		WFIFOW(fd, 1) = RFIFOW(fd, 1);
		WFIFOB(fd, 3) = abs(res);
		WFIFOSET(fd, 4);
		break;
	default:
		add_dmp(fd, RFIFOREST(fd));
		RFIFOSKIP(fd, RFIFOREST(fd));
		return 0;
	}

	RFIFOSKIP(fd, logif_table_recv[RFIFOB(fd, 0)]);
	return 0;
}
