
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../common/core.h"
#include "../common/socket.h"
#include "../common/mmo.h"
#include "../common/malloc.h"
#include "char.h"
#include "mapif.h"
#include "char_db.h"

static int mapif_table_recv[0xFF] = 
{
	71, 0, 19, 23, 0, 5, 0, 0, 0, 0, //0-9
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 10-19
	0, 5, 5, 0, 0, 0, 0, 0, 0, 0 // 20-29
};

int mapif_parse_legendsave(int fd) {
	int len,i,j,c,ps;
	struct legend tmp_legend;
	c = RFIFOB(fd, 5);
	ps= RFIFOW(fd, 6);
	
	for(i=8,j=0; i<ps && j<c;) {
		if(len) {
			
			tmp_legend.legends[j].icon=RFIFOB(fd, i); i++;
			tmp_legend.legends[j].color=RFIFOB(fd, i); i++;
			len = RFIFOB(fd, i); i++;
			memcpy(tmp_legend.legends[j].text, RFIFOP(fd, i), len);
			i+=len;
			j++;
		}
	}
	RFIFOSKIP(fd, ps);
	return 0;
}
int mapif_parse_legend(int fd) {
	struct legend tmp_legend;
	int i,j,c,len;
	mmo_request_legend(RFIFOL(fd, 1),&tmp_legend);

	WFIFOB(fd, 0) = 22; // REPLY WITH REQUESTED DATA
	WFIFOL(fd, 1) = RFIFOL(fd, 1);
	WFIFOB(fd, 5) = c = tmp_legend.legend_num;
	for(i=0,j=8;i<c;i++) {
		WFIFOB(fd, j) = tmp_legend.legends[i].icon;
		j++;
		WFIFOB(fd, j) = tmp_legend.legends[i].color;
		j++;
		WFIFOB(fd, j) = len = tmp_legend.legends[i].len;
		j++;
		memcpy(WFIFOP(fd, j), tmp_legend.legends[i].text, len);
		j+=len;
	}
	WFIFOW(fd, 6) = j;
	WFIFOSET(fd, j);
	return 0;
}

int mapif_parse_requestreg(int fd) {
	struct registry tmp_reg;
	int i,j,c,len;
	mmo_request_reg(RFIFOL(fd, 1), &tmp_reg);
	WFIFOB(fd, 0) = 21;
	WFIFOL(fd, 1) = RFIFOL(fd, 1);
	WFIFOB(fd, 5) = c = tmp_reg.global_num;
	
	for(i=0,j=8;i<c;i++) {
		WFIFOB(fd, j) = len = strlen(tmp_reg.global[i].str);
		j++;
		memcpy(WFIFOP(fd, j), tmp_reg.global[i].str, len);
		j+=len;
		WFIFOB(fd, j) = len = strlen(tmp_reg.global[i].value);
		j++;
		memcpy(WFIFOP(fd, j), tmp_reg.global[i].value, len);
		j+=len;
	}
	WFIFOW(fd, 6) = j;
	WFIFOSET(fd, j);
	return 0;
}
int mapif_parse_regsave(int fd) {
	struct registry tmp_reg;
	int packet_size = RFIFOW(fd, 6);
	int count,i,len;
	
	tmp_reg.global_num = RFIFOB(fd, 5);
	for(i=8,count=0; i<packet_size && tmp_reg.global_num > count; i++) {	
		len = RFIFOB(fd, i);
		i++;
		memcpy(tmp_reg.global[count].str, RFIFOP(fd, i), len);
		tmp_reg.global[count].str[len]=0;
		i+=len;
		len = RFIFOB(fd, i);
		i++;
		memcpy(tmp_reg.global[count].value, RFIFOP(fd, i), len);
		tmp_reg.global[count].value[len]=0;
		i+=len;
		count++;
	}

	mmo_save_reg(RFIFOL(fd, 1), &tmp_reg);
	RFIFOSKIP(fd, packet_size);
}
int mapif_send(int fd, unsigned char* buf, int len, int type) {
	int i;

	switch(type) {
	case ALL:
		for(i = 0; i < map_fifo_max; i++) {
			if (map_fifo[i].fd > 0) {
				memcpy(WFIFOP(map_fifo[i].fd,0), buf, len);
				WFIFOSET(map_fifo[i].fd,len);
			}
		}
		break;
	case ALLWOS:
		for(i = 0; i < map_fifo_max; i++) {
			if ((map_fifo[i].fd > 0) && (map_fifo[i].fd != fd)) {
				memcpy(WFIFOP(map_fifo[i].fd,0), buf, len);
				WFIFOSET(map_fifo[i].fd,len);
			}
		}
		break;
	}

	return 0;
}

int mapif_parse_auth(int fd) {
	int i;
	unsigned char *p;
	if(session[fd]->eof) {
		
		do_close(fd);
		return 0;
	}

	if(RFIFOREST(fd) < 71)
		return 0;

	if(!RFIFOB(fd,0)) {
		if ((strcmp(RFIFOP(fd,1),char_id)) && (strcmp(RFIFOP(fd,33),char_pw))) {
			WFIFOB(fd,0) = 0x00;
			WFIFOB(fd,1) = 0x01;
			WFIFOSET(fd, 2);
			session[fd]->eof = 1;
			RFIFOSKIP(fd, 71);
			return 0;
		}
		

		for(i=0;i < map_fifo_max;i++) {
			if(!map_fifo[i].fd)
				break;
		}

		if(i>=map_fifo_max)
			map_fifo_max++;

		map_fifo_n++;
		map_fifo[i].fd = fd;
		map_fifo[i].ip = RFIFOL(fd, 65);
		map_fifo[i].port = RFIFOW(fd, 69);
		session[fd]->func_parse = mapif_parse;
		realloc_fifo(fd, FIFOSIZE_SERVER, FIFOSIZE_SERVER);
		WFIFOB(fd,0) = 0x00;
		WFIFOB(fd,1) = 0x00;
		WFIFOSET(fd, 2);

		p=(unsigned char*)&map_fifo[i].ip;
		printf("Map Server #%d connected.(%d.%d.%d.%d:%d).\n", i, p[0], p[1], p[2], p[3], map_fifo[i].port);
		add_log("Map Server #%d connected(%d.%d.%d.%d:%d).\n", i, p[0], p[1], p[2], p[3], map_fifo[i].port);
		
		//prepare mmo size
		mapif_table_recv[4] = sizeof(struct mmo_charstatus)+1;
		
	} else {
		session[fd]->eof = 1;
	}

	RFIFOSKIP(fd, 71);
	return 0;
}

int mapif_parse(int fd) {
	int id,i;
	unsigned int j;
	
	for(id=0;id<map_fifo_max;id++) {
		if(map_fifo[id].fd == fd)
			break;
	}

	if(session[fd]->eof) {
		add_log("Map Server #%d connection lost.\n",id);
		printf("Map Server #%d connection lost.\n",id);
		map_fifo[id].fd = 0;
		map_fifo_n--;
		if(id==(map_fifo_max-1)) {
			for(i=map_fifo_max-1;i>=0;i--) {
				if(!map_fifo[id].fd)
					map_fifo_max--;
				else
					break;
			}
		}
		//printf("Map FIFO Server Connected: %d, FIFO Cache: %d\n", map_fifo_n, map_fifo_max);
		mmo_setallonline(0);
		do_close(fd);
		return 0;
	}

	if (RFIFOREST(fd) < mapif_table_recv[RFIFOB(fd, 0)])
		return 0;

	switch (RFIFOB(fd,0)) {
	case 0x01:
		if (RFIFOREST(fd) < RFIFOW(fd, 1)*2+3)
			return 0;

		map_fifo[id].map_n = RFIFOW(fd, 1);
		for(i=0;i<map_fifo[id].map_n;i++)
			map_fifo[id].map[i] = RFIFOW(fd, i*2+3);
		{
			unsigned char buf[MAX_MAP_PER_SERVER*2+7];
			int j,k;
			//send my presence
			if (map_fifo[id].map_n > 0) {
				WBUFB(buf, 0) = 0x01;
				WBUFL(buf, 1) = map_fifo[id].ip;
				WBUFW(buf, 5) = map_fifo[id].port;
				memcpy(WBUFP(buf,7), RFIFOP(fd,1), map_fifo[id].map_n*2+2);
				mapif_send(fd, buf, map_fifo[id].map_n*2+9, ALLWOS);
			}
			//get other map server presence
			for(i=0;i<map_fifo_max;i++) {
				if ((i != id) && (map_fifo[i].fd > 0) && (map_fifo[i].map_n > 0)) {
					WFIFOB(fd, 0) = 0x01;
					WFIFOL(fd, 1) = map_fifo[i].ip;
					WFIFOW(fd, 5) = map_fifo[i].port;
					WFIFOW(fd, 7) = map_fifo[i].map_n;
					for(j=0;j<map_fifo[i].map_n;j++) {
						WFIFOW(fd, j*2+9) = map_fifo[i].map[j];
						for(k=0;(k<map_fifo[id].map_n) && (k<map_fifo[i].map_n);k++) {
							if (map_fifo[i].map[j] == map_fifo[id].map[k]) {
								//found same map, just report and let admin fix it...
								printf("MAPERR: Same map number(#%d) at server #%d and server #%d\n", map_fifo[id].map[k], id, i);
								add_log("MAPERR: Same map number(#%d) at server #%d and server #%d\n", map_fifo[id].map[k], id, i);
							}
						}
					}
					WFIFOSET(fd, map_fifo[i].map_n*2+9);
				}
			}
			//done!
		}
		printf("Map Server #%d(%d map) is \033[1;32mready\033[0m!\n", id, map_fifo[id].map_n);
		add_log("Map Server #%d send %d map.\n", id, map_fifo[id].map_n, id);
		RFIFOSKIP(fd, map_fifo[id].map_n*2+3);
		return 0;
	case 0x02:
		WFIFOB(login_fd, 0) = 0x03;
		WFIFOW(login_fd, 1) = RFIFOW(fd, 1);
		WFIFOB(login_fd, 3) = 0x00;
		memset(WFIFOP(login_fd, 4), 0, 16);
		memcpy(WFIFOP(login_fd, 4), RFIFOP(fd, 3), 16);
		WFIFOL(login_fd, 20) = map_fifo[id].ip;
		WFIFOW(login_fd, 24) = map_fifo[id].port;
		WFIFOSET(login_fd, 26);
		break;
	case 0x03:
		{
			// Null pointer bug Fixed. [Lene]
			// We don't need to declare char_dat as global. [Lene]
			struct mmo_charstatus *char_dat;
			CREATE(char_dat, struct mmo_charstatus, 1);

			mmo_char_fromdb(RFIFOL(fd, 3), char_dat, RFIFOP(fd, 7));
			WFIFOB(fd, 0) = 0x03;
			WFIFOW(fd, 1) = RFIFOW(fd, 1);
			
			memcpy(WFIFOP(fd, 3), char_dat, sizeof(struct mmo_charstatus));
			WFIFOSET(fd, sizeof(struct mmo_charstatus)+3);
			aFree(char_dat);
		}
		break;
	case 0x04:
		mmo_char_todb((struct mmo_charstatus*)RFIFOP(fd, 1));
		break;
	case 0x05:
		logindata_del(RFIFOL(fd, 1));
		break;
	case 0x06:
		if (RFIFOREST(fd) < RFIFOW(fd, 33)+35)
			return 0;

		j=char_db_idfromname(RFIFOP(fd, 17));
		if (!j) {
			WFIFOB(fd, 0) = 0x06;
			WFIFOB(fd, 1) = 0x02;
			memcpy(WFIFOP(fd, 2), RFIFOP(fd, 1), 32);
			WFIFOW(fd, 34) = 0x00;
			WFIFOSET(fd, 36);
		} else {
			struct char_login_data *dat;
			dat = logindata_search(j);
			if (!dat) {
				WFIFOB(fd, 0) = 0x06;
				WFIFOB(fd, 1) = 0x02;
				memcpy(WFIFOP(fd, 2), RFIFOP(fd, 1), 32);
				WFIFOW(fd, 34) = 0x00;
				WFIFOSET(fd, 36);
			} else {
				//Fix nickname, an easy way is change from RFIFO
				memcpy(RFIFOP(fd, 17), dat->name, 16);
				WFIFOB(map_fifo[dat->map_server].fd, 0) = 0x06;
				WFIFOB(map_fifo[dat->map_server].fd, 1) = 0x00;
				memcpy(WFIFOP(map_fifo[dat->map_server].fd, 2), RFIFOP(fd, 1), RFIFOW(fd, 33)+34);
				WFIFOSET(map_fifo[dat->map_server].fd, RFIFOW(fd, 33)+36);
				WFIFOB(fd, 0) = 0x06;
				WFIFOB(fd, 1) = 0x01;
				memcpy(WFIFOP(fd, 2), RFIFOP(fd, 1), RFIFOW(fd, 33)+34);
				WFIFOSET(fd, RFIFOW(fd, 33)+36);
			}
		}
		RFIFOSKIP(fd, RFIFOW(fd, 33)+35);
		return 0;

	case 21: // request char_reg
		mapif_parse_requestreg(fd);
		break;

	case 20: //
		mapif_parse_regsave(fd);
		return 0;

	case 22: // request for legend data
		mapif_parse_legend(fd);
		break;
	case 23: // request for saving legend
		mapif_parse_legendsave(fd);
		return 0;
	default:
		add_dmp(fd, RFIFOREST(fd));
		RFIFOSKIP(fd, RFIFOREST(fd));
		return 0;
	}

	RFIFOSKIP(fd, mapif_table_recv[RFIFOB(fd, 0)]);
	return 0;
}
