
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#ifndef _WIN32
#include <sys/time.h>
#include <unistd.h>
#endif 
#include <time.h>

#include <signal.h>
#include <string.h>

#include "core.h"
#include "socket.h"
#include "timer.h"
#include "version.h"

static void (*term_func)(void)=NULL;
int run=1;

char dmp_filename[128];
char log_filename[128];
char date_format[32] = "%Y-%m-%d %H:%M:%S";
char SERVER_NAME[32]= "GAYA_UNKNOWN";
// Set Dump Packet File
//----------------------------
void set_dmpfile(char *dmpfilename)
{
	memset(dmp_filename, 0, 128);
	strcpy(dmp_filename, dmpfilename);
}

// Add packet to dump
// Saved from RFIFO
//----------------------------
void add_dmp(int fd, int len)
{
	FILE *dmpfp;
	int i;
	struct timeval tv;
	char timetmp[128];

	unsigned char *p = (unsigned char *) &session[fd]->client_addr.sin_addr;
	char ip[16];
	sprintf(ip, "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);

	dmpfp = fopen(dmp_filename, "a");
	if (dmpfp) {
		gettimeofday(&tv, NULL);
		strftime(timetmp, 24, date_format, localtime(&(tv.tv_sec)));
		fprintf(dmpfp, "%s IP: %s len %d\n", timetmp, ip, len);
		fprintf(dmpfp, "[HEX]:");
		for(i=0;i < len;i++) {
			fprintf(dmpfp, "[%02X]", RFIFOB(fd, i));
		}
		fprintf(dmpfp, "\n[CHR]:");
		for(i=0;i < len;i++) {
			fprintf(dmpfp, "[ %c]", RFIFOB(fd, i));
		}
		fprintf(dmpfp, "\n");
	}
	fclose(dmpfp);
}

// Set Log File
//----------------------------
void set_logfile(char *logfilename)
{
	memset(log_filename, 0, 128);
	strcpy(log_filename, logfilename);
}

// Add log line
//----------------------------
void add_log(char *fmt, ...)
{
	FILE *logfp;
	va_list ap;
	struct timeval tv;
	char tmpstr[128];

	va_start(ap, fmt);

	logfp = fopen(log_filename, "a");
	if (logfp) {
		if (fmt[0] == '\0') // jump a line if no message
			fprintf(logfp, "\n");
		else {
			gettimeofday(&tv, NULL);
			strftime(tmpstr, 24, date_format, localtime(&(tv.tv_sec)));
			sprintf(tmpstr + strlen(tmpstr), ".%03d: %s", (int)tv.tv_usec / 1000, fmt);
			vfprintf(logfp, tmpstr, ap);
		}
		fclose(logfp);
	}

	va_end(ap);
}

// Set terminate function
//----------------------------
void set_termfunc(void (*termfunc)(void))
{
	term_func = termfunc;
}

// Signal handling
//----------------------------
static void sig_proc(int sn)
{
	int i;
	switch(sn){
	case SIGINT:
	case SIGTERM:
		
		for(i=0;i<fd_max;i++){
			if(!session[i])
				continue;
			//close(i);
			do_close(i);
		}
		run=0;
		break;
	}
}

// Display Titlename
//----------------------------
static void display_title(void)
{
	ShowInfo("Gaya [%d.%d.%02d]\n\n", YURI_MAJOR_VERSION, YURI_MINOR_VERSION, YURI_PATCH_VERSION);
#if YURI_RELEASE_FLAG
	ShowInfo("**%s Developing version...\n",SERVER_NAME);
#endif
}

// Main Routine
//----------------------------
int main(int argc,char **argv)
{
	int next;
	

	SET_SERVERNAME();
	socket_init();
	malloc_init();
	timer_init();
#ifdef SIGPIPE
	signal(SIGPIPE,SIG_IGN);
#endif
	signal(SIGTERM,sig_proc);
	signal(SIGINT,sig_proc);

	

	do_init(argc,argv);
	display_title();
	while(run){
		next = do_timer(gettick_nocache());
		do_sendrecv(next);
		do_parsepacket();
	}
	
	do_term();
	socket_final();
	timer_final();
	malloc_final();
	
	
	return 0;
}
