
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../common/core.h"
#include "../common/socket.h"
#include "../common/mmo.h"
#include "../common/crypt.h"
#include "login.h"


#include "intif.h"
#include "clif.h"

static int intif_table_recv[5] = {69, 4, 4, 26, 4};

int intif_auth(int fd) {
	if(char_fd > 0) {
		do_close(fd);
		return 0;
	}

	if (RFIFOREST(fd) < intif_table_recv[0]) {
		do_close(fd);
		return 0;
	}

	if ((strcmp(RFIFOP(fd,5),login_id)) && (strcmp(RFIFOP(fd,37),login_pw))) {
		WFIFOB(fd,0) = 0x00;
		WFIFOB(fd,1) = 0x01;
		WFIFOSET(fd, 2);
		session[fd]->eof=1;
		return 0;
	}

	char_fd = fd;
	session[char_fd]->func_parse = intif_parse;
	realloc_fifo(char_fd, FIFOSIZE_SERVER, FIFOSIZE_SERVER);
	WFIFOB(char_fd,0) = 0x00;
	WFIFOB(char_fd,1) = 0x00;
	WFIFOSET(char_fd, 2);

	add_log("Connection from Char Server accepted.\n");
	printf("Connection from Char Server accepted.\n"); 
	return 0;
}

int intif_parse(int fd) {
	if(session[fd]->eof) {
		add_log("Char Server connection lost.\n");
		printf("Char Server connection lost.\n");
		char_fd = 0;
		do_close(fd);
		return 0;
	}

	if(RFIFOB(fd,0) == 0xAA) {
		int len = SWAP16(RFIFOW(fd, 1)) + 3;
		if(len <= RFIFOREST(fd))
			RFIFOSKIP(fd, len);
		return 0;
	}

	if(RFIFOREST(fd) < intif_table_recv[RFIFOB(fd, 0)])
		return 0;

	switch (RFIFOB(fd,0)) {
	case 0x01:
		if (!session[RFIFOW(fd, 1)])
			break;
		if (RFIFOB(fd, 3)==0x01)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_USEREXIST]);
		else if (RFIFOB(fd, 3))
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_ERRDB]);
		else
			clif_message(RFIFOW(fd, 1), 0x00, "\x00");
		break;
	case 0x02:
		if (!session[RFIFOW(fd, 1)])
			break;
	
		if (RFIFOB(fd, 3) == 0x01)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_USEREXIST]);
		else if (RFIFOB(fd, 3))
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_ERRDB]);
		else
			clif_message(RFIFOW(fd, 1), 0x00, login_msg[LGN_NEWCHAR]);
		break;
	case 0x03:
		if (!session[RFIFOW(fd, 1)])
			break;

		if (!RFIFOB(fd, 3)) {
			int len;
			WFIFOB(RFIFOW(fd, 1), 0) = '\xAA';
			WFIFOB(RFIFOW(fd, 1), 1) = '\x00';
			WFIFOB(RFIFOW(fd, 1), 2) = '\x05';
			WFIFOB(RFIFOW(fd, 1), 3) = '\x02';
			WFIFOB(RFIFOW(fd, 1), 4) = '\x04';
			WFIFOB(RFIFOW(fd, 1), 5) = '\x4A';
			WFIFOB(RFIFOW(fd, 1), 6) = '\x61';
			WFIFOB(RFIFOW(fd, 1), 7) = '\x7C';
			WFIFOSET(RFIFOW(fd, 1), 8);
			
			WFIFOB(RFIFOW(fd, 1), 0) = '\xAA';
			WFIFOB(RFIFOW(fd, 1), 3) = '\x03';
			WFIFOL(RFIFOW(fd, 1), 4) = SWAP32(RFIFOL(fd, 20));
			WFIFOW(RFIFOW(fd, 1), 8) = SWAP16(RFIFOW(fd, 24));
			len = strlen(RFIFOP(fd, 4));
			WFIFOB(RFIFOW(fd, 1), 10) = len;
			memcpy(WFIFOP(fd, 11), RFIFOP(fd, 4), len+1);
			WFIFOW(RFIFOW(fd, 1), 1) = SWAP16(len+9);
			WFIFOSET(RFIFOW(fd, 1), 12+len);
		} else if (RFIFOB(fd, 3) == 0x01)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_ERRDB]);
		else if (RFIFOB(fd, 3) == 0x02)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_WRONGUSER]);
		else if (RFIFOB(fd, 3) == 0x03)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_WRONGPASS]);
		else if (RFIFOB(fd, 3) == 0x04)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_BANNED]);
		else if (RFIFOB(fd, 3) == 0x05)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_ERRSERVER]);
		else if (RFIFOB(fd, 3) == 0x06)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_DBLLOGIN]);
		break;
	case 0x04:
		if (!session[RFIFOW(fd, 1)])
			break;

		if (!RFIFOB(fd, 3))
			clif_message(RFIFOW(fd, 1), 0x00, login_msg[LGN_CHGPASS]);
		else if (RFIFOB(fd, 3) == 0x01)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_ERRDB]);
		else if (RFIFOB(fd, 3) == 0x02)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_WRONGUSER]);
		else if (RFIFOB(fd, 3) == 0x03)
			clif_message(RFIFOW(fd, 1), 0x03, login_msg[LGN_WRONGPASS]);
		break;
	default:
		add_dmp(fd, RFIFOREST(fd));
		RFIFOSKIP(fd, RFIFOREST(fd));
		return 0;
	}

	RFIFOSKIP(fd, intif_table_recv[RFIFOB(fd, 0)]);
	return 0;
}

