
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../common/core.h"
#include "../common/socket.h"
#include "../common/timer.h"
#include "../common/crypt.h"

#include "version.h"
#include "login.h"
#include "clif.h"

int login_port = 2010;

int login_fd;
int char_fd;

char login_msg[MSG_MAX][256];

char login_id[32];
char login_pw[32];
int dump_save = 0;

int string_check(const char *p,int len) {
	int i;
	for(i=0;i<len;i++) {
		if(!isalnum(p[i]))
			return 1;
	}
	return 0;
}

int lang_read(const char *cfg_file) {
	char line[1024], r1[1024], r2[1024];
	int line_num = 0;
	FILE *fp;

	fp = fopen(cfg_file, "r");
	if (fp == NULL) {
		printf("CFG_ERR: Language file (%s) not found.\n", cfg_file);
		return 1;
	}

	while(fgets(line, sizeof(line), fp)) {
		line_num++;
		if (line[0] == '/' && line[1] == '/')
			continue;

		if (sscanf(line, "%[^:]: %[^\r\n]", r1, r2) == 2) {
			if(strcmpi(r1, "LGN_ERRSERVER") == 0) {
				strncpy(login_msg[LGN_ERRSERVER], r2, 256);
				login_msg[LGN_ERRSERVER][255] = '\0';
			} else if(strcmpi(r1, "LGN_WRONGPASS") == 0) {
				strncpy(login_msg[LGN_WRONGPASS], r2, 256);
				login_msg[LGN_WRONGPASS][255] = '\0';
			} else if(strcmpi(r1, "LGN_WRONGUSER") == 0) {
				strncpy(login_msg[LGN_WRONGUSER], r2, 256);
				login_msg[LGN_WRONGUSER][255] = '\0';
			} else if(strcmpi(r1, "LGN_ERRDB") == 0) {
				strncpy(login_msg[LGN_ERRDB], r2, 256);
				login_msg[LGN_ERRDB][255] = '\0';
			} else if(strcmpi(r1, "LGN_USEREXIST") == 0) {
				strncpy(login_msg[LGN_USEREXIST], r2, 256);
				login_msg[LGN_USEREXIST][255] = '\0';
			} else if(strcmpi(r1, "LGN_ERRPASS") == 0) {
				strncpy(login_msg[LGN_ERRPASS], r2, 256);
				login_msg[LGN_ERRPASS][255] = '\0';
			} else if(strcmpi(r1, "LGN_ERRUSER") == 0) {
				strncpy(login_msg[LGN_ERRUSER], r2, 256);
				login_msg[LGN_ERRUSER][255] = '\0';
			} else if(strcmpi(r1, "LGN_NEWCHAR") == 0) {
				strncpy(login_msg[LGN_NEWCHAR], r2, 256);
				login_msg[LGN_NEWCHAR][255] = '\0';
			} else if(strcmpi(r1, "LGN_CHGPASS") == 0) {
				strncpy(login_msg[LGN_CHGPASS], r2, 256);
				login_msg[LGN_CHGPASS][255] = '\0';
			} else if(strcmpi(r1, "LGN_DBLLOGIN") == 0) {
				strncpy(login_msg[LGN_DBLLOGIN], r2, 256);
				login_msg[LGN_DBLLOGIN][255] = '\0';
			} else if(strcmpi(r1, "LGN_BANNED") == 0) {
				strncpy(login_msg[LGN_BANNED], r2, 256);
				login_msg[LGN_BANNED][255] = '\0';
			} else if(strcmpi(r1, "LGN_ID_TOOLONG") == 0) {
				strncpy(login_msg[LGN_ID_TOOLONG], r2, 256);
				login_msg[LGN_ID_TOOLONG][255] = '\0';
			} else if(strcmpi(r1, "LGN_ID_TOOSHORT") == 0) {
				strncpy(login_msg[LGN_ID_TOOSHORT], r2, 256);
				login_msg[LGN_ID_TOOSHORT][255] = '\0';
			} else if(strcmpi(r1, "LGN_PW_TOOSHORT") == 0) {
				strncpy(login_msg[LGN_PW_TOOSHORT], r2, 256);
				login_msg[LGN_PW_TOOSHORT][255] = '\0';
			} else if(strcmpi(r1, "LGN_PW_TOOLONG") == 0) {
				strncpy(login_msg[LGN_PW_TOOLONG], r2, 256);
				login_msg[LGN_PW_TOOLONG][255] = '\0';
			} else if(strcmpi(r1, "LGN_ID_HANGUL_ONLY") == 0) {
				strncpy(login_msg[LGN_ID_HANGUL_ONLY], r2, 256);
				login_msg[LGN_ID_HANGUL_ONLY][255] = '\0';
			}
		}
	}
	fclose(fp);
	printf("Language File (%s) reading finished!\n", cfg_file);
	return 0;
}
	
int config_read(const char *cfg_file) {
	char line[1024], r1[1024], r2[1024];
	int line_num = 0;
	FILE *fp;

	fp = fopen(cfg_file, "r");
	if (fp == NULL) {
		printf("CFG_ERR: Configuration file (%s) not found.\n", cfg_file);
		return 1;
	}

	while(fgets(line, sizeof(line), fp)) {
		line_num++;
		if (line[0] == '/' && line[1] == '/')
			continue;

		if (sscanf(line, "%[^:]: %[^\r\n]", r1, r2) == 2) {
			if(strcmpi(r1, "login_port") == 0) {
				login_port = atoi(r2);
			} else if(strcmpi(r1, "login_id") == 0) {
				strncpy(login_id, r2, 32);
				login_id[31] = '\0';
			} else if(strcmpi(r1, "login_pw") == 0) {
				strncpy(login_pw, r2, 32);
				login_pw[31] = '\0';
			} else if(strcmpi(r1, "login_log") == 0) {
				set_logfile(r2);
			} else if(strcmpi(r1, "dump_log") == 0) {
				set_dmpfile(r2);
			} else if(strcmpi(r1, "dump_save") == 0) {
				dump_save = atoi(r2);
			}
		}
	}
	fclose(fp);
	printf("Configuration File (%s) reading finished!\n", cfg_file);
	return 0;
}

void do_term(void) {
	printf("Gaya Login Server Shutdown.\n");
	add_log("Shutdown.\n");
}

void help_screen() {
	printf("HELP LIST\n");
	printf("---------\n");
	printf(" --conf [FILENAME]  : set config file\n");
	printf(" --inter [FILENAME] : set inter file\n");
	printf(" --lang [FILENAME]  : set lang file\n");
	exit(0);
}
void SET_SERVERNAME()
{
	strcpy(SERVER_NAME,"Gaya Login-server");
}

int do_init(int argc, char **argv) {
	int i;
	char *CONF_FILE = "conf/login.conf";
	char *LANG_FILE = "conf/lang.conf";
	char *INTER_FILE = "conf/inter.conf";

	srand(gettick());
	set_logfile("log/login.log");
	set_dmpfile("log/login_dump.log");
	
	for (i = 1; i < argc ; i++) {
		if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "--h") == 0 || strcmp(argv[i], "--?") == 0 || strcmp(argv[i], "/?") == 0)
			help_screen();
		else if (strcmp(argv[i], "--conf") == 0)
			CONF_FILE=argv[i+1];
		else if (strcmp(argv[i],"--inter") == 0)
			INTER_FILE = argv[i+1];
		else if (strcmp(argv[i],"--lang") == 0)
			LANG_FILE = argv[i+1];
	}

	config_read(CONF_FILE);
	config_read(INTER_FILE);
	lang_read(LANG_FILE);
	set_termfunc(do_term);

	add_log("");
	add_log("Gaya Login Server Started.\n");

	set_defaultaccept(clif_accept);
	set_defaultparse(clif_parse);
	login_fd = make_listen_port(login_port);
	stall_time=86400;
	printf("Gaya Login Server is \033[1;32mready\033[0m! Listening at %d.\n", login_port);
	add_log("Server Ready! Listening at %d.\n", login_port);
	return 0;
}
