
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "../common/core.h"
#include "../common/socket.h"
#include "../common/crypt.h"
#include "../common/malloc.h"
#include "../common/mmo.h"
#include "../common/timer.h"

#include "clif.h"
#include "map.h"
#include "intif.h"
#include "pc.h"
#include "itemdb.h"
#include "command.h"
#include "mob.h"
#include "npc.h"
#include "magic.h"
#include "warp.h"
#include "class_db.h"
#include "../common/nullpo.h"

int clif_sendoption(struct map_sessiondata*);

struct userlist_data userlist;
//CRC Functions

unsigned long reflect(unsigned short crc, int bitnum)
{ // reflects the lower 'bitnum' bits of 'crc'
  unsigned short i, j=1, crcout=0;

  for(i=(unsigned short)1<<(bitnum-1); i; i>>=1) {
    if(crc & i)
      crcout |= j;
    j <<= 1;
  }
  return( crcout );
}

void clif_message(int fd, char type, char *msg) {
	int encs = 0;
	int len=strlen(msg);
	
	WFIFOB(fd,0) = 0xAA;
	WFIFOW(fd,1) = SWAP16(len+3);
	WFIFOB(fd,3) = 0x0A;
	WFIFOB(fd,4) = 0x00;
	WFIFOB(fd,5) = type;
	WFIFOW(fd,6) = SWAP16(len);

	memcpy(WFIFOP(fd,7),msg,len);
	WFIFOSET(fd, crypt(WFIFOP(fd,0)));
}

unsigned short crcbitbybit(unsigned char *p, unsigned short len)
{ /* bit by bit algorithm with augmented zero bytes. */
  unsigned short i, j, c, bit;
  unsigned short crc = 0;

  for(i=0; i<len; i++) {
    c = (unsigned short)*p++;
    c = reflect(c, 8);

    for (j=0x80; j; j>>=1) {
      bit = crc & 0x8000;
      crc <<= 1;
      if(c & j)
        crc |= 1;
      if(bit)
        crc ^= 0xA001;
    }
  }     

  for(i=0; i < 16; i++) {
    bit = crc & 0x8000;
    crc <<= 1;
    if(bit)
      crc ^= 0xA001;
  }
  crc = reflect(crc, 16);
  return(crc);
}


//End CRC Functions

int clif_user_list(struct map_sessiondata* sd) {
	char buf[1024]={0,};
	
	int i;
	int len;
//	int path;
	struct map_sessiondata *tsd;
//	int x;
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,3)=0x36;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOW(sd->fd,5)=SWAP16(userlist.user_count); //max users
	WFIFOW(sd->fd,7)=SWAP16(userlist.user_count); //users on server
	WFIFOB(sd->fd,9)=0;
	len=0;
	
	for(i=0;i<(userlist.user_count);i++) {

		tsd=map_id2sd(userlist.user[i]);
		if(tsd==NULL) continue;
		if(tsd->status.class==0) {
			WFIFOB(sd->fd,len+10)=16*tsd->status.country+5;
		}
		else {
			WFIFOB(sd->fd,len+10)=16*tsd->status.country+tsd->status.class;
		}
		WFIFOB(sd->fd,len+11)=tsd->status.subclass*16;
		WFIFOB(sd->fd,len+12)=0x8F;
		WFIFOB(sd->fd,len+13)=strlen(tsd->status.name);
		strcpy(WFIFOP(sd->fd,len+14),tsd->status.name);
		len=len+strlen(tsd->status.name)+4;
	}
	
	WFIFOW(sd->fd,1)=SWAP16(len+7);
	crypt(WFIFOP(sd->fd,0));
	//strcpy(WFIFOP(sd->fd,0),buf);
	
	WFIFOSET(sd->fd,len+10);
	return 0;
}

int pc_battle_values(struct map_sessiondata *sd,struct map_sessiondata *mob,int dm) {
	int damage=dm;
	int dx;
	int dy;
	
	dx=abs(sd->bl.x-mob->bl.x);
	dy=abs(sd->bl.y-mob->bl.y);
	if(sd->rage>0) damage=damage*sd->rage;
	if(sd->status.state==2 || sd->status.state==5) { 
		damage=damage*8;
	}
	
	
	if(sd->status.state==2 || sd->status.state==5){
		sd->status.state=0;
		clif_sendstatus(sd);
	}
	sd->tooclose=1;
	
	
	return damage;
}
int pc_damage(struct block_list *bl,va_list ap) {
	struct map_sessiondata *src;
	struct map_sessiondata *sd;
	int damage,hit;
	damage=va_arg(ap,int);
	nullpo_ret(sd=va_arg(ap,struct map_sessiondata*));
	nullpo_ret(src=(struct map_sessiondata*)bl);
	
	
	
	//printf("Src type: %d\n",sd->status.hp);
	if(src->status.state==1) return 0;
	
	
	damage=pc_battle_values(sd,src,damage);
	hit=clif_calc_critical(sd);
	sd->target=src->bl.id;
	run_script(itemdb_script(sd->status.equip[EQ_WEAP].id),0,sd->bl.id,0);
	
	
	if(hit>0) {
		if(src->status.side==sd->status.side) {
			damage=damage*2;
			}
		else if(abs(src->status.side-sd->status.side)==2) {
		}
		else {
			damage=damage*1.5;
		}
		if(hit==2) {
			damage=damage*3;
		}
		
		pc_damaged(sd,damage);
		clif_sendstatus(src);
		if(hit==1) { 
			clif_send_pc_health(src,sd,1,33);
		} else { 
			clif_send_pc_health(src,sd,1,255);
		}
		

	}	
	
	return 0;
}

int pc_damage2(struct block_list *bl,struct map_sessiondata *sd, int damage) {
	struct map_sessiondata *src;
	int hit;
	nullpo_ret(src=(struct map_sessiondata*)bl);
	
	
	
	//printf("Src type: %d\n",sd->status.hp);
	if(src->status.state==1) return 0;
	
	
	damage=pc_battle_values(sd,src,damage);
//	printf("pc_battle_values() returned %d\n",damage);
	hit=clif_calc_critical(sd);
	sd->target=src->bl.id;
	run_script(itemdb_script(sd->status.equip[EQ_WEAP].id),0,sd->bl.id,0);
	
	
	if(hit>0) {
		if(src->status.side==sd->status.side) {
			damage=damage*2;
		}
		else if(abs(src->status.side-sd->status.side)!=2) {
			damage=damage*1.5;
		}
		if(hit==2) {
			damage=damage*3;
		}
		
		pc_damaged(src,damage);
		clif_sendstatus(src);
		if(sd->status.equip[EQ_WEAP].id) clif_sendsound(sd, 196957); 
		if(hit==1) { 
			clif_send_pc_health(src,sd,damage,33);
		} else { 
			clif_send_pc_health(src,sd,damage,255);
		}
		
		
	}	
	
	return 0;
}

int clif_send_pc_health(struct map_sessiondata* src, struct map_sessiondata *sd,int damage,int critical){
	
	int maxvita;
	int currentvita;
	float percentage;
	char buf[15];
	maxvita=src->max_hp;
	currentvita=src->status.hp;
	
	if(currentvita<=0){
		percentage=0;
	}else{
	percentage=(float)currentvita/(float)maxvita;
	percentage=(float)percentage*100;
	}
	
	
	
	
	

	WBUFB(buf,0)=0xAA;
	WBUFW(buf,1)=SWAP16(12);
	WBUFB(buf,3)=0x13;
	WBUFB(buf,4)=0x03;
	WBUFL(buf,5)=SWAP32(src->bl.id);
	WBUFB(buf,9)=critical;
	WBUFB(buf,10)=percentage;
	WBUFL(buf,11)=SWAP32(damage);
	crypt(WBUFP(buf,0));
	clif_send(buf,15,&src->bl,AREA);
	
	if(currentvita<=0){
		src->status.state=1;
		src->status.hp=0;
		pc_uncast(src);
		clif_sendstatus(src);
		clif_sendchararea(src);
		clif_getchararea(src);
	}
	
	return 0;
}
int clif_findspell_pos(struct map_sessiondata* sd,int id) {
	int x;
	for(x=0;x<26;x++) {
		if(sd->status.skill[x]==id) {
			return x;
		}
	}
	
	return -1;
}

int clif_calc_critical(struct map_sessiondata* sd) {
	int hit;
	int chance;

	hit=sd->hit*2+5;
	chance=rand()%100;
	if(hit>95) hit=95;
	if(chance<hit) {
		return 2; //critical!
	} else {
		
	   //hit=sd->status.grace*2;
	   //chance=rand()%100;
	   //if(chance<hit) {
		return 1; //normal!
	   //}
	}

	return 0; //nothing....
}
int clif_has_aethers(struct map_sessiondata *sd,int spell) {
	int x;
	for(x=0;x<50;x++) {
		if(sd->status.dura_aether[x].id==spell) {
			return sd->status.dura_aether[x].aether;
		}
	}

	return 0;
}

int clif_handle_bfp(struct map_sessiondata *sd,int side, int damage) {
	struct block_list *bl;
	
	switch(side){
		
	case 0: //UP
		bl=map_firstincell_Reverse(sd->bl.m,sd->bl.x,sd->bl.y-1,BL_MOB);
		if(bl==NULL) return 0;
		if( bl->type== BL_MOB ) clif_mob_damage2(bl,sd,damage);
		else if( bl->type== BL_PC ) pc_damage2(bl,sd,damage);
		
		//PVP
//		map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x,sd->bl.y-1,BL_MOB,sd);
//		map_foreachincell(pc_damage,sd->bl.m,sd->bl.x,sd->bl.y-1,BL_PC,damage,sd);
/*		if(sd->tooclose==0) {
			if(sd->polearm>0) {
				
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x,sd->bl.y-2,BL_MOB,sd);
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x-1,sd->bl.y-1,BL_MOB,sd);
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x+1,sd->bl.y-1,BL_MOB,sd);
			}
		}
		*/		break;
	case 1: //RIGHT
		bl=map_firstincell_Reverse(sd->bl.m,sd->bl.x+1,sd->bl.y,BL_MOB);
		if(bl==NULL) return 0;
		if( bl->type== BL_MOB ) clif_mob_damage2(bl,sd,damage);
		else if( bl->type== BL_PC ) pc_damage2(bl,sd,damage);
		
		//		map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x+1,sd->bl.y,BL_MOB,sd);
		//		map_foreachincell(pc_damage,sd->bl.m,sd->bl.x+1,sd->bl.y,BL_PC,damage,sd);
		/*		if(sd->tooclose==0) {	
		if(sd->polearm>0) {
		
		 map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x+2,sd->bl.y,BL_MOB,sd);
		 map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x+1,sd->bl.y-1,BL_MOB,sd);
		 map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x+1,sd->bl.y+1,BL_MOB,sd);
		 }
		 }
		*/		break;
	case 2: //DOWN
		bl=map_firstincell_Reverse(sd->bl.m,sd->bl.x,sd->bl.y+1,BL_MOB);
		if(bl==NULL) return 0;
		if( bl->type== BL_MOB ) clif_mob_damage2(bl,sd,damage);
		else if( bl->type== BL_PC ) pc_damage2(bl,sd,damage);
		
//		map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x,sd->bl.y+1,BL_MOB,sd);
//		map_foreachincell(pc_damage,sd->bl.m,sd->bl.x,sd->bl.y+1,BL_PC,damage,sd);
/*		if(sd->tooclose==0) {
			if(sd->polearm>0) {
				
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x,sd->bl.y+2,BL_MOB,sd);
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x-1,sd->bl.y+1,BL_MOB,sd);
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x+1,sd->bl.y+1,BL_MOB,sd);
			}
		}
*/		break;
		
	case 3: //LEFT
		bl=map_firstincell_Reverse(sd->bl.m,sd->bl.x-1,sd->bl.y,BL_MOB);
		if(bl==NULL) return 0;
		if( bl->type== BL_MOB ) clif_mob_damage2(bl,sd,damage);
		else if( bl->type== BL_PC ) pc_damage2(bl,sd,damage);
		
//		map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x-1,sd->bl.y,BL_MOB,sd);
//		map_foreachincell(pc_damage,sd->bl.m,sd->bl.x-1,sd->bl.y,BL_PC,damage,sd);
/*		if(sd->polearm>0) {
			if(sd->tooclose==0) {
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x-2,sd->bl.y,BL_MOB,sd);
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x-1,sd->bl.y-1,BL_MOB,sd);
				map_foreachincell(clif_mob_damage,sd->bl.m,sd->bl.x-1,sd->bl.y+1,BL_MOB,sd);
			}
		}
*/		break;
	default:
		break;
	}
	return 0;
}
int clif_battle_values(struct map_sessiondata *sd,struct mobspawn_data *mob,int dm) {
	int damage=dm;
	int dx;
	int dy;
	
	dx=abs(sd->bl.x-mob->bl.x);
	dy=abs(sd->bl.y-mob->bl.y);
	if(sd->rage>0) damage=damage*sd->rage;
	if(sd->status.state==2 || sd->status.state==5) { 
		damage=damage*8;
	}
	

	if(sd->polearm>0) {
		if((dx==1 && dy==0) || (dx==0 && dy==1)) {
		  if(sd->status.state==2 || sd->status.state==5){	  
			sd->status.state=0;
			  clif_sendstatus(sd);
		}
		sd->tooclose=1;
		} else {
		  damage=damage*(1.5*(sd->status.subclass+1));
		}
	} else {
		if(sd->status.state==2 || sd->status.state==5){
		sd->status.state=0;
		clif_sendstatus(sd);
		}
		sd->tooclose=1;
	}
	
	return damage;
}
int clif_mob_look(struct block_list *bl,va_list ap) {
	int type;
	struct map_sessiondata *sd;
	struct mobspawn_data *mob;
	int look;
	int color;
//	char newname[32];
//	int len_name;
	type=va_arg(ap,int);
	
	if(type==LOOK_GET) {
		nullpo_ret( sd=va_arg(ap,struct map_sessiondata*));
		nullpo_ret( mob=(struct mobspawn_data*)bl);
		if(mob->state==MOB_DEAD)
			return 0;
	}else{
		nullpo_ret(sd=(struct map_sessiondata*)bl);
		nullpo_ret(mob=va_arg(ap,struct mobspawn_data*));
		if(mob->state==MOB_DEAD)
			return 0;
	}
	look=mobdb_look(mob->id);
	color=mobdb_color(mob->id);
	
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOW(sd->fd,1)=SWAP16(18);
	WFIFOB(sd->fd,3)=0x07; //send mob graphic
	WFIFOB(sd->fd,4)=0x03; //incremental
	WFIFOW(sd->fd,5)=SWAP16(1);
	WFIFOW(sd->fd,7)=SWAP16(mob->bl.x);
	WFIFOW(sd->fd,9)=SWAP16(mob->bl.y);
	WFIFOB(sd->fd,11)=0x05;
	WFIFOL(sd->fd,12)=SWAP32(mob->bl.id);
	WFIFOW(sd->fd,16)=SWAP16(32768+look);
	WFIFOB(sd->fd,18)=color;
	WFIFOB(sd->fd,19)=mob->side;
	WFIFOB(sd->fd,20)=0;
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 21);
	return 0;
}
int clif_send_duration(struct map_sessiondata* sd,int id,int time) {
	int len;
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,3)=0x3A;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOB(sd->fd,5)=strlen(magicdb_name(id));
	len=strlen(magicdb_name(id));
	strcpy(WFIFOP(sd->fd,6),magicdb_name(id));
	WFIFOL(sd->fd,len+6)=SWAP32(time);
	WFIFOW(sd->fd,1)=SWAP16(len+7);
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,len+10);
	
	return 0;
}

int clif_send_aether(struct map_sessiondata* sd,int id,int time) {
	int pos;
	
	pos=clif_findspell_pos(sd,id);
	if(pos<0) return 0;

	WFIFOB(sd->fd,0)=0xAA;
	WFIFOW(sd->fd,1)=SWAP16(8);
	WFIFOB(sd->fd,3)=63; //Aether packet
	WFIFOB(sd->fd,4)=3;
	WFIFOW(sd->fd,5)=SWAP16(pos+1);
	WFIFOL(sd->fd,7)=SWAP32(time);
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,11);
	return 0;
}

int clif_mob_move(struct block_list *bl, va_list ap) {
	int type;
	struct map_sessiondata *sd;
	struct mobspawn_data *mob;
	type=va_arg(ap,int);
	
	if(type==LOOK_GET) {
		nullpo_ret(sd=va_arg(ap,struct map_sessiondata*));
		nullpo_ret(mob=(struct mobspawn_data*)bl);
		if(mob->state==MOB_DEAD)
			return 0;
	}else{
		nullpo_ret(sd=(struct map_sessiondata*)bl);
		nullpo_ret(mob=va_arg(ap,struct mobspawn_data*));
		if(mob->state==MOB_DEAD)
			return 0;
	}
	
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOW(sd->fd,1)=SWAP16(11);
	WFIFOB(sd->fd,3)=0x0C;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOL(sd->fd,5)=SWAP32(mob->bl.id);
	WFIFOW(sd->fd,9)=SWAP16(mob->bx);
	WFIFOW(sd->fd,11)=SWAP16(mob->by);
	WFIFOB(sd->fd,13)=mob->side;
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,14);
	return 0;
}
int clif_mob_damage(struct block_list *bl, va_list ap){
	struct mobspawn_data *mob;
	struct map_sessiondata *sd;
	int damage;
	int hit;


	nullpo_ret(sd=va_arg(ap,struct map_sessiondata*));
	nullpo_ret(mob=(struct mobspawn_data*)bl);

	damage=pc_calcdamage(sd,mobdb_sizetype(mob->id));
	
	if(mob->state==MOB_DEAD) return 0;

	damage=clif_battle_values(sd,mob,damage);
	hit=clif_calc_critical(sd);
	sd->target=mob->bl.id;
	run_script(itemdb_script(sd->status.equip[EQ_WEAP].id),0,sd->bl.id,0);
	if(hit>0) {
		
		if(mobdb_type(mob->id)==MOB_NORMAL || mobdb_type(mob->id)==MOB_AGGRESSIVE)
		{
			mob->attacker=sd->bl.id;
			mob->state=MOB_HIT;
		
		// just running away from attacker
		} else if (mobdb_type(mob->id)==MOB_REGRESSIVE && ( (mobdb_getvita(mob->id)/mob->current_vita) >= 10 ) ) {
			mob->attacker=sd->bl.id;
			mob->state=MOB_RUN;
		}
		if(mob->side==sd->status.side) {
			damage=damage*2;
		}
		else if(abs(mob->side-sd->status.side)!=2) {
			damage=damage*1.5;
		}
		if(hit==2) {
			damage=damage*3;
		}
		
		damage=mob_damaged(mob,damage);

		if(hit==1) { 
			clif_send_mob_health(mob,sd,damage,33);
		} else { 
			clif_send_mob_health(mob,sd,damage,255);
		}
		
		
	}
	return 0;
}

int clif_mob_damage2(struct block_list *bl, struct map_sessiondata *sd, int damage) {
	struct mobspawn_data *mob;
	int hit;

	nullpo_ret(mob=(struct mobspawn_data*)bl);

	if(mobdb_sizetype(mob->id) != 's') damage=pc_calcdamage(sd,'l');
	
	if(mob->state==MOB_DEAD) return 0;

	damage=clif_battle_values(sd,mob,damage);
	hit=clif_calc_critical(sd);
	sd->target=mob->bl.id;
	run_script(itemdb_script(sd->status.equip[EQ_WEAP].id),0,sd->bl.id,0);
	if(hit>0) {
		
		if(mobdb_type(mob->id)==MOB_NORMAL || mobdb_type(mob->id)==MOB_AGGRESSIVE)
		{
			mob->attacker=sd->bl.id;
			mob->state=MOB_HIT;
		
		// just running away from attacker
		} else if (mobdb_type(mob->id)==MOB_REGRESSIVE && ( (mobdb_getvita(mob->id)/mob->current_vita) >= 10 ) ) {
			mob->attacker=sd->bl.id;
			mob->state=MOB_RUN;
		}
		if(mob->side==sd->status.side) {
			damage=damage*2;
		}
		else if(abs(mob->side-sd->status.side)!=2) {
			damage=damage*1.5;
		}
		if(hit==2) {
			damage=damage*3;
		}
		
		damage=mob_damaged(mob,damage);
		if(sd->status.equip[EQ_WEAP].id) clif_sendsound(sd, 196957); // ´¼Ҹ 
		if(hit==1) { 
			clif_send_mob_health(mob,sd,damage,33);
		} else { 
			clif_send_mob_health(mob,sd,damage,255);
		}
		
		
	}
	return 0;
}
int clif_send_mob_health(struct mobspawn_data* mob, struct map_sessiondata *sd,int damage,int critical){
	
	int maxvita;
	int currentvita;
	float percentage;
	
	maxvita=mobdb_getvita(mob->id);
	currentvita=mob->current_vita;
	
	if(currentvita<=0){
		percentage=0;
	}else{
		percentage=(float)currentvita/(float)maxvita;
		percentage=(float)percentage*100;
	}
	
	
	
	
	
	
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOW(sd->fd,1)=SWAP16(12);
	WFIFOB(sd->fd,3)=0x13;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOL(sd->fd,5)=SWAP32(mob->bl.id);
	WFIFOB(sd->fd,9)=critical;
	WFIFOB(sd->fd,10)=percentage;
	WFIFOL(sd->fd,11)=SWAP32(damage);
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,15);
	
	if(currentvita<=0){
		mobdb_drops(mob->id,mob->bl.m,mob->bl.x,mob->bl.y);
		clif_mob_kill(sd,mob);
		pc_givexp(sd,mobdb_exp(mob->id),xp_rate);
	}
	
	return 0;
}


int clif_mob_kill(struct map_sessiondata *sd, struct mobspawn_data* mob){
	

	// run script when the mob died
	run_script(mobdb[mob->id]->attack_script,0,sd->bl.id,0);
	mob->state=MOB_DEAD;
	mob->attacker=0;
	mob->time=0;
	map_delblock(&mob->bl);
	map_foreachinarea(clif_send_destroy,mob->bl.m,mob->bl.x,mob->bl.y,AREA,BL_PC,LOOK_GET,mob);
	//map_addblock(&mob->bl);
	
	return 0;
}

int clif_send_destroy(struct block_list *bl,va_list ap){
	struct map_sessiondata *sd;
	struct mobspawn_data *mob;
	int type;
	type=va_arg(ap,int);
	
	nullpo_ret(sd=(struct map_sessiondata*)bl);
	nullpo_ret(mob=va_arg(ap,struct mobspawn_data*));
	
	
	
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOW(sd->fd,1)=SWAP16(6);
	WFIFOB(sd->fd,3)=0x5F;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOL(sd->fd,5)=SWAP32(mob->bl.id);
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,9);
	return 0;
}

int clif_parsemenu2(struct map_sessiondata *sd) {
    int npc_choice=RFIFOB(sd->fd,13);
    int npc_id=0;
    npc_id=SWAP32(RFIFOL(sd->fd,6));

	switch (npc_choice) {
	case 1:
		sd->npc_pos=-1;
		sd->st->pos=-1;
		sd->st->state = END; 
		sd->npc_menu=0;	
		sd->state.menu_or_input=0;	
		sd->npc_prev=0;
		sd->npc_gc=0;
		sd->npc_id=0;
		sd->npc_g=0;
		npc_scriptcont(sd,(int)npc_id,(int)sd->npc_pos);
	//	printf("\n");
		break;

	default:
		sd->npc_menu = RFIFOB(sd->fd,15);
		npc_scriptcont(sd,npc_id,sd->npc_pos);
		break;
	}
}

int clif_parsenpcdialog(struct map_sessiondata *sd) {
    int npc_choice=RFIFOB(sd->fd,13);
    int npc_id=0;
    npc_id=SWAP32(RFIFOL(sd->fd,6));
    
    //printf("npc_choice: %d\n",npc_choice);
	
    switch (npc_choice) {
	case 0: //previous
		
		sd->npc_pos=sd->npc_prev;
		
		npc_scriptcont(sd,npc_id,sd->npc_pos);
		break;
	case 1: //quit
		sd->npc_pos=-1;
		sd->st->pos=-1;

		
		npc_scriptcont(sd,(int)npc_id,(int)sd->npc_pos);
		sd->npc_prev=0;
		sd->npc_gc=0;
		sd->npc_id=0;
		sd->npc_g=0;
		break;
	case 2: //next
		
		npc_scriptcont(sd,npc_id,sd->npc_pos);
        break;
	}
    return 0;
}
int clif_send_sub(struct block_list *bl, va_list ap) {
	unsigned char *buf;
	int len;
	struct block_list *src_bl;
	int type;
	struct map_sessiondata *sd;

	//nullpo_ret( bl);
	nullpo_ret( ap);
	nullpo_ret( sd=(struct map_sessiondata *)bl);

	buf=va_arg(ap,unsigned char*);
	len=va_arg(ap,int);
	nullpo_ret( src_bl=va_arg(ap,struct block_list*));
	type=va_arg(ap,int);

	switch(type) {
	case AREA_WOS:
		if(bl==src_bl)
			return 0;
		break;
	case SAMEAREA_WOS:
		if(bl==src_bl)
			return 0;
		break;
	}

	if ((sd) && (WFIFOP(sd->fd,0) != buf))
		memcpy(WFIFOP(sd->fd,0),buf,len);
	if (sd) WFIFOSET(sd->fd,len);
	return 0;
}

int clif_send(unsigned char *buf,int len,struct block_list *bl,int type) {
	struct map_sessiondata *sd;
	int i;
	switch(type) {
	case ALL_CLIENT:
	case SAMESRV:
		for(i=0;i<fd_max;i++){
			if(session[i] && (sd=session[i]->session_data)){
				memcpy(WFIFOP(i,0),buf,len);
				WFIFOSET(i,len);
			}
		}
		break;
	case SAMEMAP:
		for(i=0;i<fd_max;i++){
			if(session[i] && (sd=session[i]->session_data) && sd->bl.m == bl->m){
				memcpy(WFIFOP(i,0),buf,len);
				WFIFOSET(i,len);
			}
		}
		break;
	case SAMEMAP_WOS:
		for(i=0;i<fd_max;i++){
			if(session[i] && (sd=session[i]->session_data) && sd->bl.m == bl->m && sd != (struct map_sessiondata *)bl){
				memcpy(WFIFOP(i,0),buf,len);
				WFIFOSET(i,len);
			}
		}
		break;
	case AREA:
	case AREA_WOS:
		map_foreachinarea(clif_send_sub,bl->m,bl->x,bl->y,AREA,BL_PC,buf,len,bl,type);
		break;
	case SAMEAREA:
	case SAMEAREA_WOS:
		map_foreachinarea(clif_send_sub,bl->m,bl->x,bl->y,SAMEAREA,BL_PC,buf,len,bl,type);
		break;
	case CORNER:
		map_foreachinarea(clif_send_sub,bl->m,bl->x,bl->y,CORNER,BL_PC,buf,len,bl,type);
		break;
	case SELF:
		sd=(struct map_sessiondata *)bl;
		memcpy(WFIFOP(sd->fd,0),buf,len);
		WFIFOSET(sd->fd,len);
		break;
	}
	return 0;
}

int clif_quit(struct map_sessiondata *sd) {
	map_delblock(&sd->bl);
	clif_lookgone(&sd->bl);
	return 0;
}
int clif_mystaytus(struct map_sessiondata *sd) {
	int len;
	int x;
	int i;
	int count;
	int tmp;
	int exp_need;
	int packetlen;
	int templen=0;
	int legendcount=0;

	if(sd->status.level<99) {
		if(sd->status.level==5 && sd->status.class ==0 ){
			exp_need=0;
			
		}
		else {
			exp_need=exp_require[sd->status.class][sd->status.level+1]-sd->status.exp;
	
		} 
	} else {
		exp_need=0;
	}
	len=0;

	
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,3)=57;
	WFIFOB(sd->fd,4)=3;
	WFIFOB(sd->fd,5)=(sd->ac+256)%256; //AC Value
	//WFIFOB(sd->fd,5)=0;
	WFIFOB(sd->fd,6)=sd->hit;
	WFIFOB(sd->fd,7)=sd->dam;
	WFIFOB(sd->fd,8)=14; // clan name size
	strcpy(WFIFOP(sd->fd,9),"ӽùĻ"); //clan name
	len+=14;
	WFIFOB(sd->fd,len+9)= tmp = strlen(sd->status.clan_title);
	strcpy(WFIFOP(sd->fd,len+10),sd->status.clan_title);
	len+=tmp;
	WFIFOB(sd->fd,len+10)= tmp = strlen(sd->status.title);
	strcpy(WFIFOP(sd->fd,len+11),sd->status.title);
	len+=tmp;
	WFIFOB(sd->fd,len+11)=0;//ƹ͵ƴ 
	WFIFOB(sd->fd,len+12)=sd->status.opt.al_group;//1;
	WFIFOL(sd->fd,len+13)=SWAP32(exp_need); // remaining exp to level up
	
	WFIFOB(sd->fd,len+17)= tmp = class[sd->status.class][sd->status.subclass].class_name_len; //Class name length
	memcpy(WFIFOP(sd->fd,len+18),class[sd->status.class][sd->status.subclass].class_name,class[sd->status.class][sd->status.subclass].class_name_len);
	len+=tmp;
	
	for(x=0;x<9;x++) {
		WFIFOW(sd->fd,len+18)=SWAP16(itemdb_icon(sd->status.equip[x].id));
		WFIFOB(sd->fd,len+20)=itemdb_iconcolor(sd->status.equip[x].id);
		WFIFOB(sd->fd,len+21)=0x00;
		WFIFOB(sd->fd,len+22)=0x00;
		WFIFOB(sd->fd,len+23)=0x00;
		WFIFOB(sd->fd,len+24)=0x00;
		WFIFOB(sd->fd,len+25)=0x00;
		WFIFOB(sd->fd,len+26)=0x00;
		len+=9;
	}
	len-=9;

	WFIFOB(sd->fd,len+27)=sd->status.opt.al_exchg; //ȯ
	WFIFOB(sd->fd,len+28)=sd->status.opt.al_pvp; //PVP

	
	for(i=0; i<MAX_LEGEND; i++) {
		if(sd->legend_list.legends[i].len > 0) {
			WFIFOB(sd->fd,len+30+templen) = sd->legend_list.legends[i].icon; //ܹȣ
			WFIFOB(sd->fd,len+31+templen) = sd->legend_list.legends[i].color; //۾Į
			WFIFOB(sd->fd,len+32+templen) = sd->legend_list.legends[i].len; //۾Ʈ 
			memcpy(WFIFOP(sd->fd,len+33+templen),sd->legend_list.legends[i].text, sd->legend_list.legends[i].len);
			
			templen += 3+sd->legend_list.legends[i].len;
			legendcount++;
		}
	}
	WFIFOW(sd->fd,len+29)=legendcount; // īƮ 
	WFIFOB(sd->fd,len+34+templen) = '0';

	WFIFOW(sd->fd,1)=SWAP16(len+34+templen);
	packetlen=crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,packetlen);
	return 0;
}
int clif_send_worldmap(struct map_sessiondata *sd) {
	
	int count=1;
	int c_size=4;
	int len;
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,1)=SWAP16(26+c_size);
	WFIFOB(sd->fd,3)=0x2E;
	WFIFOB(sd->fd,4)=0;
	WFIFOB(sd->fd,5)=2;
	WFIFOB(sd->fd,6)='W';
	WFIFOB(sd->fd,7)='M';
	WFIFOB(sd->fd,8)=count;
	WFIFOB(sd->fd,9)=0; //target area
	WFIFOB(sd->fd,10)=6;
	WFIFOB(sd->fd,11)=(char)130;
	WFIFOB(sd->fd,12)=3;
	WFIFOB(sd->fd,13)=(char)60;
	WFIFOB(sd->fd,14) =c_size;
	memcpy(WFIFOP(sd->fd,15),"̸",c_size);
	WFIFOB(sd->fd,15+c_size) = 0;
	WFIFOB(sd->fd,16+c_size) =0;
	WFIFOB(sd->fd,17+c_size) =0;
	WFIFOB(sd->fd,18+c_size) =0;
	WFIFOB(sd->fd,19+c_size) =0;
	WFIFOB(sd->fd,20+c_size) =0;
	WFIFOB(sd->fd,21+c_size) =0;
	WFIFOB(sd->fd,22+c_size) =0;
	WFIFOB(sd->fd,23+c_size) =0;
	WFIFOB(sd->fd,24+c_size) =2;
	WFIFOB(sd->fd,25+c_size) =0;
	WFIFOB(sd->fd,26+c_size) =0;
	WFIFOB(sd->fd,27+c_size) =0;
	WFIFOB(sd->fd,28+c_size) =2;

	len = crypt(WFIFOP(sd->fd,0));	
	WFIFOSET(sd->fd,len);
}

int clif_lookgone(struct block_list *bl) {	
	char buf[9];
	WBUFB(buf, 0) = 0xAA;
	WBUFB(buf, 1) = 0x00;
	WBUFB(buf, 2) = 0x06;
	if(bl->type==BL_PC)
		WBUFB(buf, 3) = 0x0E;
	else 
		WBUFB(buf, 3) = 0x5F;
	WBUFB(buf, 4) = 0x03;
	WBUFL(buf, 5) = SWAP32(bl->id);
	crypt(WBUFP(buf, 0));
	clif_send(buf, 9, bl, AREA);
	return 0;
}

int clif_itemlook_sub(struct block_list *bl, va_list ap) {
	struct map_sessiondata *sd;
	struct flooritem_data *fl;
	int type;
	type=va_arg(ap,int);
	
	if (type==LOOK_GET) {
		nullpo_ret( sd=va_arg(ap, struct map_sessiondata*));
		nullpo_ret( fl=(struct flooritem_data*)bl);
	} else {
		nullpo_ret( sd=(struct map_sessiondata*)bl);
		nullpo_ret( fl=va_arg(ap, struct flooritem_data*));
	}

	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x11;
	WFIFOB(sd->fd, 3) = 0x07;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOB(sd->fd, 5) = 0x00;
	WFIFOB(sd->fd, 6) = 0x01;
	WFIFOW(sd->fd, 7) = SWAP16(fl->bl.x);
	WFIFOW(sd->fd, 9) = SWAP16(fl->bl.y);
	WFIFOB(sd->fd, 11) = 0x02; //for item thing
	WFIFOL(sd->fd, 12) = SWAP32(fl->bl.id);
	WFIFOW(sd->fd, 16) = SWAP16(itemdb_icon(fl->data.id));
	WFIFOB(sd->fd, 18) = itemdb_iconcolor(fl->data.id);
	WFIFOB(sd->fd, 19) = 0x00;
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 20);
	
	return 0;
}

int clif_npc_look(struct block_list *bl, va_list ap) {
    
	int type;
	struct map_sessiondata *sd=NULL;
    struct npc_data *npc=NULL;
	
    type=va_arg(ap, int);
    //printf("npc_look done!\n");
    
	if (type==LOOK_GET) {
		nullpo_ret( sd=va_arg(ap,struct map_sessiondata*));
		nullpo_ret( npc=(struct npc_data*)bl);
		//printf("npc id: %d\n",npc->bl.graphic_id);
        if (npc->bl.subtype==FLOOR)
           return 0;
	}
    //printf("npc_graphic: %d\n",npc->bl.graphic_id);
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOW(sd->fd, 1) = SWAP16(19); //packet size
	WFIFOB(sd->fd, 3) = 0x07;
	WFIFOB(sd->fd, 4) = 0x03;
    WFIFOW(sd->fd, 5) = SWAP16(1); //only 1 npc/monster
    WFIFOW(sd->fd, 7) = SWAP16(npc->bl.x); //X coord
    WFIFOW(sd->fd, 9) = SWAP16(npc->bl.y); //Y Coord
    WFIFOB(sd->fd, 11) = 12; //Clickable NPC
    WFIFOL(sd->fd, 12) = SWAP32(npc->bl.id); //Id of npc
    WFIFOW(sd->fd, 16) = SWAP16(32768+npc->bl.graphic_id);
    WFIFOB(sd->fd, 18) = npc->bl.graphic_color;
    WFIFOB(sd->fd, 19) = 2; //This is side
    WFIFOW(sd->fd, 20) = 0;
    crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 22);
	
	return 0;
}

int clif_charlook_sub(struct block_list *bl, va_list ap) {
	int len;
	int type;
	int tmp=0;
	struct map_sessiondata *sd, *src_sd;
	type=va_arg(ap, int);

	if (type==LOOK_GET) {
		nullpo_ret( sd=(struct map_sessiondata*)bl);
		nullpo_ret( src_sd=va_arg(ap,struct map_sessiondata*));
		if (src_sd==sd)
			return 0;
	} else {
		nullpo_ret( src_sd=(struct map_sessiondata*)bl);
		nullpo_ret( sd=va_arg(ap,struct map_sessiondata*));
	}
	
	WFIFOB(src_sd->fd, 0) = 0xAA;
	WFIFOB(src_sd->fd, 3) = 0x33;
	WFIFOB(src_sd->fd, 4) = 0x03;
	WFIFOW(src_sd->fd, 5) = SWAP16(sd->bl.x);
	WFIFOW(src_sd->fd, 7) = SWAP16(sd->bl.y);
	WFIFOB(src_sd->fd, 9) = sd->status.side;
	WFIFOL(src_sd->fd, 10) = SWAP32(sd->status.id);
	WFIFOB(src_sd->fd, 14) = 0x00;
	WFIFOB(src_sd->fd, 15) = SWAP16(sd->status.sex);
	WFIFOB(src_sd->fd, 16) = sd->status.state;
	WFIFOB(src_sd->fd, 17) = 0;
	WFIFOB(src_sd->fd, 18) = 0;
	WFIFOB(src_sd->fd, 19) = 75; // Speed for movement
	WFIFOB(src_sd->fd, 20) = 0;
	WFIFOB(src_sd->fd, 21) = sd->status.face; //face
	WFIFOB(src_sd->fd, 22) = sd->status.hair; //hair
	WFIFOB(src_sd->fd, 23) = sd->status.hair_color; //hair color
	//if(sd->coat<255) {
	//	WFIFOB(src_sd->fd,24)=39;
	//} else {

	//}
	//armor
	if (!(tmp=pc_isequip(sd, EQ_ARMOR))) {		
			WFIFOB(src_sd->fd, 24) = 0x00;
			WFIFOB(src_sd->fd, 25) = 0x00;
	} else {
	//if(sd->coat<255) {
		//WFIFOB(src_sd->fd, 25) = itemdb_look(pc_isequip(sd, EQ_ARMOR));
//		WFIFOB(src_sd->fd,25)=sd->coat+16;
//		WFIFOB(src_sd->fd, 26) = sd->coat_color;
//	} else{
		WFIFOB(src_sd->fd, 24) = itemdb_look(tmp);
		WFIFOB(src_sd->fd, 25)=	sd->status.armor_color;
	}
	//weapon
	if (!(tmp=pc_isequip(sd, EQ_WEAP))) {
		WFIFOB(src_sd->fd, 26) = 0xFF;
		WFIFOB(src_sd->fd, 27) = 0xFF;
	} else {
		WFIFOW(src_sd->fd, 26) = SWAP16(itemdb_look(tmp));
		WFIFOB(src_sd->fd, 28) = itemdb_lookcolor(tmp);
	
	}	
//shield
	if (!(tmp=pc_isequip(sd, EQ_SHIELD)) || src_sd->status.opt.showhelm == 0)
		WFIFOW(src_sd->fd, 29) = 0xFFFF;
	else
		WFIFOW(src_sd->fd, 29) = SWAP16(itemdb_look(tmp));
		WFIFOB(src_sd->fd, 31) = itemdb_lookcolor(tmp);
	if (!(tmp=pc_isequip(sd, EQ_HELM))) {
	//helm stuff goes here
		WFIFOB(src_sd->fd, 32)=0; // supposed to be 1=Helm, 0=No helm
		WFIFOB(src_sd->fd, 33)=0xFF; // supposed to be Helm num
		WFIFOB(src_sd->fd, 34)=0XFF; // helm color
       } else {
		WFIFOB(src_sd->fd, 32)=1;
		WFIFOB(src_sd->fd, 33)=itemdb_look(tmp);
		WFIFOB(src_sd->fd, 34)=itemdb_lookcolor(tmp);
	}
	
	if (!(tmp=pc_isequip(sd,EQ_FACEACC))) {
	//beard stuff
		WFIFOW(src_sd->fd, 35)=0xFFFF;
		WFIFOB(src_sd->fd, 37)=0xFF;
	} else {
		WFIFOW(src_sd->fd, 35)=SWAP16(itemdb_look(tmp)); //beard num
		WFIFOB(src_sd->fd, 37)=itemdb_lookcolor(tmp); //beard color
	}

	WFIFOB(src_sd->fd, 38) = 4;
	strcpy(WFIFOP(src_sd->fd, 40), sd->status.name);
	len = strlen(sd->status.name);
	WFIFOB(src_sd->fd, 39) = len;
	WFIFOW(src_sd->fd, 1) = SWAP16(len+40);

	WFIFOSET(src_sd->fd, crypt(WFIFOP(src_sd->fd, 0)));

	if(sd->disguise>0) {
	WFIFOB(src_sd->fd,0)=0xAA;
	WFIFOB(src_sd->fd,3)=29;
	WFIFOB(src_sd->fd,4)=0x03;
	WFIFOL(src_sd->fd,5)=SWAP32(sd->bl.id);
	WFIFOB(src_sd->fd,9)=1;
	WFIFOB(src_sd->fd,10)=5;
	WFIFOB(src_sd->fd,11)=4;
	WFIFOW(src_sd->fd,12)=SWAP16(sd->disguise+32768);
	WFIFOB(src_sd->fd,14)=sd->disguise_color;
	WFIFOB(src_sd->fd,16)=strlen(sd->status.name);
	len=strlen(sd->status.name);
	strcpy(WFIFOP(src_sd->fd,17),sd->status.name);
	WFIFOW(src_sd->fd,1)=SWAP16(len+14);
	crypt(WFIFOP(src_sd->fd,0));
	WFIFOSET(src_sd->fd,len+17);
	}
	return 0;
}

// except location data
int clif_charlook_sub2(struct block_list *bl, va_list ap) {
	int len;
	int type;
	int tmp=0;
	struct map_sessiondata *sd, *src_sd;
	type=va_arg(ap, int);
	
	if (type==LOOK_GET) {
		nullpo_ret( sd=(struct map_sessiondata*)bl);
		nullpo_ret( src_sd=va_arg(ap,struct map_sessiondata*));
		if (src_sd==sd)
			return 0;
	} else {
		nullpo_ret( src_sd=(struct map_sessiondata*)bl);
		nullpo_ret( sd=va_arg(ap,struct map_sessiondata*));
	}
	
	WFIFOB(src_sd->fd, 0) = 0xAA;
	WFIFOB(src_sd->fd, 3) = 0x1D;
	WFIFOB(src_sd->fd, 4) = 0x03;
	WFIFOL(src_sd->fd, 5) = SWAP32(sd->status.id);
	WFIFOB(src_sd->fd, 9) = 0x00;
	WFIFOB(src_sd->fd, 10) = SWAP16(sd->status.sex);
	WFIFOB(src_sd->fd, 11) = sd->status.state;
	WFIFOB(src_sd->fd, 12) = 0;
	WFIFOB(src_sd->fd, 13) = 0;
	WFIFOB(src_sd->fd, 14) = 75; // Speed for movement
	WFIFOB(src_sd->fd, 15) = 0;
	WFIFOB(src_sd->fd, 16) = sd->status.face; //face
	WFIFOB(src_sd->fd, 17) = sd->status.hair; //hair
	WFIFOB(src_sd->fd, 18) = sd->status.hair_color; //hair color
	//if(sd->coat<255) {
	//	WFIFOB(src_sd->fd,24)=39;
	//} else {
	
	//}
	//armor
	if (!(tmp=pc_isequip(sd, EQ_ARMOR))) {		
		WFIFOB(src_sd->fd, 19) = 0x00;
		WFIFOB(src_sd->fd, 20) = 0x00;
	} else {
		//if(sd->coat<255) {
		//WFIFOB(src_sd->fd, 25) = itemdb_look(pc_isequip(sd, EQ_ARMOR));
		//		WFIFOB(src_sd->fd,25)=sd->coat+16;
		//		WFIFOB(src_sd->fd, 26) = sd->coat_color;
		//	} else{
		WFIFOB(src_sd->fd, 19) = itemdb_look(tmp);
		WFIFOB(src_sd->fd, 20)=	sd->status.armor_color;
	}
	//weapon
	if (!(tmp=pc_isequip(sd, EQ_WEAP))) {
		WFIFOB(src_sd->fd, 21) = 0xFF;
		WFIFOB(src_sd->fd, 23) = 0xFF;
	} else {
		WFIFOW(src_sd->fd, 21) = SWAP16(itemdb_look(tmp));
		WFIFOB(src_sd->fd, 23) = itemdb_lookcolor(tmp);
		
	}	
	//shield
	if (!(tmp=pc_isequip(sd, EQ_SHIELD)) || src_sd->status.opt.showhelm == 0)
		WFIFOW(src_sd->fd, 24) = 0xFFFF;
	else
		WFIFOW(src_sd->fd, 24) = SWAP16(itemdb_look(tmp));
	WFIFOB(src_sd->fd, 26) = itemdb_lookcolor(tmp);
	if (!(tmp=pc_isequip(sd, EQ_HELM))) {
		//helm stuff goes here
		WFIFOB(src_sd->fd, 27)=0; // supposed to be 1=Helm, 0=No helm
		WFIFOB(src_sd->fd, 28)=0xFF; // supposed to be Helm num
		WFIFOB(src_sd->fd, 29)=0XFF; // helm color
	} else {
		WFIFOB(src_sd->fd, 27)=1;
		WFIFOB(src_sd->fd, 28)=itemdb_look(tmp);
		WFIFOB(src_sd->fd, 29)=itemdb_lookcolor(tmp);
	}
	
	if (!(tmp=pc_isequip(sd,EQ_FACEACC))) {
		//beard stuff
		WFIFOW(src_sd->fd, 30)=0xFFFF;
		WFIFOB(src_sd->fd, 32)=0xFF;
	} else {
		WFIFOW(src_sd->fd, 30)=SWAP16(itemdb_look(tmp)); //beard num
		WFIFOB(src_sd->fd, 32)=itemdb_lookcolor(tmp); //beard color
	}
	
	WFIFOB(src_sd->fd, 33) = 4;
	strcpy(WFIFOP(src_sd->fd, 35), sd->status.name);
	len = strlen(sd->status.name);
	WFIFOB(src_sd->fd, 34) = len;
	WFIFOW(src_sd->fd, 1) = SWAP16(len+35);
	
	WFIFOSET(src_sd->fd, crypt(WFIFOP(src_sd->fd, 0)));
	
	if(sd->disguise>0) {
		WFIFOB(src_sd->fd,0)=0xAA;
		WFIFOB(src_sd->fd,3)=29;
		WFIFOB(src_sd->fd,4)=0x03;
		WFIFOL(src_sd->fd,5)=SWAP32(sd->bl.id);
		WFIFOB(src_sd->fd,9)=1;
		WFIFOB(src_sd->fd,10)=5;
		WFIFOB(src_sd->fd,11)=4;
		WFIFOW(src_sd->fd,12)=SWAP16(sd->disguise+32768);
		WFIFOB(src_sd->fd,14)=sd->disguise_color;
		WFIFOB(src_sd->fd,16)=strlen(sd->status.name);
		len=strlen(sd->status.name);
		strcpy(WFIFOP(src_sd->fd,17),sd->status.name);
		WFIFOW(src_sd->fd,1)=SWAP16(len+14);
		crypt(WFIFOP(src_sd->fd,0));
		WFIFOSET(src_sd->fd,len+17);
	}
	return 0;
}


int clif_getchararea(struct map_sessiondata *sd) {
	map_foreachinarea(clif_charlook_sub,sd->bl.m,sd->bl.x,sd->bl.y,SAMEAREA,BL_PC,LOOK_GET,sd);
	return 0;
}

int clif_getitemarea(struct map_sessiondata *sd) {
	map_foreachinarea(clif_itemlook_sub,sd->bl.m,sd->bl.x,sd->bl.y,SAMEAREA,BL_ITEM,LOOK_GET,sd);
	return 0;
}

int clif_sendchararea(struct map_sessiondata *sd) {
	map_foreachinarea(clif_charlook_sub,sd->bl.m,sd->bl.x,sd->bl.y,AREA,BL_PC,LOOK_SEND,sd);
	return 0;
}

int clif_sendchararea2(struct map_sessiondata *sd) {
	map_foreachinarea(clif_charlook_sub2,sd->bl.m,sd->bl.x,sd->bl.y,AREA,BL_PC,LOOK_SEND,sd);
	return 0;
}

/*
int clif_charspecific(int sender, int id) {
	int len;
//	int type;
	struct map_sessiondata *sd,*src_sd;
	int tmp=0;
	sd=map_id2sd(sender);
	src_sd=map_id2sd(id);
	
	WFIFOB(src_sd->fd, 0) = 0xAA;
	WFIFOB(src_sd->fd, 3) = 0x33;
	WFIFOB(src_sd->fd, 4) = 0x03;
	WFIFOW(src_sd->fd, 5) = SWAP16(sd->bl.x);
	WFIFOW(src_sd->fd, 7) = SWAP16(sd->bl.y);
	WFIFOB(src_sd->fd, 9) = sd->status.side;
	WFIFOL(src_sd->fd, 10) = SWAP32(sd->status.id);
	WFIFOB(src_sd->fd, 14) = 0x00;
	WFIFOB(src_sd->fd, 15) = SWAP16(sd->status.sex);
	WFIFOB(src_sd->fd, 16) = sd->status.state;
	WFIFOB(src_sd->fd, 17) = 0;
	WFIFOB(src_sd->fd, 18) = 0;
	WFIFOB(src_sd->fd, 19) = 60; // Speed for movement
	WFIFOB(src_sd->fd, 20) = 0;
	WFIFOB(src_sd->fd, 21) = sd->status.face; //face
	WFIFOB(src_sd->fd, 22) = sd->status.hair; //hair
	WFIFOB(src_sd->fd, 23) = sd->status.hair_color; //hair color
	//if(sd->coat<255) {
	//	WFIFOB(src_sd->fd,24)=39;
	//} else {

	//}
	//armor
	if (!(tmp=pc_isequip(sd, EQ_ARMOR))) {		
			WFIFOB(src_sd->fd, 24) = 0x00;
			WFIFOB(src_sd->fd, 25) = 0x00;
	} else {
	//if(sd->coat<255) {
		//WFIFOB(src_sd->fd, 25) = itemdb_look(pc_isequip(sd, EQ_ARMOR));
//		WFIFOB(src_sd->fd,25)=sd->coat+16;
//		WFIFOB(src_sd->fd, 26) = sd->coat_color;
//	} else{
		WFIFOB(src_sd->fd, 24) = itemdb_look(tmp);
		WFIFOB(src_sd->fd,25)=	sd->status.armor_color;
	}
	//weapon
	if (!(tmp=pc_isequip(sd, EQ_WEAP))) {
		WFIFOB(src_sd->fd, 26) = 0xFF;
		WFIFOB(src_sd->fd, 27) = 0xFF;
	} else {
		WFIFOW(src_sd->fd, 26) = SWAP16(itemdb_look(tmp));
		WFIFOB(src_sd->fd, 28) = itemdb_lookcolor(tmp);
	
	}	
//shield
	if (!(tmp=pc_isequip(sd, EQ_SHIELD)))
		WFIFOW(src_sd->fd, 29) = 0xFFFF;
	else
		WFIFOW(src_sd->fd, 29) = SWAP16(itemdb_look(tmp));
		WFIFOB(src_sd->fd, 31) = itemdb_lookcolor(tmp);
	if (!(tmp=pc_isequip(sd,EQ_HELM))) {
	//helm stuff goes here
		WFIFOB(src_sd->fd, 32)=0; // supposed to be 1=Helm, 0=No helm
		WFIFOB(src_sd->fd, 33)=0xFF; // supposed to be Helm num
		WFIFOB(src_sd->fd, 34)=0XFF; // helm color
        } else {
		WFIFOB(src_sd->fd, 32)=1;
		WFIFOB(src_sd->fd, 33)=itemdb_look(tmp);
		WFIFOB(src_sd->fd, 34)=itemdb_lookcolor(tmp);
	}
	
	if (!(tmp=pc_isequip(sd,EQ_FACEACC))) {
	//beard stuff
		WFIFOW(src_sd->fd, 35)=0xFFFF;
		WFIFOB(src_sd->fd, 37)=0xFF;
	} else {
		WFIFOW(src_sd->fd, 35)=SWAP16(itemdb_look(tmp)); //beard num
		WFIFOB(src_sd->fd, 37)=itemdb_lookcolor(tmp); //beard color
	}

	WFIFOB(src_sd->fd, 38) = 3;
	strcpy(WFIFOP(src_sd->fd, 40), sd->status.name);
	len = strlen(sd->status.name);

	WFIFOB(src_sd->fd, 39) = len;
	WFIFOW(src_sd->fd, 1) = SWAP16(len+53);

	WFIFOSET(src_sd->fd, crypt(WFIFOP(src_sd->fd, 0)));
	if(sd->disguise>0) {
	WFIFOB(src_sd->fd,0)=0xAA;
	WFIFOB(src_sd->fd,3)=29;
	WFIFOB(src_sd->fd,4)=0x03;
	WFIFOL(src_sd->fd,5)=SWAP32(sd->bl.id);
	WFIFOB(src_sd->fd,9)=1;
	WFIFOB(src_sd->fd,10)=5;
	WFIFOB(src_sd->fd,11)=4;
	WFIFOW(src_sd->fd,12)=SWAP16(sd->disguise+32768);
	WFIFOW(src_sd->fd,14)=0;
	WFIFOB(src_sd->fd,16)=strlen(sd->status.name);
	len=strlen(sd->status.name);
	strcpy(WFIFOP(src_sd->fd,17),sd->status.name);
	WFIFOW(src_sd->fd,1)=SWAP16(len+14);
	crypt(WFIFOP(src_sd->fd,0));
	WFIFOSET(src_sd->fd,len+17);
	}
	return 0;

}*/
int clif_sendack(struct map_sessiondata *sd) {
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x05;
	WFIFOB(sd->fd, 3) = 0x1E;
	WFIFOB(sd->fd, 4) = 0x00;
	WFIFOB(sd->fd, 5) = 0x48;
	WFIFOB(sd->fd, 6) = 0x65;
	WFIFOB(sd->fd, 7) = 0x78;
	WFIFOSET(sd->fd, 8);
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,1)=0x00;
	WFIFOB(sd->fd,2)=10;
	WFIFOB(sd->fd,3)=0x08;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOB(sd->fd,5)=0x08;
	WFIFOL(sd->fd,6)=0;
	WFIFOW(sd->fd,10)=0;
	WFIFOB(sd->fd,12)=0;
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,13);
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,1)=0x00;
	WFIFOB(sd->fd,2)=0x04;
	WFIFOB(sd->fd,3)=0x49;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOW(sd->fd,5)=0;
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,7);

	return 0;
}

int clif_sendtime(struct map_sessiondata *sd) {
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x04;
	WFIFOB(sd->fd, 3) = 0x20;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOB(sd->fd, 5) = 0x0C; //current time
	WFIFOB(sd->fd, 6) = 0x01; //current year
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 7);
	return 0;
}

int clif_sendid(struct map_sessiondata *sd) {
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x0E;
	WFIFOB(sd->fd, 3) = 0x05;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOL(sd->fd, 5) = SWAP32(sd->status.id);
	WFIFOL(sd->fd, 9) = SWAP32(2);
	WFIFOL(sd->fd, 13) = SWAP32(0);
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 17);
	return 0;
}

int clif_sendmapinfo(struct map_sessiondata *sd) {
	char len;
	//Map Title and Map X-Y
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 3) = 0x15;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOW(sd->fd, 5) = SWAP16(sd->bl.m);
	WFIFOW(sd->fd, 7) = SWAP16(map[sd->bl.m].xs);
	WFIFOW(sd->fd, 9) = SWAP16(map[sd->bl.m].ys);
	WFIFOB(sd->fd, 11) = 0x00;
	WFIFOB(sd->fd, 12) = 0x00;
	len = strlen(map[sd->bl.m].title);
	WFIFOB(sd->fd, 13) = len;
	memcpy(WFIFOP(sd->fd, 14), map[sd->bl.m].title, len);
	WFIFOW(sd->fd, 1) = SWAP16(11+len);
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 14+len);
	//Map BGM
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x14;
	WFIFOB(sd->fd, 3) = 0x19;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOB(sd->fd, 5) = 0x01;
	WFIFOB(sd->fd, 6) = 0x00;
	WFIFOB(sd->fd, 7) = 0x00;
	WFIFOB(sd->fd, 8) = map[sd->bl.m].bgm%256;
	WFIFOB(sd->fd, 9) = 0x00;
	WFIFOB(sd->fd, 10) = 0x00;
	WFIFOB(sd->fd, 11) = 0xC0;
	WFIFOL(sd->fd, 12) = SWAP32(sd->status.id);
	WFIFOB(sd->fd, 16) = 0x01;
	WFIFOB(sd->fd, 17) = 0x00;
	WFIFOB(sd->fd, 18) = 0x02;
	WFIFOB(sd->fd, 19) = 0x02;
	WFIFOB(sd->fd, 20) = 0x00;
	WFIFOB(sd->fd, 21) = 0x04;
	WFIFOB(sd->fd, 22) = 0x00;
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 23);
	

	return 0;
}

int clif_sendxy(struct map_sessiondata *sd) {
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x0B;
	WFIFOB(sd->fd, 3) = 0x04;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOW(sd->fd, 5) = SWAP16(sd->bl.x);
	WFIFOW(sd->fd, 7) = SWAP16(sd->bl.y);
	
	if (sd->bl.x < 8)
		WFIFOW(sd->fd, 9) = SWAP16(sd->bl.x);
	else if (sd->bl.x >= map[sd->bl.m].xs - 8)
		WFIFOW(sd->fd, 9) = SWAP16(sd->bl.x - map[sd->bl.m].xs + 17);
	else
		WFIFOW(sd->fd, 9) = SWAP16(8);

	if (sd->bl.y < 7)
		WFIFOW(sd->fd, 11) = SWAP16(sd->bl.y);
	else if (sd->bl.y >= map[sd->bl.m].ys - 7)
		WFIFOW(sd->fd, 11) = SWAP16(sd->bl.y - map[sd->bl.m].ys + 15);
	else
		WFIFOW(sd->fd, 11) = SWAP16(7);

	WFIFOB(sd->fd, 13) = 0x00;
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 14);
	
	return 0;
}

int clif_sendstatus(struct map_sessiondata *sd) {
	int path,level;
	unsigned char percentage;
	int exp_need,exp_now;
	path=sd->status.class;
	level=sd->status.level;
	if(level<99) {
		exp_need=exp_require[path][level+1]-exp_require[path][level];
		exp_now=sd->status.exp-exp_require[path][level];

		if(exp_need == 0) {
			// avoid crash..
		//	printf("clif_sendstatus() funtion has error : 1\n");
		//	printf("temporary users exp percentage setted 0\n");
			percentage=0;
		} else {
			percentage=((float)exp_now/exp_need)*100;
		}
		
	} else {
		percentage=((float)sd->status.exp/0xFFFFFFFF)*100;
	}
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 3) = 0x08;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOB(sd->fd, 5) = 0x78;
	WFIFOB(sd->fd, 6) = 0x01;
	WFIFOB(sd->fd, 7) = sd->status.country;
	WFIFOB(sd->fd, 8) = sd->status.totem;
	WFIFOB(sd->fd, 9) = 0x01;
	WFIFOB(sd->fd, 10) = sd->status.level;
	WFIFOL(sd->fd, 11) = SWAP32(sd->max_hp);
	WFIFOL(sd->fd, 15) = SWAP32(sd->max_mp);
	WFIFOB(sd->fd, 19) = sd->might;
	WFIFOB(sd->fd, 20) = sd->will;
	WFIFOB(sd->fd, 21) = 1;
	WFIFOB(sd->fd, 22) = 1;
	WFIFOB(sd->fd, 23) = sd->grace;
	WFIFOB(sd->fd, 24) = 1;
	WFIFOB(sd->fd, 25) = 1;//sd->status.country;
	WFIFOB(sd->fd, 26) = 0x63;
	WFIFOB(sd->fd, 27) = 0x6D;
	WFIFOB(sd->fd, 28) = 0x15;
	WFIFOB(sd->fd, 29) = 0xB9;
	WFIFOB(sd->fd, 30) = 0;
	WFIFOB(sd->fd, 31) = 0;
	WFIFOB(sd->fd, 32) = 0;
	WFIFOB(sd->fd, 33) = 2;
	WFIFOL(sd->fd, 34) = SWAP32(sd->status.hp);
	WFIFOL(sd->fd, 38) = SWAP32(sd->status.mp);
	WFIFOL(sd->fd, 42) = SWAP32(sd->status.exp);
	WFIFOL(sd->fd, 46) = SWAP32(sd->status.money);
	WFIFOB(sd->fd, 50) = (unsigned char)percentage; //exp percent
	WFIFOB(sd->fd, 51) = 0; // reversed key 
	WFIFOB(sd->fd, 52) = 0; // dark sight
	WFIFOB(sd->fd, 53) = 0; // if it setted 1, this character is being cannot say
	WFIFOB(sd->fd, 54) = 0;
	WFIFOB(sd->fd, 55) = 1;
	WFIFOB(sd->fd, 56) = 1;
	WFIFOB(sd->fd, 57) = 1;
	WFIFOB(sd->fd, 58) = 1;
	WFIFOB(sd->fd, 59) = 1;
	WFIFOB(sd->fd, 60) = 0xB3;
	WFIFOB(sd->fd, 61) = 1; //  
	WFIFOB(sd->fd, 62) = 0xB3; // нƮ 
	WFIFOB(sd->fd, 63) = 0xB3;
	WFIFOB(sd->fd, 64) = 0xB3;
	WFIFOB(sd->fd, 65) = 0xB3;
	
	WFIFOW(sd->fd, 1) = SWAP16(65);

	WFIFOSET(sd->fd, crypt(WFIFOP(sd->fd, 0)) );
	return 0;
}

int clif_spawn(struct map_sessiondata *sd) {
	map_addblock(&sd->bl);
	clif_sendchararea(sd);
	return 0;
}

void dumptest1( char *data, int size ) 
{
	int i,j,pos,npos;
	int retval;
	int line=0;
	for( i=0; i<size; i++) {
		retval=printf("%d",data[i]);
		for(j=0;j<4-retval;j++) printf(" ");
		pos=i%10;
		if( pos == 9) {
			printf("     ");
			npos=line*10-1;
			for(j=0; j<10; j++) printf("%c",data[npos+j]);

			printf("\n");
			line++;
		}

	}
//	printf("\ntest result: i:%d\n",i);
}
int clif_parsewalk(struct map_sessiondata *sd) {
	int moveblock;
	int dx, dy, xold, yold;
	int x0=0,y0=0,x1=0,y1=0;
	unsigned short checksum;
	char *buf;
//	dumptest1(RFIFOP(sd->fd, 0), (int)SWAP16(RFIFOW(sd->fd,1) ));
	int ret;
	struct block_list *bl=NULL;
	
	

	if(RFIFOB(sd->fd,3)==6) {
		
		xold=dx = SWAP16(RFIFOW(sd->fd,7));
		yold=dy = SWAP16(RFIFOW(sd->fd,9));
		checksum=SWAP16(RFIFOW(sd->fd,17));
		
	}else{

		xold=dx=sd->bl.x;
		yold=dy=sd->bl.y;
	}
	
	x0=SWAP16(RFIFOW(sd->fd,11));
	y0=SWAP16(RFIFOW(sd->fd,13));
	x1=RFIFOB(sd->fd,15);
	y1=RFIFOB(sd->fd,16);

	switch(RFIFOB(sd->fd, 5)) {
	case 0:
		dy-=1;
		break;
	case 1:
		dx+=1;
		break;
	case 2:
		dy+=1;
		break;
	case 3:
		dx-=1;
		break;
	}

	bl=map_firstincell2(sd->bl.m,dx,dy,0);
	if(bl!=NULL) {
		clif_sendxy(sd);
		return 0;
	}
	ret = do_warp(sd,sd->bl.m,dx,dy);
	if(ret > 0) {
		clif_sendminitext(sd,map_msg[ret].message,map_msg[ret].len);
		clif_sendxy(sd);
		return 0;
	} else if( ret == 0) {
		return 0;
	}

	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x0C;
	WFIFOB(sd->fd, 3) = 0x26;
	WFIFOB(sd->fd, 4) = 0x03;

	WFIFOB(sd->fd, 5) = RFIFOB(sd->fd, 5);
	WFIFOW(sd->fd, 6) = SWAP16(xold);
	WFIFOW(sd->fd, 8) = SWAP16(yold);
	if (RFIFOB(sd->fd, 3) == 0x06) {
		WFIFOW(sd->fd, 10) = SWAP16(8);
		WFIFOW(sd->fd, 12) = SWAP16(7);
	} else {
		WFIFOW(sd->fd, 10) = SWAP16(dx);
		WFIFOW(sd->fd, 12) = SWAP16(dy);
	}
	WFIFOB(sd->fd, 14) = RFIFOB(sd->fd, 6);
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 15);


	if (dx == sd->bl.x && dy == sd->bl.y)
		return 0;

	CREATE(buf, char, 15);
	WBUFB(buf, 0) = 0xAA;
	WBUFB(buf, 1) = 0x00;
	WBUFB(buf, 2) = 0x0C;
	WBUFB(buf, 3) = 0x0C;
	WBUFB(buf, 4) = 0x03;
	WBUFL(buf, 5) = SWAP32(sd->status.id);
	WBUFW(buf, 9) = SWAP16(xold);
	WBUFW(buf, 11) = SWAP16(yold);
	WBUFB(buf, 13) = RFIFOB(sd->fd, 5);
	WBUFB(buf, 14) = RFIFOB(sd->fd, 6);
	crypt(WBUFP(buf, 0));
	clif_send(buf, 15, &sd->bl, AREA_WOS);
	aFree(buf);

	moveblock = (sd->bl.x/BLOCK_SIZE != dx/BLOCK_SIZE || sd->bl.y/BLOCK_SIZE != dy/BLOCK_SIZE);

	if(moveblock) map_delblock(&sd->bl);
	sd->bl.x = dx;
	sd->bl.y = dy;
	if(moveblock) map_addblock(&sd->bl);
	
	
	if (RFIFOB(sd->fd, 3) == 0x06) {
		clif_sendmapdata(sd, sd->bl.m, x0, y0, x1, y1,checksum);
		//this is where all the "finding" code goes
		map_foreachinblock(clif_npc_look,sd->bl.m,x0,y0,x0+(x1-1),y0+(y1-1),BL_NPC,LOOK_GET,sd);
		map_foreachinblock(clif_mob_look,sd->bl.m,x0,y0,x0+(x1-1),y0+(y1-1),BL_MOB,LOOK_GET,sd);
		map_foreachinblock(clif_charlook_sub,sd->bl.m,x0,y0,x0+(x1-1),y0+(y1-1),BL_PC,LOOK_GET,sd);
		map_foreachinblock(clif_charlook_sub,sd->bl.m,x0,y0,x0+(x1-1),y0+(y1-1),BL_PC,LOOK_SEND,sd);
   }

	map_foreachincell(pc_runfloor_sub,sd->bl.m,sd->bl.x,sd->bl.y,BL_NPC,sd);
	
//	printf("ret:%d\n",ret);
//	printf("info: dx=%d dy=%d xold=%d yold=%d x0=%d y0=%d x1=%d y1=%d\n",dx, dy, xold, yold, x0, y0, x1, y1);
	return 0;
}

int clif_parsemap(struct map_sessiondata *sd) {
	int x0 = SWAP16(RFIFOW(sd->fd, 5));
	int y0 = SWAP16(RFIFOW(sd->fd, 7));
	int x1 = RFIFOB(sd->fd, 9);
	int y1 = RFIFOB(sd->fd, 10);
	unsigned short checksum = SWAP16(RFIFOW(sd->fd,11));
	
//	printf("parse_map: %d,%d,%d,%d\n",x0,y0,x1,y1);
//	{ int i; printf("parsemap: ");for(i=0;i< 20; i++) printf("%d ",RFIFOB(sd->fd, i)); printf("\n"); }
	clif_sendmapdata(sd, sd->bl.m, x0, y0, x1, y1,checksum);
	
	
//	map_foreachincell(pc_runfloor_sub,sd->bl.m,sd->bl.x,sd->bl.y,BL_NPC,sd);
	map_foreachinarea(clif_npc_look,sd->bl.m,sd->bl.x,sd->bl.y,AREA,BL_NPC,LOOK_GET,sd);
	map_foreachinarea(clif_mob_look,sd->bl.m,sd->bl.x,sd->bl.y,AREA,BL_MOB,LOOK_GET,sd);
	
    return 0;
}

int clif_sendmapdata(struct map_sessiondata *sd, int m, int x0, int y0, int x1, int y1,unsigned short check) {
	int x,y,pos,xv,yv;
//	unsigned short checksum;
//	checksum=0;
	if (x0 < 0)
		x0 = 0;
	if (y0 < 0)
		y0 = 0;
	if (x1 > map[m].xs)
		x1 = map[m].xs;
	if (y1 > map[m].ys)
		y1 = map[m].ys;
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 3) = 0x06;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOB(sd->fd, 5)= 0;
	WFIFOW(sd->fd, 6) = SWAP16(x0);
	WFIFOW(sd->fd, 8) = SWAP16(y0);
	WFIFOB(sd->fd, 10) = x1;
	WFIFOB(sd->fd, 11) = y1;
	pos = 12;

	for (y = 0; y < y1; y++) {
		if (y+y0 >= map[m].ys)
			break;
		for (x = 0; x < x1; x++) {
			if (x+x0 >= map[m].xs)
				break;
			xv=x0+x;
			yv=y0+y;

			WFIFOW(sd->fd, pos) = SWAP16(read_tile(m, xv, yv));
			pos+=2;
			WFIFOW(sd->fd, pos) = SWAP16(read_pass(m, xv, yv));
			pos+=2;
			WFIFOW(sd->fd, pos) = SWAP16(read_obj(m, xv, yv));
			pos+=2;
		}
	}
	
//	checksum=crcbitbybit(buf,len);

	if (pos <= 12)
		return 0;
	//printf("Checksum Created: %d\n, Checksum required: %d\n",checksum,check);
	WFIFOW(sd->fd, 1) = SWAP16(pos-3);
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, pos);

	
	return 0;
}

int clif_sendside(struct map_sessiondata *sd) {
	unsigned char buf[10];
	WBUFB(buf, 0) = 0xAA;
	WBUFB(buf, 1) = 0x00;
	WBUFB(buf, 2) = 0x07;
	WBUFB(buf, 3) = 0x11;
	WBUFB(buf, 4) = 0x03;
	WBUFL(buf, 5) = SWAP32(sd->status.id);
	WBUFB(buf, 9) = sd->status.side;
	crypt(WBUFP(buf, 0));
	clif_send(buf, 10, &sd->bl, AREA_WOS);
	return 0;
}
int clif_sendmob_side(struct mobspawn_data *mob) {
	unsigned char buf[10];
	WBUFB(buf, 0) = 0xAA;
	WBUFB(buf, 1) = 0x00;
	WBUFB(buf, 2) = 0x07;
	WBUFB(buf, 3) = 0x11;
	WBUFB(buf, 4) = 0x03;
	WBUFL(buf, 5) = SWAP32(mob->bl.id);
	WBUFB(buf, 9) = mob->side;
	crypt(WBUFP(buf, 0));
	clif_send(buf, 10, &mob->bl, AREA_WOS);
	return 0;
}
int clif_parseside(struct map_sessiondata *sd) {
	sd->status.side = RFIFOB(sd->fd, 5);
	clif_sendside(sd);
	return 0;
}

int clif_parseemotion(struct map_sessiondata *sd) {
	clif_sendaction(sd, RFIFOB(sd->fd, 5)+11, 0x4E, 0);
	return 0;
}

int clif_sendmsg(struct map_sessiondata *sd, int type, char *buf, int len) {
/*	Type:
	0 = Wisp (blue text)
	3 = Mini Text/Status Text
*/
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOW(sd->fd, 1) = SWAP16(5+len);
	WFIFOB(sd->fd, 3) = 0x0A;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOW(sd->fd, 5) = type;
	WFIFOB(sd->fd, 7) = len;
	memcpy(WFIFOP(sd->fd, 8), buf, len);
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 8+len);
	return 0;
}

int clif_sendminitext(struct map_sessiondata *sd, char *msg, int len) {
	clif_sendmsg(sd, 3, msg, len);
	return 0;
}

int clif_sendwisp(struct map_sessiondata *sd, char *srcname, unsigned char *msg, int msglen) {
	int srclen = strlen(srcname);
	unsigned char *buf;

	CREATE(buf, unsigned char, 2+srclen+msglen);
	memcpy(WBUFP(buf, 0), srcname, srclen);
	WBUFB(buf, srclen) = '"';
	WBUFB(buf, 1+srclen) = ' ';
	memcpy(WBUFP(buf, 2+srclen), msg, msglen);
	clif_sendmsg(sd, 0, buf, 2+srclen+msglen);
	aFree(buf);
	return 0;
}

int clif_retrwisp(struct map_sessiondata *sd, char *dstname, unsigned char *msg, int msglen) {
	int dstlen = strlen(dstname);
	unsigned char buf[256]={0,};


	memcpy(WBUFP(buf, 0), dstname, dstlen);
	WBUFB(buf, dstlen) = '>';
	WBUFB(buf, 1+dstlen) = ' ';
	memcpy(WBUFP(buf, 2+dstlen), msg, msglen);
	clif_sendmsg(sd, 0, buf, 2+dstlen+msglen);

	return 0;
}

int clif_failwisp(struct map_sessiondata *sd, const char *name) {
	char buf[256]={0,};
	sprintf(buf,map_msg[MAP_WHISPFAIL].message, name);
	clif_sendmsg(sd, 0, buf, strlen(buf));
	return 0;
}

int clif_parsedropitem(struct map_sessiondata *sd) {
	int id=RFIFOB(sd->fd, 5)-1;
	int all=RFIFOB(sd->fd, 6);


	clif_sendaction(sd, 5, 20, 0);
	pc_dropitemmap(sd, id, all);
	return 0;
}

int clif_senddelitem(struct map_sessiondata *sd, int num, int type) {
/*	Type:
	0 = Remove	5 = Shot	10 = Sold
	1 = Drop	6 = Used	11 = Removed
	2 = Eat		7 = Posted	12 = *Item name*
	3 = Smoke	8 = Decayed	13 = Broke
	4 = Throw	9 = Gave
*/
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x06;
	WFIFOB(sd->fd, 3) = 0x10;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOB(sd->fd, 5) = num+1;
	WFIFOB(sd->fd, 6) = type;
	WFIFOB(sd->fd, 7) = 0x00;
	WFIFOB(sd->fd, 8) = 0x00;
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 9);
	return 0;
}

int clif_sendadditem(struct map_sessiondata *sd, int num) {
	char buf[256]={0,};
	char *name;
	int namelen;
	int len;
	int id=sd->status.inventory[num].id;
	int type;

	if(!id) return 0; // ݴ Ѥ;
	if (strlen(sd->status.inventory[num].real_name))
		name = sd->status.inventory[num].real_name;
	else
		name = itemdb_name(id);

	type=itemdb_type(sd->status.inventory[num].id);
	if( type== ITM_SMOKE /*|| type== ITM_DRINK*/) {
		namelen=sprintf(buf, "%s [ %d %s]", sd->status.inventory[num].real_name, sd->status.inventory[num].dura, itemdb_unitchar(id));

	} else if (sd->status.inventory[num].amount > 1) {
		namelen=sprintf(buf, "%s %d", sd->status.inventory[num].real_name, sd->status.inventory[num].amount);
		
	} else {
		namelen=sprintf(buf, "%s", sd->status.inventory[num].real_name);
	}
	
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 3) = 0x0F;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOB(sd->fd, 5) = num+1;
	WFIFOW(sd->fd, 6) = SWAP16(itemdb_icon(id));
	WFIFOB(sd->fd, 8) = itemdb_iconcolor(id);
	WFIFOB(sd->fd, 9) = namelen;
	memcpy(WFIFOP(sd->fd, 10), buf, namelen);
	len=namelen;
	WFIFOW(sd->fd, 1) = SWAP16(len+8);
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd, namelen+11);
	
	return 0;
}

int clif_sendequip(struct map_sessiondata *sd, int id) {
	char buff[256];
	char *name;
	int msgnum=0;
	
	switch (id) {
	case EQ_HELM:
		msgnum=MAP_EQHELM;
		break;
	case EQ_WEAP:
		msgnum=MAP_EQWEAP;
		break;
	case EQ_ARMOR:
		msgnum=MAP_EQARMOR;
		break;
	case EQ_SHIELD:
		msgnum=MAP_EQSHIELD;
		break;
	case EQ_RIGHT:
		msgnum=MAP_EQRIGHT;
		break;
	case EQ_LEFT:
		msgnum=MAP_EQLEFT;
		break;
	default:
		return -1;
		break;
	}


	name = sd->status.equip[id].real_name;

	//if(itemdb_equipscript(sd->status.equip[id].id)) {
	//sd->target=sd->bl.id;
	//run_script(itemdb_equipscript(sd->status.equip[id].id),0,sd->bl.id,0);
	//}
	sprintf(buff, map_msg[msgnum].message, name);
	clif_sendminitext(sd, buff, strlen(buff));
	return 0;
}

int clif_parsewisp(struct map_sessiondata *sd) {
	char dst_name[16];
	struct map_sessiondata *dst_sd;
	int dstlen, srclen, msglen;
	dstlen = RFIFOB(sd->fd, 5);
	if(dstlen > 12) { // Avoid crash by buffer overflow
		return 0;
	}
	srclen = strlen(sd->status.name);
	msglen = RFIFOB(sd->fd, 5+dstlen);
	memset(dst_name, 0, 16);
	memcpy(dst_name, RFIFOP(sd->fd, 6), dstlen);
	dst_sd = map_name2sd(dst_name);
	if (!dst_sd) {
		intif_wisp(sd->status.name, dst_name, RFIFOP(sd->fd, 7+dstlen), msglen);
	} else {
		clif_sendwisp(dst_sd, sd->status.name, RFIFOP(sd->fd, 7+dstlen), msglen);
		clif_retrwisp(sd, dst_sd->status.name, RFIFOP(sd->fd, 7+dstlen), msglen);
	}
	return 0;
}

int clif_sendsay(struct map_sessiondata *sd, char *msg, int msglen, int type) {
/*	Type:
	0 = Talk
	1 = Shout
	2 = Skill
*/
	char *buf;
	int namelen = strlen(sd->status.name);
	CREATE(buf, char, 13+namelen+msglen);
	WBUFB(buf, 0) = 0xAA;
	WBUFW(buf, 1) = SWAP16(10+namelen+msglen);
	WBUFB(buf, 3) = 0x0D;
	WBUFB(buf, 4) = 0x03;
	WBUFB(buf, 5) = type;
	WBUFL(buf, 6) = SWAP32(sd->status.id);
	WBUFB(buf, 10) = namelen+msglen+2;
	memcpy(WBUFP(buf, 11), sd->status.name, namelen);
	if (type==1)
		WBUFB(buf, 11+namelen) = '!';
	else
		WBUFB(buf, 11+namelen) = ':';
	WBUFB(buf, 12+namelen) = ' ';
	memcpy(WBUFP(buf, 13+namelen), msg, msglen);
	crypt(WBUFP(buf, 0));
	if (type==1) {
		clif_send(buf, 13+namelen+msglen, &sd->bl, SAMEMAP);
		
	} else {
		clif_send(buf, 13+namelen+msglen, &sd->bl, AREA);
	}
	
	aFree(buf);
	return 0;
}

int clif_parsesay(struct map_sessiondata *sd) {
	if (is_command(sd, RFIFOP(sd->fd, 7), RFIFOB(sd->fd, 6)))
		return 0;
	clif_sendsay(sd, RFIFOP(sd->fd, 7), RFIFOB(sd->fd, 6), RFIFOB(sd->fd, 5));
	return 0;
	
}

int clif_parserefresh(struct map_sessiondata *sd) {
	clif_sendxy(sd);
	clif_getchararea(sd);
	return 0;
}

int clif_sendupdatestatus(struct map_sessiondata *sd) {
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 1) = 0x00;
	WFIFOB(sd->fd, 2) = 0x1E;
	WFIFOB(sd->fd, 3) = 0x08;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOB(sd->fd, 5) = 0x38;
	WFIFOL(sd->fd, 6) = SWAP32(sd->status.hp);
	WFIFOL(sd->fd, 10) = SWAP32(sd->status.mp);
	WFIFOL(sd->fd, 14) = SWAP32(sd->status.exp);
	WFIFOL(sd->fd, 18) = SWAP32(sd->status.money);
	WFIFOL(sd->fd, 22) = 0x00;
	WFIFOB(sd->fd, 26) = 0x00;
	WFIFOB(sd->fd, 27) = 0x00;
	WFIFOB(sd->fd, 28) = 0x01;
	WFIFOB(sd->fd, 29) = 0x00;
	WFIFOB(sd->fd, 30) = 0x00;
	WFIFOB(sd->fd, 31) = 0x73;
	WFIFOB(sd->fd, 32) = 0x35;
	crypt(WFIFOP(sd->fd, 0));
	WFIFOSET(sd->fd, 33);
	return 0;
}

int clif_sendbluemessage(struct map_sessiondata *sd, char *msg) {
	
	WFIFOB(sd->fd, 0) = 0xAA;
	WFIFOB(sd->fd, 3) = 0x0A;
	WFIFOB(sd->fd, 4) = 0x03;
	WFIFOW(sd->fd, 5) = 0;
	WFIFOB(sd->fd, 7) = strlen(msg);
	strcpy(WFIFOP(sd->fd,8), msg);
	WFIFOW(sd->fd, 1)= SWAP16(strlen(msg)+5);
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd, strlen(msg)+8);
	return 0;

}

int clif_sendaction(struct map_sessiondata *sd, int type, int time, int sound) {
/*	Type:
	0 = Stand/None	6 = Magic
	1 = Attack	7 = Eat*
	2 = Throw	8 = Eat
	3 = Shot
	4 = Sit
	5 = Sit*
*/
	unsigned char buf[14];
	WBUFB(buf, 0) = 0xAA;
	WBUFB(buf, 1) = 0x00;
	WBUFB(buf, 2) = 0x0B;
	WBUFB(buf, 3) = 0x1A;
	WBUFB(buf, 4) = 0x03;
	WBUFL(buf, 5) = SWAP32(sd->status.id);
	WBUFB(buf, 9) = type;
	WBUFB(buf, 10) = 0x00;
	WBUFB(buf, 11) = time;
	WBUFB(buf, 12) = sound;
	WBUFB(buf, 13) = 0x00;
	crypt(WBUFP(buf, 0));
	clif_send(buf, 14, &sd->bl, AREA);
	return 0;
}

int clif_sendsound(struct map_sessiondata *sd, int sound) {
	char buf[32]={0,};

	
	WBUFB(buf, 0) = 0xAA;
	WBUFW(buf, 1) = SWAP16(18);
	WBUFB(buf, 3) = 0x19;
	WBUFL(buf, 5) = SWAP32(sound);
//	printf("%x %x %x %d Ҹ\n",buf[4],buf[5],buf[6],buf[7]);
	WBUFB(buf, 9) = 0x64;
	WBUFB(buf, 10) = 0x00;
	WBUFB(buf, 11) = 0x04;
	WBUFL(buf, 12) = SWAP32(sd->status.id);
	WBUFB(buf, 16) = 1;
	WBUFB(buf, 17) = 0;
	WBUFB(buf, 18) = 2;
	WBUFB(buf, 19) = 2;
	WBUFB(buf, 20) = 0;
	WBUFB(buf, 21) = 4;

	
	clif_send(buf, crypt(WBUFP(buf, 0)), &sd->bl, AREA);
	
}
int clif_sendmob_action(struct mobspawn_data *mob, int type, int sound, int time) {
	
/*	Type:
	0 = Stand/None	6 = Magic
	1 = Attack	7 = Eat*
	2 = Throw	8 = Eat
	3 = Shot
	4 = Sit
	5 = Sit*
*/
	unsigned char buf[14];
	WBUFB(buf, 0) = 0xAA;
	WBUFB(buf, 1) = 0x00;
	WBUFB(buf, 2) = 0x0B;
	WBUFB(buf, 3) = 0x1A;
	WBUFB(buf, 4) = 0x03;
	WBUFL(buf, 5) = SWAP32(mob->bl.id);
	WBUFB(buf, 9) = type;
	WBUFB(buf, 10) = 0x00;
	WBUFB(buf, 11) = time;
	WBUFB(buf, 12) = sound;
	WBUFB(buf, 13) = 0x00;
	crypt(WBUFP(buf, 0));
	clif_send(buf, 14, &mob->bl, AREA);
	return 0;
}
int clif_sendanimation(struct block_list *bl, va_list ap) {
	struct block_list *t;
	struct map_sessiondata *sd;
	int anim;
	
	anim=va_arg(ap,int);
	nullpo_ret(t=va_arg(ap,struct block_list*));
	nullpo_ret(sd=(struct map_sessiondata*)bl);
	


	WFIFOB(sd->fd,0)=0xAA;
	WFIFOW(sd->fd,1)=SWAP16(0x0B);
	WFIFOB(sd->fd,3)=41;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOL(sd->fd,5)=SWAP32(t->id);
	WFIFOW(sd->fd,9)=SWAP16(anim);
	WFIFOW(sd->fd,11)=0x00;
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,14);
	//clif_send(buf, 14, bl, AREA);
	return 0;
}

int clif_sendmagic(struct map_sessiondata *sd,int pos) {
/*	int len;
	int id;
	int type;
	char* name;
	char* question;
	
	id=sd->status.skill[pos];
	name=magicdb_name(id);
	question=magicdb_question(id);
	type=magicdb_type(id);
	
	
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,3)=0x17;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOB(sd->fd,5)=pos+1;
	WFIFOB(sd->fd,6)=type; //this is type
	WFIFOB(sd->fd,7)=strlen(name);
	len=strlen(name);
	strcpy(WFIFOP(sd->fd,8),name);
	WFIFOB(sd->fd,len+8)=strlen(question);
	strcpy(WFIFOP(sd->fd,len+9),question);
	len=len+strlen(question);
	WFIFOW(sd->fd,1)=SWAP16(len+5);
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,len+8);
*/	return 0;
}

int clif_parsemagic(struct map_sessiondata* sd) {
//	char* answer;
	int pos;
	int time;
	int i;
//	int newtime;

	char msg[255];
	int len;
	
	pos=RFIFOB(sd->fd,5)-1;
	
	i=clif_has_aethers(sd,sd->status.skill[pos]);
	if(i>0) {
		time=i/1000;
		len=sprintf(msg,"%d Ŀ  ֽϴ",time);
		clif_sendminitext(sd,msg,len);
		return 0;
	}
	
	switch(magicdb_type(sd->status.skill[pos])) {
	case 2:
		sd->target=SWAP32(RFIFOL(sd->fd,6));
		break;
	case 5:
		sd->target=sd->bl.id;
		break;
	
	default:
		break;
	}
	
	clif_sendaction(sd,6,30,0);
	run_script(magicdb_script(sd->status.skill[pos]),0,sd->bl.id,0);
	
	return 0;
}

int clif_scriptmes(struct map_sessiondata *sd, int id,  char *msg, int previous, int next) {
    unsigned short graphic_id=sd->npc_g;
	unsigned char color=sd->npc_gc;
	char *buf;
	int len=strlen(msg);	
	int i;
	CREATE(buf, char, len);
	memcpy(buf, msg, len);
	for(i=0;i<len;i++) {
		if(buf[i]=='\\' && buf[i+1]=='n') {
			memmove(&buf[i],&buf[i]+1,len-i);
			buf[i]='\n';
			len--;
			buf[len]='\0';
		}
	}
	WFIFOB(sd->fd,0)=0xAA;
    WFIFOB(sd->fd,3)=0x30;
	WFIFOB(sd->fd,4)=0x03;

	WFIFOW(sd->fd,5)=SWAP16(1);
	WFIFOL(sd->fd,7)=SWAP32(id);
//	printf("npc_id: %d\n",id);
    if (graphic_id==0)
       WFIFOB(sd->fd,11)=0;
    else if (graphic_id>49152)
       WFIFOB(sd->fd,11)=2;
    else
        WFIFOB(sd->fd,11)=1;
    
	WFIFOB(sd->fd,12)=1;
	WFIFOW(sd->fd,13)=SWAP16(graphic_id); // graphic id
	
	WFIFOB(sd->fd,15)=color; //graphic color
	WFIFOB(sd->fd,16)=1;
	WFIFOW(sd->fd,17)=SWAP16(graphic_id);
	WFIFOB(sd->fd,19)=color;
    WFIFOL(sd->fd,20)=SWAP32(1);
	WFIFOB(sd->fd,24)=next; // Previous
	WFIFOB(sd->fd,25)=previous; // Next
	WFIFOB(sd->fd,26)=0;
	WFIFOB(sd->fd,27)=len;
	strcpy(WFIFOP(sd->fd,28),buf);
	WFIFOW(sd->fd,1)=SWAP16(len+25);
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,len+28);

//	printf("script debug: %s\n",msg);
//    printf("total count: %d\n",strlen(msg)+28);
    aFree(buf);
	return 0;
}
int clif_scriptinput(struct map_sessiondata *sd, int id, char* msg) {
	unsigned short graphic_id=sd->npc_g;
	unsigned char color=sd->npc_gc;
	int len=strlen(msg);	
	int packetlen;
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,1) = (27+len)>>8;
	WFIFOB(sd->fd,2) = (27+len)%256;  
	WFIFOB(sd->fd,3)='\x2F';
	WFIFOB(sd->fd,4)=0x00;
	WFIFOB(sd->fd,5)=3; // 3 is write window
	WFIFOB(sd->fd,6)=3;
	WFIFOL(sd->fd,7)=SWAP32(id);
	WFIFOB(sd->fd,12)=1;
	WFIFOW(sd->fd,13)=SWAP16(graphic_id); // graphic id
	WFIFOB(sd->fd,15)=color; //graphic color
	WFIFOB(sd->fd,16)=1;
	WFIFOW(sd->fd,17)=SWAP16(graphic_id);
	WFIFOB(sd->fd,19)=color;
    WFIFOL(sd->fd,20)=0;
	
	WFIFOB(sd->fd,21)=len;
	strcpy(WFIFOP(sd->fd,22),msg);

	WFIFOB(sd->fd,22+len)=4;
	WFIFOB(sd->fd,23+len)='0x4d';
	WFIFOB(sd->fd,24+len)='0x65';
	WFIFOB(sd->fd,25+len)='0x61';
	WFIFOB(sd->fd,26+len)='0x74';
	WFIFOB(sd->fd,27+len)='0x00';
	WFIFOB(sd->fd,28+len)='0x4c';
	WFIFOB(sd->fd,29+len)='0x00';

	packetlen = crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,packetlen);
}
int clif_scriptmenu(struct map_sessiondata *sd, int id, char* dialog, char* menu[], int size) {
    int graphic=sd->npc_g;
    int color=sd->npc_gc;
   	int x;
   	int len;

	int menulen;
	printf("npc_id_menu: %d\n",id);

	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,3)=0x2F; //THis is a menu packet
	WFIFOB(sd->fd,4)=3;
	WFIFOB(sd->fd,5)=1;
	WFIFOB(sd->fd,6)=1; // 1 is menu window
	WFIFOL(sd->fd,7)=SWAP32(id);
	
	WFIFOB(sd->fd, 11) = 0;
	WFIFOB(sd->fd,12)=1;
    
	WFIFOW(sd->fd,13)=SWAP16(graphic);
	WFIFOB(sd->fd,15)=color;
	WFIFOB(sd->fd,16)=0;
	WFIFOW(sd->fd,17)=0;
	WFIFOB(sd->fd,19)=0;
	WFIFOB(sd->fd,20)=0;
   	len=strlen(dialog);
	WFIFOB(sd->fd,21)=len&0xFF;
	memcpy(WFIFOP(sd->fd,22),dialog,len);
	WFIFOB(sd->fd,len+22)=0;
	WFIFOB(sd->fd,len+23)=size&0xFF;
    for(x=1;x<size+1;x++){
		menulen=strlen(menu[x]);
        WFIFOB(sd->fd,len+24)=menulen;
        memcpy(WFIFOP(sd->fd,len+25),menu[x],menulen);
        len+=menulen;
        WFIFOW(sd->fd,len+25)=SWAP16(x);
        len+=3;
    }
    WFIFOW(sd->fd,1)=SWAP16(len+22);
    WFIFOSET(sd->fd,crypt(WFIFOP(sd->fd,0)));
	return 0;
}
int clif_scriptmenu2(struct map_sessiondata *sd, int id, char* dialog, char* menu[], int size) {

	int graphic=sd->npc_g;
    int color=sd->npc_gc;
   	int x;
	int menulen;
	
	int len = strlen(dialog);

	WFIFOB(sd->fd,0)=0xAA;
    WFIFOB(sd->fd,3)=0x30;
	WFIFOB(sd->fd,4)=0x03;

	WFIFOW(sd->fd,5)=2;
	WFIFOW(sd->fd,6)=2;
	WFIFOL(sd->fd,7)=SWAP32(id);
    WFIFOB(sd->fd,11)=1;
	WFIFOB(sd->fd,12)=1;
	WFIFOW(sd->fd,13)=SWAP16(graphic); // graphic id
	WFIFOB(sd->fd,15)=color; //graphic color
	WFIFOB(sd->fd,16)=1;
	WFIFOW(sd->fd,17)=SWAP16(graphic);
	WFIFOB(sd->fd,19)=color;
    WFIFOL(sd->fd,20)=SWAP32(1);
	WFIFOB(sd->fd,24)=0;
	WFIFOB(sd->fd,25)=0;
	WFIFOB(sd->fd,26)=0;
	WFIFOB(sd->fd,27)=len;
	strcpy(WFIFOP(sd->fd,28),dialog);
	WFIFOB(sd->fd,28+len)=size;

    for(x=1;x<size+1;x++){
		menulen=strlen(menu[x]);
        WFIFOB(sd->fd,len+29)=menulen;
        memcpy(WFIFOP(sd->fd,len+30),menu[x],menulen);
		len+=menulen+1;
    }
    

	WFIFOW(sd->fd,1)=SWAP16(len+30);

	WFIFOSET(sd->fd,crypt(WFIFOP(sd->fd,0)));
}
int clif_usermenu(struct map_sessiondata *sd, int id, char* dialog, char* menu[], int size) {
    int graphic=sd->npc_g;
    int color=sd->npc_gc;
   	int x;
   	int len;

	int menulen;
	printf("npc_id_menu: %d\n",id);

	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,3)=0x2F; //THis is a menu packet
	WFIFOB(sd->fd,4)=3;
	WFIFOB(sd->fd,5)=1;
	WFIFOB(sd->fd,6)=8; // 1 is menu window
	WFIFOL(sd->fd,7)=SWAP32(id);
	
	WFIFOB(sd->fd, 11) = 0;
	WFIFOB(sd->fd,12)=1;
    
	WFIFOW(sd->fd,13)=SWAP16(graphic);
	WFIFOB(sd->fd,15)=color;
	WFIFOB(sd->fd,16)=0;
	WFIFOW(sd->fd,17)=0;
	WFIFOB(sd->fd,19)=0;
	WFIFOB(sd->fd,20)=0;
   	len=strlen(dialog);
	WFIFOB(sd->fd,21)=len&0xFF;
	memcpy(WFIFOP(sd->fd,22),dialog,len);
	WFIFOB(sd->fd,len+22)=0;
	WFIFOB(sd->fd,len+23)=size&0xFF;
    for(x=1;x<size+1;x++){
		menulen=strlen(menu[x]);
        WFIFOB(sd->fd,len+24)=menulen;
        memcpy(WFIFOP(sd->fd,len+25),menu[x],menulen);
        len+=menulen;
        WFIFOW(sd->fd,len+25)=SWAP16(x);
        len+=3;
    }
    WFIFOW(sd->fd,1)=SWAP16(len+22);
    WFIFOSET(sd->fd,crypt(WFIFOP(sd->fd,0)));
	return 0;
}

int clif_scriptcallshopbuy(struct map_sessiondata *sd, struct npc_shop_data *nsd, int id, char mainmsg[]) {
	
	int graphic=sd->npc_g;
    int color=sd->npc_gc;
   	int i;
	int tempsize = 0;
	int itemnamelen;
	int mainlen = strlen(mainmsg);

	if(nsd == NULL) { 
		return 0;
	}

	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,3)=0x2F;
	WFIFOB(sd->fd,4)=0;

	WFIFOB(sd->fd,5)=4;
	WFIFOB(sd->fd,6)=4; // Ʈ â 
    WFIFOL(sd->fd,7)=SWAP32(id);

	WFIFOB(sd->fd,11)=0;
	WFIFOB(sd->fd,12)=1;
	WFIFOW(sd->fd,13)=SWAP16(graphic); // graphic id
	WFIFOB(sd->fd,15)=color; //graphic color
	WFIFOB(sd->fd,16)=0;
	WFIFOB(sd->fd,17)=0;
	WFIFOB(sd->fd,18)=0;
	WFIFOB(sd->fd,19)=0;
	WFIFOB(sd->fd,20)=0;
	WFIFOB(sd->fd,21)=mainlen%256;
	memcpy(WFIFOP(sd->fd,22),mainmsg,mainlen);

	WFIFOW(sd->fd,22+mainlen)=SWAP16(nsd->id);
	WFIFOB(sd->fd,24+mainlen)=0;
	WFIFOB(sd->fd,25+mainlen)=nsd->itemcount;
	
	for(i=0; i<nsd->itemcount; i++) {
		itemnamelen = strlen(item_db[nsd->itemid[i]]->name);
		
		WFIFOW(sd->fd,26+tempsize+mainlen)=SWAP16(item_db[nsd->itemid[i]]->icon);
		WFIFOB(sd->fd,28+tempsize+mainlen)=item_db[nsd->itemid[i]]->icon_color;
		WFIFOL(sd->fd,29+tempsize+mainlen)=SWAP32(nsd->buyprice[i]);
		WFIFOB(sd->fd,33+tempsize+mainlen)=itemnamelen;
		memcpy(WFIFOP(sd->fd,34+tempsize+mainlen),item_db[nsd->itemid[i]]->name,itemnamelen);
		WFIFOB(sd->fd,34+tempsize+itemnamelen+mainlen)=0;
		
		tempsize+=itemnamelen+9; //15
	}
	WFIFOW(sd->fd,1) = SWAP16(35+tempsize+mainlen);
	WFIFOSET(sd->fd,crypt(WFIFOP(sd->fd,0)));
}
int clif_scriptcallshopsell(struct map_sessiondata *sd, struct npc_shop_data *nsd, int id, char mainmsg[]) {
	
	int graphic=sd->npc_g;
    int color=sd->npc_gc;
   	int i, j;
	int tempsize = 0;

	int mainlen = strlen(mainmsg);
	int cansellitemcount = 0;
	if(nsd == NULL) { 
		return 0;
	}

	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,3)=0x2F;
	WFIFOB(sd->fd,4)=0;
	WFIFOB(sd->fd,5)=5;
	WFIFOB(sd->fd,6)=5;
    WFIFOL(sd->fd,7)=SWAP32(id);
	WFIFOB(sd->fd,11)=0;
	WFIFOB(sd->fd,12)=1;
	WFIFOW(sd->fd,13)=SWAP16(graphic); // graphic id
	WFIFOB(sd->fd,15)=color; //graphic color
	WFIFOB(sd->fd,16)=0;
	WFIFOB(sd->fd,17)=0;
	WFIFOB(sd->fd,18)=0;
	WFIFOB(sd->fd,19)=0;
	WFIFOB(sd->fd,20)=0;
	WFIFOB(sd->fd,21)=mainlen%256;
	memcpy(WFIFOP(sd->fd,22),mainmsg,mainlen);
	WFIFOW(sd->fd,22+mainlen)=SWAP16(nsd->id);

	for(i=0; i<26; i++) {
		for(j=0; j<nsd->itemcount; j++) {
		if (!sd->status.inventory[i].id) continue;
			if(nsd->itemid[j] == sd->status.inventory[i].id) {
				WFIFOB(sd->fd,25+mainlen+tempsize)= i+1;
				cansellitemcount++;
				tempsize++;	
			}	
		}
	}
	WFIFOB(sd->fd,24+mainlen)=cansellitemcount; //
	WFIFOW(sd->fd,1) = SWAP16(28+mainlen);
	WFIFOSET(sd->fd,crypt(WFIFOP(sd->fd,0)));
	
}

int clif_userinput(struct map_sessiondata *sd, int id, int usertype,char msg[]) {
	unsigned short graphic_id=sd->npc_g;
	unsigned char color=sd->npc_gc;
	int len=strlen(msg);	
	int packetlen;

	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,1) = (27+len)>>8;
	WFIFOB(sd->fd,2) = (27+len)%256;  
	WFIFOB(sd->fd,3)='\x2F';
	WFIFOB(sd->fd,4)=0x00;
	WFIFOB(sd->fd,5)=3; // 3 is write window
	WFIFOB(sd->fd,6)=usertype;
	WFIFOL(sd->fd,7)=SWAP32(id);
	WFIFOB(sd->fd,12)=1;
	WFIFOW(sd->fd,13)=SWAP16(graphic_id); // graphic id
	WFIFOB(sd->fd,15)=color; //graphic color
	WFIFOB(sd->fd,16)=1;
	WFIFOW(sd->fd,17)=SWAP16(graphic_id);
	WFIFOB(sd->fd,19)=color;
    WFIFOL(sd->fd,20)=0;
	
	WFIFOB(sd->fd,21)=len;
	strcpy(WFIFOP(sd->fd,22),msg);

	WFIFOB(sd->fd,22+len)=4;
	WFIFOB(sd->fd,23+len)='0x4d';
	WFIFOB(sd->fd,24+len)='0x65';
	WFIFOB(sd->fd,25+len)='0x61';
	WFIFOB(sd->fd,26+len)='0x74';
	WFIFOB(sd->fd,27+len)='0x00';
	WFIFOB(sd->fd,28+len)='0x4c';
	WFIFOB(sd->fd,29+len)='0x00';

	packetlen = crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,packetlen);
}

int clif_parseuseitem(struct map_sessiondata *sd) {
	pc_useitem(sd, RFIFOB(sd->fd, 5)-1);
	return 0;
}

int clif_parsegetitem(struct map_sessiondata *sd) {
	clif_sendaction(sd, 4, 40, 0);
	pc_getitemmap(sd, RFIFOB(sd->fd, 5));
	return 0;
}

int clif_parseunequip(struct map_sessiondata *sd) {
	int type;
	
	switch(RFIFOB(sd->fd, 5)) {
	case 0x01:
		type=EQ_WEAP;
		break;
	case 0x02:
		type=EQ_ARMOR;
		break;
	case 0x03:
		type=EQ_SHIELD;
		break;
	case 0x04:
		type=EQ_HELM;
		break;
	case 0x07:
		type=EQ_LEFT;
		break;
	case 0x08:
		type=EQ_RIGHT;
		break;
/*	?? Sub-Acc ??
	case 0x14:
	case 0x15:	*/
	default:
		return 0;
	}
//	printf("unequip amount: %d\n",sd->status.equip[type].amount);
	pc_unequip(sd, type);
	return 0;
}

int clif_parseattack(struct map_sessiondata *sd) {
	int damage;
	int id;
	id=sd->status.equip[EQ_WEAP].id;
	if(sd->status.state==1 || sd->status.state==3) return 0;
	//if(itemdb_look(id)>19999 && itemdb_look(id)<30000) { //Bows
	//	clif_sendaction(sd,1,0x4E,itemdb_sound(id));
	//} else { //Everything else
		clif_sendaction(sd, 1, 14, itemdb_sound(sd->status.equip[EQ_WEAP].id));
	//}
		
	damage=pc_calcdamage(sd,'s');
	switch(sd->status.side){
	
	case 0: //UP
        clif_handle_bfp(sd,0, damage);

		break;
	case 1: //RIGHT
		clif_handle_bfp(sd,1, damage);

		break;
	case 2: //DOWN
		clif_handle_bfp(sd,2, damage);

		break;
	
	case 3: //LEFT
		clif_handle_bfp(sd,3, damage);
		break;
	default:
		break;
	}
	sd->tooclose=0;	
	if(sd->backstab==1) { clif_handle_bfp(sd,(sd->status.side+2)%4,damage); sd->tooclose=0; }
	if(sd->flank==1) { clif_handle_bfp(sd,(sd->status.side+1)%4,damage); sd->tooclose=0; clif_handle_bfp(sd,(sd->status.side+3)%4,damage);}
	
	clif_sendsound(sd, itemdb_sound(sd->status.equip[EQ_WEAP].id));
	sd->tooclose=0;
	return 0;
}

int clif_parsechangepos(struct map_sessiondata *sd) {
	if (!RFIFOB(sd->fd, 5))
		pc_changeitem(sd, RFIFOB(sd->fd, 6)-1, RFIFOB(sd->fd, 7)-1);
	return 0;
}

int clif_accept(int fd) {
	int i;
	for(i=0;i<AUTH_FIFO_SIZE;i++) {
		if(auth_fifo[i].ip == session[fd]->client_addr.sin_addr.s_addr) {
			intif_load(fd, auth_fifo[i].id, auth_fifo[i].name);
			auth_fifo[i].ip = 0;
			auth_fifo[i].id = 0;
			return 0;
		}
	}

	session[fd]->eof = 1;
	return 0;
}

int clif_parse(int fd) {
	unsigned short len;
	struct map_sessiondata *sd;
	struct npc_data *nd;
	struct block_list *bl;
    sd = session[fd]->session_data;

	if(session[fd]->eof) {
		if (sd) {
			map_deliddb(&sd->bl);
			pc_stoptimer(sd);
			map_deluserlist(sd);
			intif_save(sd);
			intif_quit(sd);
			clif_quit(sd);
			
		}
		do_close(fd);
		return 0;
	}

	nullpo_ret(sd);

	if (RFIFOB(fd, 0) != 0xAA) {
		int head_err=0;
		while (RFIFOB(fd, 0) != 0xAA && RFIFOREST(fd)) {
			RFIFOSKIP(fd, 1);
			if (head_err++ > 30) {
				session[fd]->eof=1;
				return 0;
			}
		}
	}

	if (RFIFOREST(fd) < 3)
		return 0;

	len = SWAP16(RFIFOW(fd, 1)) + 3;
//	if (RFIFOREST(fd) < len ) // We don't need this, it will be checked in the while condition [Lene]
//		return 0; 

	while( RFIFOB(fd,0) == 0xAA // processing parse if there are remained packets [Lene]
		&& RFIFOREST(fd) >= len ) {	// when it comes completly [Lene]

		len=crypt(RFIFOP(fd, 0));
//		printf("%d: %x\n",fd,RFIFOB(fd,3));
		switch (RFIFOB(fd, 3)) {
		case 0x05:
			clif_parsemap(sd);
			break;
		case 0x06:
		//	printf("0x06 request\n");
			clif_parsewalk(sd);
			break;
		case 0x07:
			clif_parsegetitem(sd);
			break;
		case 0x08:
			clif_parsedropitem(sd);
			break;
		case 0x0B:
			if(sd) {	
				map_deliddb(&sd->bl);
				pc_stoptimer(sd);
				map_deluserlist(sd);
				intif_save(sd);
				intif_quit(sd);
				clif_quit(sd);
			}
			do_close(fd);
			return 0;
		case 0x0C: // < missing object/char/monster
			clif_getchararea(sd);
			//clif_charspecific(sd->bl.id,SWAP32(RFIFOL(sd->fd,5)));
			//clif_charspecific(SWAP32(RFIFOL(sd->fd,5)),sd->bl.id);
			break;
		case 0x0E:
			clif_parsesay(sd);
			break;
		case 0x0F: //magic
			sd->time=gettick();
			clif_parsemagic(sd);			
			break;
		case 0x10:
			if(session[fd]->session_data) {
				struct map_sessiondata *sd=session[fd]->session_data;
			//	printf("̸: %s\n",sd->status.equip[EQ_WEAP].real_name);
				pc_itemname_check(sd);
				pc_setpos(sd,sd->status.last_pos.m,sd->status.last_pos.x,sd->status.last_pos.y);
				pc_chkinvalid_pos(sd);
				pc_calcstat(sd);
				pc_starttimer(sd);
				clif_sendack(sd);
				clif_sendtime(sd);
				clif_sendid(sd);
				clif_sendmapinfo(sd);
				clif_sendxy(sd);
				clif_sendstatus(sd);
				clif_spawn(sd);
				clif_getchararea(sd);
				clif_getitemarea(sd);
				clif_mystaytus(sd);

				map_adduserlist(sd);
				pc_loaditem(sd);
				pc_loadequip(sd);
				pc_loadmagic(sd);
				map_addiddb(&sd->bl);
				pc_magic_startup(sd);
				clif_sendoption(sd);
				clif_sendminitext(sd, map_msg[MAP_WELCOME].message,map_msg[MAP_WELCOME].len);
				
				
				// don't send anything before sending ack [Lene]
				//		clif_sendminitext(sd, "Hyul testing server... develop version 1",sizeof("Hyul testing server... develop version 1")-1);
			} else {
				return 0;
			}
			break;
		case 0x11:
			clif_parseside(sd);
			break;
		case 0x13:
			clif_parseattack(sd);
			break;
			
			// Currently it doesn't work in the Nexus [Lene]
			// We have to figure out that packet elements [Lene]
		case 0x18:
			clif_user_list(sd);
			break;
			
			
		case 0x19:
			clif_parsewisp(sd);
			break;

		case 0x1B:
		//		{int i; for(i=0;i<20;i++) { printf("%x ",RFIFOB(fd, i)); } printf("\n"); }
				// [6] 1: Ӹɼ
				// [6] 3:  ɼ - 9
				// [6] 4:  ɼ - 8
				// [6] 5: ȿ - 7
				// [6] 6: ȿ - 6
				// [6] 7: ̵
				// [6] 9: ̵ - 9
				// [6] D: ȿ - 10
				/*WFIFOB(fd, 0) = 0xAA;
				WFIFOW(fd, 1) = SWAP16(9);
				WFIFOB(fd, 3) = 0x23;
				WFIFOB(fd, 4) = 3;
				WFIFOB(fd, 5) = 1; 
				WFIFOB(fd, 6) = 1;
				WFIFOB(fd, 7) = 1;
				WFIFOB(fd, 8) = 1;
				WFIFOB(fd, 9) = 1;
				WFIFOB(fd, 10) = 1;
				WFIFOB(fd, 11) = 1;
				WFIFOSET(fd, crypt(WFIFOP(fd,0)) );*/
			clif_parsetoggle(sd);

			break;
		case 0x1C:
			clif_parseuseitem(sd);
			break;
		case 0x1D:
			clif_parseemotion(sd);
			break;
		case 0x1F:
			clif_parseunequip(sd);
			break;

		case 0x24: // request drop money
			clif_parsedropmoney(sd);
			break;
		case 0x2D:
			if(RFIFOB(fd,5) == 0) {
				clif_mystaytus(sd);
			} else {
				group_sendmember(sd);
			}
			break;
		case 0x2E: // 
			clif_parsegroup(sd);
			break;
		
		case 0x30:
			clif_parsechangepos(sd);
			break;
		case 0x32:
		//	printf("0x32 request\n");
			
			clif_parsewalk(sd);
			break;
		case 0x38:
			
			clif_parserefresh(sd);
			break;
		case 0x39: //menu & input
			

			switch(RFIFOB(sd->fd,5)){
			case 0: //menu
				
				break;
			case 1:
				clif_parsemenu(sd);
				break;
				
			case 3: //Input
				clif_parseinput(sd);

			case 4: //BuyWindow //ۼù ó 
				clif_parsebuy(sd);
				break;
			case 5: //SellWindow ۼù ó 
				clif_parsesell(sd);
				break;
			case 8: //Sellparse sub2 ȱ /ƴϿ ó
				clif_parsesell_sub2(sd);
				break;
			case 9: //Buyparse sub1 Է¹ ó 
				clif_parsebuy_sub1(sd);
				break;
			case 10: // Sellparse sub1 Է¹ ó
				clif_parsesell_sub1(sd);
				break;
			}
			break;
		case 0x3A:

			switch(RFIFOB(sd->fd,5)){
			case 2:
				clif_parsemenu2(sd);
				break;
			default:
				clif_parsenpcdialog(sd);
			}
			break;
		case 0x42:
			break;
		case 0x43:
			//clif_sendtest(sd);
			bl=map_id2bl(SWAP32(RFIFOL(sd->fd,6)));
	//		printf("bl:%d,",bl);
			if(bl) {
				switch(bl->type) {
				case BL_PC: 
					clif_senduserinfo(sd,bl);
					
					break;
				case BL_NPC:
					nd=(struct npc_data *)bl;
	//				printf("script_mem: %d\n",nd->script);
					run_script(nd->u.scr.script,0,sd->bl.id,nd->bl.id);
					
					break;
				}
			}
			break;
		case 0x46: // ǰ..?
			break;
		case 0x4a: // trade stuff
			clif_parsetrade(sd);
			break;
		case 0x69:
			break;
		default:
			if (dump_save)
				add_dmp(fd, len);
			break;
			
				
		}
		RFIFOSKIP(fd, len);
	}
	
	return 0;
}
int clif_walkback(struct map_sessiondata *sd) {
	struct block_list *bl=map_firstincell2(sd->bl.m,sd->bl.x,sd->bl.y,0);
	if(bl==NULL) return 0;
	switch(sd->status.side) {
	case 0:
		sd->bl.y++;
		break;
	case 1:
		sd->bl.x--;
		break;
	case 2:
		sd->bl.y--;
		break;
	case 3:
		sd->bl.x++;
		break;
	}
	return 1;
}
int clif_sendtest(struct map_sessiondata *sd) {
	
	static int number;
	WFIFOB(sd->fd,0)=0xAA;
	WFIFOB(sd->fd,1)=0x00;
	WFIFOB(sd->fd,2)=0x04;
	WFIFOB(sd->fd,3)=0x63;
	WFIFOB(sd->fd,4)=0x03;
	WFIFOB(sd->fd,5)=number;
	WFIFOB(sd->fd,6)=0;
	crypt(WFIFOP(sd->fd,0));
	WFIFOSET(sd->fd,7);
	number++;
	return 0;
}
int clif_parsemenu(struct map_sessiondata *sd) {
    int selection;
    int id;
    int i;
	id=SWAP32(RFIFOL(sd->fd,6));

    selection=SWAP16(RFIFOW(sd->fd,10));
    sd->npc_menu=selection;
    npc_scriptcont(sd,id,sd->npc_pos);
    return 0;
}
int clif_parseinput(struct map_sessiondata *sd) {
    int id;
	id=SWAP32(RFIFOL(sd->fd,6));
	memset(sd->npc_str,0,256);
    memcpy(sd->npc_str,RFIFOP(sd->fd, 18),RFIFOB(sd->fd, 17));
    npc_scriptcont(sd,id,sd->npc_pos);
    return 0;
}
int clif_parsebuy(struct map_sessiondata *sd) {
    int id, itemid, shopid, shopcount = -1;
	struct item it;
	struct npc_data *nd;
	struct block_list *bl;
	char buyname[256]={0,};	
	
	id=SWAP32(RFIFOL(sd->fd,6));
	memcpy(buyname, RFIFOP(sd->fd, 13), RFIFOB(sd->fd,12));
	itemid = itemdb_id(buyname);
	if(!itemid) { return 0; }
	shopid=SWAP16(RFIFOW(sd->fd,10));
	shopcount = search_shopidcount(shopid, itemid); 
	if(shopcount < 0) { return 0; }
	if(itemdb_maxhave(itemid) == 0) {
		sd->selectshopid = shopid;
		sd->selectshopcount = shopcount;
		sd->targetitemid = itemid;
		clif_userinput(sd, id, 9, "  ʿϽŰ?");
		return 0;
	}
	if (pc_isinvenspace(sd, it.id)>=26) {
		clif_scriptmenu(sd, id, "ǰ  áϴ", NULL, 0);
		return 0;
	}
	if (!pc_delmoney(sd, npcshop_buyprice(shopid,shopcount))) {
		clif_scriptmenu(sd, id, " Ͻó׿", NULL, 0);
		return 0;
	}
	pc_giveitem(sd, itemid, 1);

	bl=map_id2bl(id);
	nd=(struct npc_data *)bl;
	run_script(nd->u.scr.script,0,sd->bl.id,nd->bl.id);
	return 0;
}
int clif_parsebuy_sub1(struct map_sessiondata *sd) {
	int id, itemid, shopid, totalamount, realamount,shopcount;
	unsigned int totalpay;
	struct npc_data *nd;
	struct block_list *bl;
	char str[256]={0,};	
	id=SWAP32(RFIFOL(sd->fd,6));
    memcpy(str,RFIFOP(sd->fd, 18),RFIFOB(sd->fd, 17));
	
	if(RFIFOB(sd->fd, 17)>5) {
		clif_userinput(sd, id, 9, " ڷθ ֽð, ̹  ŰͰ Ͽ  ִġ Ѱܼ ȵ˴ϴ");
		return 0;	
	}
	if(!allnum_check(str)) {
		clif_userinput(sd, id, 9, " ڷθ ֽð, ̹  ŰͰ Ͽ  ִġ Ѱܼ ȵ˴ϴ.");
		return 0;	
	}
	
	shopid = sd->selectshopid;
	shopcount = sd->selectshopcount;

	itemid = sd->targetitemid;
	
	realamount = atoi(str);
	totalamount = realamount+pc_itemcount(sd, itemid);
	
	if(totalamount > itemdb_maxamount(itemid)) {
		clif_userinput(sd, id, 9, " ڷθ ֽð, ̹  ŰͰ Ͽ  ִġ Ѱܼ ȵ˴ϴ...");
		return 0;	
	}
	totalpay = realamount * (npcshop_buyprice(shopid,shopcount));
	if (sd->status.money < totalpay) {
		clif_scriptmenu(sd, id, " Ͻó׿", NULL, 0);
		return 0;
	}

	if(pc_giveitem(sd, itemid, realamount)) {
		clif_scriptmenu(sd, id, "ǰ  áϴ", NULL, 0);
		return 0;
	}
	pc_delmoney(sd,totalpay);
	bl=map_id2bl(id);
	nd=(struct npc_data *)bl;
	run_script(nd->u.scr.script,0,sd->bl.id,nd->bl.id);
}
int clif_parsesell(struct map_sessiondata *sd) {
    int id, itemid, shopid, inventorynum,shopcount = -1;
	char *buf[5];
	char msg[256]={0,};
	id=SWAP32(RFIFOL(sd->fd,6));

	inventorynum = RFIFOB(sd->fd,12) - 1;
	itemid = sd->status.inventory[inventorynum].id;
	if(!itemid) { return 0; }
	shopid=SWAP16(RFIFOW(sd->fd,10));

	shopcount = search_shopidcount(shopid, itemid); 
	if(shopcount < 0) { return 0; }
	
	sd->selectshopid = shopid;
	sd->selectshopcount = shopcount;
	sd->targetslot = inventorynum;
	sd->targetitemid = itemid;
	sd->itemamount = 1;

	if(itemdb_maxhave(itemid) == 0) {
		clif_userinput(sd, id, 10, "  Ľǰǰ?");
		return 0;
	}

	sprintf(msg, "%d 帱 ֽϴ. ĽǷ?",npcshop_sellprice(shopid,shopcount));
	buf[1] = "";
	buf[2] = "ƴϿ";
	clif_usermenu(sd, id, msg, buf, 2);
    return 0;
}
int clif_parsesell_sub1(struct map_sessiondata *sd) {
    int id, itemid, shopid, myamount, amount ,shopcount = -1;
	char *buf[5];
	char str[256]={0,};
	char msg[256]={0,};
	id=SWAP32(RFIFOL(sd->fd,6));
	itemid = 0;
	memcpy(str,RFIFOP(sd->fd, 18),RFIFOB(sd->fd, 17));
	
	if(RFIFOB(sd->fd, 17)>5) {
		clif_userinput(sd, id, 10, " ڷθ ֽð, ̹   ں Ŭ ϴ");
		return 0;	
	}
	if(!allnum_check(str)) {
		clif_userinput(sd, id, 10, " ڷθ ֽð, ̹   ں Ŭ ϴ.");
		return 0;	
	}
	itemid = sd->targetitemid;
	shopid = sd->selectshopid;
	shopcount = sd->selectshopcount;
	myamount = sd->status.inventory[sd->targetslot].amount;
	amount = atoi(str);

	if(amount > myamount) {
		clif_userinput(sd, id, 10, " ڷθ ֽð, ̹   ں Ŭ ϴ..");
		return 0;
	}
	sd->itemamount = amount;

	sprintf(msg, "%d  %d 帱 ֽϴ. ĽǷ?",amount,(npcshop_sellprice(shopid,shopcount))*amount );
	buf[1] = "";
	buf[2] = "ƴϿ";
	clif_usermenu(sd, id, msg, buf, 2);
}
int clif_parsesell_sub2(struct map_sessiondata *sd) {
    int id, shopid, shopcount = -1;
	int selection;
	struct npc_data *nd;
	struct block_list *bl;
	unsigned int totalpay;
	char msg[256]={0,};
	id=SWAP32(RFIFOL(sd->fd,6));
    selection=SWAP16(RFIFOW(sd->fd,10));

	shopid = sd->selectshopid;
	shopcount = sd->selectshopcount;
	
	if(npc_shop[shopid] == NULL) { 
		return 0;
	}

	if(selection != 1) {
		return 0;
	}
	totalpay = (npcshop_sellprice(shopid,shopcount))*sd->itemamount;

	if(!pc_getmoney(sd, totalpay)) {
		clif_scriptmenu(sd, id, " ʹ   ó׿", NULL, 0);
		return 0;
	}
	pc_delitem(sd, sd->targetslot, sd->itemamount, 3);

	bl=map_id2bl(id);
	nd=(struct npc_data *)bl;
	run_script(nd->u.scr.script,0,sd->bl.id,nd->bl.id);
    return 0;
}
int clif_parsedropmoney(struct map_sessiondata *sd) {
	int money=SWAP32(RFIFOL(sd->fd,5));
//	printf("%d\n",money);
	if(money > MAX_MONEY) {
		// Err or...?
		return 0;
	}
	clif_sendaction(sd, 5, 20, 0);
	pc_dropmoney(sd,money);
}

int clif_parsetoggle(struct map_sessiondata *sd) {
	int fd=sd->fd;
	int msg;
//	printf("Parse toggle: %d %d %d %d\n", RFIFOB(fd, 5), RFIFOB(fd, 6), RFIFOB(fd, 7), RFIFOB(fd, 8))
	switch(RFIFOB(fd, 6)) {;
		// [6] 1: Ӹɼ
				// [6] 3:  ɼ - 9
				// [6] 4:  ɼ - 8
				// [6] 5: ȿ - 7
				// [6] 6: ȿ - 6
				// [6] 7: ̵
				// [6] 9: ̵ - 9
				// [6] D: ȿ - 10
	case 0: // get riding horse?
		break;
	case 1: // allow whisper 
		sd->status.opt.al_whisp=(sd->status.opt.al_whisp)? 0:1;
		msg=(sd->status.opt.al_whisp)? MAP_WHISPER_ON:MAP_WHISPER_OFF;
		clif_sendminitext(sd, map_msg[msg].message, map_msg[msg].len);
		break;
	case 2: // group
		sd->status.opt.al_group=(sd->status.opt.al_group)? 0:1;
		msg=(sd->status.opt.al_group)? MAP_GROUP_ON:MAP_GROUP_OFF;
		if(sd->status.opt.al_group == 0) { if(sd->group_info) { group_delmember(sd->group_info, sd); } }
		clif_sendminitext(sd, map_msg[msg].message, map_msg[msg].len);
		break;
	case 8: // trade
		sd->status.opt.al_exchg=(sd->status.opt.al_exchg)? 0:1;
		msg=(sd->status.opt.al_exchg)? MAP_EX_ON:MAP_EX_OFF;
		clif_sendminitext(sd, map_msg[msg].message, map_msg[msg].len);
		break;
	case 11: // pvp
		sd->status.opt.al_pvp=(sd->status.opt.al_pvp)? 0:1;
		msg=(sd->status.opt.al_pvp)? MAP_PVP_ON:MAP_PVP_OFF;
		clif_sendminitext(sd, map_msg[msg].message, map_msg[msg].len);
		break;
	case 13: // sound effect
		sd->status.opt.eff_sound=(sd->status.opt.eff_sound)? 0:1;
		break;
	case 14: // show helm
		sd->status.opt.showhelm=(sd->status.opt.showhelm)? 0:1;
		clif_sendchararea2(sd);
		break;
	}
	clif_sendoption(sd);
}
int clif_senduserinfo(struct map_sessiondata *sd, struct map_sessiondata *src) {
	char buf[512]={0,};
	int len;
	int fd=sd->fd;
	int tmp;
	int i;
	int templen = 0;
	int legendcount = 0;
	WFIFOB(fd, 0) = 0xAA;
	WFIFOB(fd, 3) = 0x34;
	WFIFOB(fd, 4) = 0;
	// title
	WFIFOB(fd, 5)= len =strlen(src->status.title);
	memcpy(WFIFOP(fd, 6),src->status.title,len);
	
	WFIFOB(fd,6+len) = 18;
	memcpy(WFIFOP(fd,7+len),"߼ӽùĻ",18);
	len+=18;
	WFIFOB(fd,7+len)=tmp=strlen(src->status.clan_title);
	memcpy(WFIFOP(fd,8+len),src->status.clan_title,tmp);
	len+=tmp;
	WFIFOB(fd,8+len)=tmp=class[src->status.class][src->status.subclass].class_name_len;
	memcpy(WFIFOP(fd,9+len),class[src->status.class][src->status.subclass].class_name,tmp);
	len+=tmp;
	WFIFOB(fd,9+len)=tmp=strlen(src->status.name);
	memcpy(WFIFOP(fd,10+len),src->status.name,tmp);
	len+=tmp;
	WFIFOW(fd,len+10)=SWAP16(src->status.sex);
	WFIFOB(fd,len+12)=src->status.state;
	WFIFOB(fd,len+13)=0;
	WFIFOB(fd,len+14)=2;
	WFIFOB(fd,len+15)=78;
	WFIFOB(fd,len+16)=0;
	WFIFOB(fd,len+17)=src->status.face&0xFF;
	WFIFOB(fd,len+18)=src->status.hair&0xFF;
	WFIFOB(fd,len+19)=src->status.hair_color&0xFF;

	if (!pc_isequip(src, EQ_ARMOR)) {		
			WFIFOB(fd, len+20) = 0x00;
			WFIFOB(fd, len+21) = 0x00;
	} else {
		WFIFOB(fd, len+20) = itemdb_look(pc_isequip(src, EQ_ARMOR));
		WFIFOB(fd, len+21)=	sd->status.armor_color;
	}

	if (!pc_isequip(src, EQ_WEAP)) {		
			WFIFOW(fd, len+22) = 0xFFFF;
			WFIFOB(fd, len+24) = 0x00;
	} else {
		WFIFOW(fd, len+22) = SWAP16(itemdb_look(pc_isequip(src, EQ_WEAP)));
		WFIFOB(fd, len+24)=	 itemdb_lookcolor(pc_isequip(src, EQ_WEAP));
	}

	if (!pc_isequip(src, EQ_SHIELD)) {		
			WFIFOW(fd, len+25) = 0xFFFF;
			WFIFOB(fd, len+27) = 0x00;
	} else {
		WFIFOW(fd, len+25) = SWAP16(itemdb_look(pc_isequip(src, EQ_SHIELD)));
		WFIFOB(fd, len+27)=	itemdb_lookcolor(pc_isequip(src, EQ_SHIELD));
	}

	if (!pc_isequip(src, EQ_HELM)) {		
			WFIFOB(fd, len+28) = 0x00;
			WFIFOB(fd, len+29) = 0x00;
			WFIFOB(fd, len+30) = 0x00;
	} else {
		WFIFOB(fd, len+28) = 0x01;
		WFIFOB(fd, len+29) = itemdb_look(pc_isequip(src, EQ_HELM));
		WFIFOB(fd, len+30)=	itemdb_lookcolor(pc_isequip(src, EQ_HELM));
	}
	
	if (!pc_isequip(src, EQ_FACEACC)) {		
			WFIFOW(fd, len+31) = 0xFFFF;
			WFIFOB(fd, len+33) = 0x00;
	} else {
		WFIFOW(fd, len+31) = SWAP16(itemdb_look(pc_isequip(src, EQ_FACEACC)));
		WFIFOB(fd, len+33)=	itemdb_lookcolor(pc_isequip(src, EQ_FACEACC));
	}
	
	for(i=0;i<9;i++) {
		WFIFOW(fd, len+34) = SWAP16(itemdb_icon(src->status.equip[i].id));
		WFIFOB(fd, len+36) = itemdb_iconcolor(src->status.equip[i].id);
		WFIFOB(fd, len+37)=0x00;
		WFIFOB(fd, len+38)=0x00;
		WFIFOB(fd, len+39)=0x00;
		WFIFOB(fd, len+40)=0x00;
		WFIFOB(fd, len+41)=0x00;
		WFIFOB(fd, len+42)=0x00;
		len+=9;
	}
	// copy the item names
	tmp=sprintf(buf,item_infos,
		src->status.equip[EQ_WEAP].real_name,
		src->status.equip[EQ_ARMOR].real_name,
		src->status.equip[EQ_SHIELD].real_name,
		src->status.equip[EQ_HELM].real_name,
		src->status.equip[EQ_LEFT].real_name,
		src->status.equip[EQ_RIGHT].real_name,
		src->status.equip[EQ_SCRIPT1].real_name,
		src->status.equip[EQ_SCRIPT2].real_name);
	


	WFIFOB(fd, len+34) = tmp;
	memcpy(WFIFOP(fd, len+35), buf, tmp);
	len+=tmp;
	WFIFOL(fd, len+35) = SWAP32(src->status.id); //ε1

	WFIFOB(fd, len+39) = src->status.opt.al_group; //׷
	WFIFOB(fd, len+40) = src->status.opt.al_exchg; //ȯ
	WFIFOB(fd, len+41) = 1; //?
	WFIFOB(fd, len+42) = 100>>8; //ǰ
	WFIFOB(fd, len+43) = 100%256;//ǰ
	WFIFOB(fd, len+44) = 0; //?
	WFIFOB(fd, len+45) = 0; //?
	WFIFOB(fd, len+46) = 1; //ٲٸ鷹
	WFIFOB(fd, len+47) = 2; //ٲٸ鷹

	
	for(i=0; i<MAX_LEGEND; i++) {
		if(sd->legend_list.legends[i].len > 0) {
			WFIFOB(fd, len+49+templen) = sd->legend_list.legends[i].icon; //ܹȣ
			WFIFOB(fd, len+50+templen) = sd->legend_list.legends[i].color; //۾Į
			WFIFOB(fd, len+51+templen) = sd->legend_list.legends[i].len; //۾Ʈ 
			memcpy(WFIFOP(fd, len+52+templen),sd->legend_list.legends[i].text, sd->legend_list.legends[i].len);
			templen += 3+sd->legend_list.legends[i].len;
			legendcount++;
		}
	}
	WFIFOB(fd, len+48) = legendcount; // 
	
	WFIFOW(fd,1)= SWAP16(len+52+templen);

	WFIFOSET(fd, crypt(WFIFOP(fd,0)));

}

int clif_sendoption(struct map_sessiondata *sd) {
	int fd=sd->fd;
	WFIFOB(fd, 0) = 0xAA;
	WFIFOW(fd, 1) = SWAP16(8);
	WFIFOB(fd, 3) = 0x23;
	WFIFOB(fd, 4) = 0;
	WFIFOB(fd, 5) = sd->status.opt.eff_weather; // weather effect
	WFIFOB(fd, 6) = sd->status.opt.eff_magic; // magic effect
	WFIFOB(fd, 7) = sd->status.opt.announce; // listening announcement or not
	WFIFOB(fd, 8) = sd->status.opt.fastmove; // fast movement
	WFIFOB(fd, 9) = sd->status.opt.eff_sound; // sound effect
	WFIFOB(fd, 10) = sd->status.opt.showhelm;
	WFIFOSET(fd, crypt(WFIFOP(fd, 0)));
}