
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../common/core.h"
#include "../common/socket.h"
#include "../common/timer.h"
#include "../common/mmo.h"
#include "../common/malloc.h"
#include "../common/mmo.h"
#include "../common/nullpo.h"
#include "map.h"
#include "intif.h"
#include "pc.h"
#include "clif.h"

int intif_parse_registry(int fd);
int intif_parse_legend(int fd);

static int intif_table_recv[0xFF] = 
{
	2, 0, 37, 0, 5, 0, 0, 0, 0, 0, // 0-9
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 10-19
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 20-29
};


int CheckForCharServer(void) {
	return ((char_fd <= 0) || session[char_fd] == NULL || session[char_fd]->wdata == NULL);
}

int check_connect_char() {
	if (char_fd <= 0 || session[char_fd] == NULL) {
		printf("Attempt to connect to char-server...%d\n",char_port);
		char_fd = make_connection(char_ip, char_port);
		if(char_fd > 0) {
			session[char_fd]->func_parse = intif_parse;
			realloc_fifo(char_fd, FIFOSIZE_SERVER, FIFOSIZE_SERVER);
			WFIFOB(char_fd,0) = 0x00;
			memcpy(WFIFOP(char_fd,1), char_id, 32);
			memcpy(WFIFOP(char_fd,33), char_pw, 32);
			WFIFOL(char_fd, 65) = map_ip;
			WFIFOW(char_fd, 69) = map_port;
			WFIFOSET(char_fd, 71);
		} else {
			printf("Cannot connect to char-server...\n");
		}
	}
	return 0;
}

int intif_quit(struct map_sessiondata *sd) {
	if (char_fd <= 0)
		return -1;

	WFIFOB(char_fd, 0) = 0x05;
	WFIFOL(char_fd, 1) = sd->status.id;
	WFIFOSET(char_fd, 5);
	return 0;
}

int intif_load(int fd, int id, char *name) {
	if (char_fd <= 0)
		return -1;

	WFIFOB(char_fd, 0) = 0x03;
	WFIFOW(char_fd, 1) = fd;
	WFIFOL(char_fd, 3) = id;
	memcpy(WFIFOP(char_fd, 7), name, 16);
	WFIFOSET(char_fd, 23);
	return 0;
}

int intif_save(struct map_sessiondata *sd) {
	if (char_fd <= 0)
		return -1;

	sd->status.last_pos.m = sd->bl.m;
	sd->status.last_pos.x = sd->bl.x;
	sd->status.last_pos.y = sd->bl.y;
	WFIFOB(char_fd, 0) = 0x04;
	memcpy(WFIFOP(char_fd, 1), &sd->status, sizeof(struct mmo_charstatus));
	WFIFOSET(char_fd, sizeof(struct mmo_charstatus)+1);
	if(sd->state.leg_dirty) {
		intif_savelegend(sd);
	}
	
	return 0;
}

int intif_mmo_tosd(int fd, struct mmo_charstatus* p) {
	struct map_sessiondata *sd;
	int n;
	if (!p->id) {
		session[fd]->eof=1;
		return 0;
	}	
	CREATE(sd, struct map_sessiondata, 1);
	memcpy(&sd->status, p, sizeof(struct mmo_charstatus));
	intif_request_registry(sd);
	intif_request_legend(sd);

	sd->fd = fd;
	session[fd]->session_data = sd;
	sd->bl.id = sd->status.id;
	sd->bl.prev = sd->bl.next = NULL;
	
}

int intif_parse(int fd) {
	int i, j = 0;
	struct map_sessiondata *sd;
	if(session[fd]->eof) {
		add_log("Can't connect to Char Server.\n");
		printf("Can't connect to Char Server.\n");
		char_fd = 0;
		do_close(fd);
		return 0;
	}

	if(RFIFOREST(fd) < intif_table_recv[RFIFOB(fd, 0)])
		return 0;

	

	switch (RFIFOB(fd,0)) {
	case 0x00:
		if (RFIFOB(fd,1)) {
			printf("CFG_ERR: Username or password to connect Login Server is Invalid!\n");
			add_log("CFG_ERR: Username or password to connect Login Server is Invalid!\n");
			break;
		}
		add_log("Connected to Char Server.\n");
		printf("Connected to Char Server.\n");

		WFIFOB(fd, 0) = 0x01;
		WFIFOW(fd, 1) = map_n;
		for(i=0;i<MAX_MAP_PER_SERVER;i++) {
			if(map[i].tile != NULL) {
				WFIFOW(fd, j*2+3) = i;
				j++;
			}
			if(j>=map_n)
				break;
		}
		WFIFOSET(fd, map_n*2+3);
		break;
	case 0x01:
		if(RFIFOREST(fd) < RFIFOW(fd, 7)*2+9)
			return 0;

		for(i=0;i<RFIFOW(fd, 7);i++)
			map_setmapip(RFIFOW(fd, i*2+9), RFIFOL(fd, 1), RFIFOW(fd, 5));
		{
			unsigned char *p;
			p=RFIFOP(fd, 1);
			printf("Received %d maps from map server %d.%d.%d.%d:%d\n", RFIFOW(fd, 7), p[0], p[1], p[2], p[3], RFIFOW(fd, 5));
		}
		RFIFOSKIP(fd, RFIFOW(fd, 7)*2+9);
		return 0;
	case 0x02:
		for(i=0;i<AUTH_FIFO_SIZE;i++) {
			if(auth_fifo[i].id == RFIFOL(fd, 3)) {
				auth_fifo[auth_n].ip = RFIFOL(fd, 33);
				break;
			}
		}
		if (i>=AUTH_FIFO_SIZE) {
			auth_fifo[auth_n].id = RFIFOL(fd, 3);
			auth_fifo[auth_n].ip = RFIFOL(fd, 33);
			memset(auth_fifo[auth_n].name, 0, 16);
			memcpy(auth_fifo[auth_n].name, RFIFOP(fd, 7), 16);
			auth_n++;
			if (auth_n >= AUTH_FIFO_SIZE)
				auth_n = 0;
		}
		WFIFOB(fd, 0) = 0x02;
		WFIFOW(fd, 1) = RFIFOW(fd, 1);
		memset(WFIFOP(fd, 3), 0, 16);
		memcpy(WFIFOP(fd, 3), RFIFOP(fd, 3), 16);
		WFIFOSET(fd, 19);
		break;
	case 0x03:
		if (!session[RFIFOW(fd, 1)])
			break;
		intif_mmo_tosd(RFIFOW(fd,1), (struct mmo_charstatus*)RFIFOP(fd, 3));
		break;
	case 0x04:
		sd = map_id2sd(RFIFOL(fd, 1));
		if (sd) {
			session[sd->fd]->eof = 1;
		} else {
			WFIFOB(fd, 0) = 0x05;
			WFIFOL(fd, 1) = RFIFOL(fd, 1);
			WFIFOSET(fd, 5);
		}
		break;
	case 0x06:
		{
			
			if (RFIFOREST(fd) < RFIFOW(fd, 34)+36)
				return 0;
			
			{
				struct map_sessiondata *sd;
				if (RFIFOB(fd, 1) == 0x00) {
					sd = map_name2sd(RFIFOP(fd, 18));
					if (sd)
						clif_sendwisp(sd, RFIFOP(fd, 2), RFIFOP(fd, 36), RFIFOW(fd, 36));
				} else if (RFIFOB(fd, 1) == 0x01) {
					sd = map_name2sd(RFIFOP(fd, 2));
					if (sd)
						clif_retrwisp(sd, RFIFOP(fd, 18), RFIFOP(fd, 36), RFIFOW(fd, 36));
				} else {
					sd = map_name2sd(RFIFOP(fd, 2));
					if (sd) {
						char dest_name[16];
						memcpy(dest_name,RFIFOP(fd,18),16);
						clif_failwisp(sd,dest_name);
					}
				}
			}
			RFIFOSKIP(fd, RFIFOW(fd, 34)+36);
		}
		return 0;

	case 21: // contains reg data 
		intif_parse_registry(fd);
		return 0;
	case 22: // constains legend data
		intif_parse_legend(fd);
		return 0;
	default:
		add_dmp(fd, RFIFOREST(fd));
		RFIFOSKIP(fd, RFIFOREST(fd));
		return 0;
	}
	RFIFOSKIP(fd, intif_table_recv[RFIFOB(fd, 0)]);
	return 0;
}

int intif_wisp(char *srcname, char *dstname, unsigned char *msg, int len) {
	if (char_fd <= 0)
		return -1;

	WFIFOB(char_fd, 0) = 0x06;
	memcpy(WFIFOP(char_fd, 1), srcname, 16);
	memcpy(WFIFOP(char_fd, 17), dstname, 16);
	WFIFOW(char_fd, 33) = len;
	memcpy(WFIFOP(char_fd, 35), msg, len);
	WFIFOSET(char_fd, len+35);
	return 0;
}

int intif_timer(int data1, int data2) {
	int i;
	struct map_sessiondata *sd;
	for(i=0;i<fd_max;i++){
		if(session[i] && (sd=session[i]->session_data)) {
			intif_save(sd);
		}
	}
	return 0;
}

int intif_init() {
	add_timer(gettick()+save_time,intif_timer,save_time,0,save_time);
	intif_table_recv[3] = sizeof(struct mmo_charstatus)+3;
	return 0;
}


int intif_regtostr(char* str, struct global_reg *reg, int qty) {
	int len =0, i;

	for (i = 0; i < qty; i++) {
		len+= sprintf(str+len, "%s", reg[i].str)+1; //We add 1 to consider the '\0' in place.
		len+= sprintf(str+len, "%s", reg[i].value)+1;
	}
	return len;
}

//Request for saving registry values.
int intif_saveregistry(struct map_sessiondata *sd)
{
	struct global_reg *reg;
	int count;
	int i,p;
	int len;
	if (CheckForCharServer())
		return -1;


	//Character reg
	reg = sd->save_reg.global;
	count = sd->save_reg.global_num;
	sd->state.reg_dirty &= ~0x4;

	WFIFOHEAD(char_fd, 288 * MAX_REG_NUM+13);
	WFIFOB(char_fd,0)=20; // - request saving registry ( û)
	WFIFOL(char_fd,1)=sd->status.id;
	WFIFOW(char_fd,5)=count;
	
	
	for (p=8,i = 0; i < count; i++) {
		if (reg[i].str[0] && reg[i].value != 0) {
			WFIFOB(char_fd, p)= len = strlen(reg[i].str);
			p++;
			memcpy(WFIFOP(char_fd, p), reg[i].str, len);
			p+=len;
			WFIFOB(char_fd, p)= len = strlen(reg[i].value);
			p++;
			memcpy(WFIFOP(char_fd, p), reg[i].value, len);
			p+=len;
		}
	}
	WFIFOW(char_fd,6)=p; // save packet size
	
	WFIFOSET(char_fd,p);
	return 0;
}

//Request the registries for this player.
int intif_request_registry(struct map_sessiondata *sd)
{
	nullpo_retr(0, sd);

	sd->save_reg.global_num = -1;

	if (CheckForCharServer())
		return 0;
	
	WFIFOB(char_fd,0) = 21; // - request registry data (û)
	WFIFOL(char_fd,1) = sd->status.id;
	WFIFOSET(char_fd,5);

	return 0;
}

int intif_parse_registry(int fd) {
	int c,i,j,len;
	struct map_sessiondata *sd=map_id2sd(RFIFOL(fd, 1));
	struct registry *reg;
	if(sd==NULL) return 0; // user doesn't found
	reg=&sd->save_reg;
	reg->global_num=RFIFOB(fd, 5);
	c=RFIFOW(fd, 6);
//	printf("size: %d\n",c);
	for(i=8,j=0; i<c && j<reg->global_num; j++) {
		len=RFIFOB(fd, i);
		i++;
		memcpy(reg->global[j].str, RFIFOP(fd, i), len);
		reg->global[j].str[len]='\0';
		i+=len;
		len=RFIFOB(fd, i);
		i++;
		memcpy(reg->global[j].value, RFIFOP(fd, i), len);
		reg->global[j].value[len]='\0';
		i+=len;
//		printf("%s[%d]'s reg data- %s:%s\n", sd->status.name, j, reg->global[j].str, reg->global[j].value);
	}
	reg->global_num=j;
	

	RFIFOSKIP(fd, c);
	return 1;
}

int intif_request_legend(struct map_sessiondata *sd) {

	sd->legend_list.legend_num = -1;
	if (CheckForCharServer())
		return 0;

	WFIFOB(char_fd, 0) = 0x22; // Request for legend data
	WFIFOL(char_fd, 1) = sd->status.id;
	WFIFOSET(char_fd, 5);

}

int intif_parse_legend(int fd) {
	int c,i,j,len;
	struct map_sessiondata *sd;
	struct legend *leg;
	sd=map_id2sd(RFIFOL(fd, 1));
	if(sd==NULL) {
		ShowError("intif_parse_legend(): map session does not found! id:%u\n",RFIFOL(fd, 1));
		return 1;
	}
	leg=&sd->legend_list;
	for(i=8,j=0;i<c && j<leg->legend_num;j++) {
		leg->legends[j].icon = RFIFOB(fd, i);i++;
		leg->legends[j].color = RFIFOB(fd, i);i++;
		leg->legends[j].len = len = RFIFOB(fd, i);i++;
		memcpy(leg->legends[j].text, RFIFOP(fd, i), len);
	}
	leg->legend_num=j;
	RFIFOSKIP(fd, j);
	return 0;
}
int intif_savelegend(struct map_sessiondata *sd) {
	int p,i,c,len;
	struct legend *leg;
	if (CheckForCharServer())
		return 0;
	leg=&sd->legend_list;
	c=leg->legend_num;
	WFIFOB(char_fd, 0) = 23; //request for saving legend
	WFIFOL(char_fd, 1) = sd->status.id;
	WFIFOB(char_fd, 5) = c;

	for(p=8,i=0;i<c;) {
		len = leg->legends[i].len;
		if(len) {			
			WFIFOB(char_fd, p) = leg->legends[i].icon; p++;
			WFIFOB(char_fd, p) = leg->legends[i].color; p++;
			WFIFOB(char_fd, p) = len; p++;
			memcpy(WFIFOP(char_fd, p), leg->legends[i].text, len);
			p+=len;
			i++;
		}
	}

	WFIFOW(char_fd, 6) = p;
	WFIFOSET(char_fd, p);
	return 0;
}