
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#ifdef _WIN32
#include <winsock.h>
#else
#include <arpa/inet.h>
#endif
#include "../common/core.h"
#include "../common/socket.h"
#include "../common/timer.h"
#include "version.h"
#include "map.h"
#include "../common/crypt.h"
#include "clif.h"
#include "intif.h"
#include "../common/mmo.h"
#include "../common/malloc.h"
#include "../common/db.h"
#include "../common/nullpo.h"
#include "class_db.h"
#include "itemdb.h"
#include "magic.h"
#include "pc.h"
#include "npc.h"
#include "script.h"
#include "mob.h"
#include "../common/utils.h"
#include "warp.h"
struct dbt *id_db;
struct dbt *class_db;

char db_path[256]="db";

struct map_msg_data map_msg[MSG_MAX];

//item id pool
char *object;
int object_n=0;

struct map_src_list{
	struct map_src_list *next;
	int id;
	char title[64];
	int bgm;
	char *mapfile;
};
struct map_src_list *map_src_first = NULL;
struct map_src_list *map_src_last = NULL;

char map_ip_s[16];
int map_ip;
unsigned short map_port = 2010;

int char_fd;
int map_fd;

char char_id[32];
char char_pw[32];
char char_ip_s[16];
int char_ip;
unsigned short char_port;
int save_time = 60000;
int dump_save = 0;
int xp_rate;
int d_rate;
char item_infos[256]="";

struct auth_data auth_fifo[AUTH_FIFO_SIZE];
int auth_n = 0;
struct userlist_data userlist;
struct map_data map[MAX_MAP_PER_SERVER];
int map_n = 0;

//block_list
#define BL_LIST_MAX 32768
#define BL_FREE_MAX 32768
struct block_list *bl_list[BL_LIST_MAX];
struct block_list *bl_free[BL_FREE_MAX];
struct block_list *bl_head;
int bl_list_count = 0;
int bl_free_count = 0;
int bl_free_lock = 0;

#define BROADCAST_MAX 256
char broadcast_msg[BROADCAST_MAX][256];
int broadcast_count=0;

// map_warp list


void map_read_autobroadcast( const char *file ) {
	FILE *fp;
	fp=fopen(file,"r");
	if(fp==NULL) return;
	
	while(fgets(broadcast_msg[broadcast_count],1023,fp)) {
		broadcast_msg[broadcast_count][strlen(broadcast_msg[broadcast_count])-1]=0;
		broadcast_count++;
	}

	fclose(fp);
}

void map_autobroadcast_freebuff() {
	// we don't have to free this. :)
}

void map_autobroadcast(int fd) {
	int rand_val;
	rand_val=rand()%broadcast_count;
	
	// progressing...
	//-----------------------------------
	//clif_sendbluemessage();

}
struct block_list* map_id2bl(unsigned int id) {
	struct block_list *bl=NULL;
    bl = idb_get(id_db,id);
	return bl;
}

void map_deliddb(unsigned int id) {
	idb_remove(id_db,id);
}

void map_addiddb(struct block_list *bl) {
	nullpo_retv(bl);
	idb_put(id_db,bl->id,bl);
}

void map_initiddb() {
	id_db = db_alloc(__FILE__,__LINE__,DB_INT,DB_OPT_BASE,sizeof(int));
}

int map_finaliddb(void *key,void *data,va_list ap) {
	struct block_list *db;
	struct npc_data *nd;
	struct mobspawn_data *md;
	nullpo_ret(db=data);

	map_delblock(db);
	switch (db->id) {
	case BL_NPC:
		nd=data;
		switch(db->subtype) {
		case FLOOR:
			script_free_code(nd->u.scr.script);
		case SCRIPT:
			break;
		}
	case BL_PC:
	case BL_ITEM:
		aFree(db);
		break;
	case BL_MOB:
		aFree(db);
	}
	return 0;
}

void map_termiddb() {
	if(id_db){
		id_db->destroy(id_db, NULL);
	}
}

void map_clritem() {
	aFree(object);
	object_n = 0;
}

void map_delitem(unsigned int id) {
	struct block_list *bl;
	bl = map_id2bl(id);
	if (!bl)
		return;

	map_deliddb(id);
	map_delblock(bl);
	aFree(bl);

	id -= FLOORITEM_START_NUM;
	if (id>=object_n && id<0)
		return;

	object[id] = 0;
}

void map_additem(struct block_list *bl) {
	unsigned int i;
	for (i=0;i<object_n;i++) {
		if (!object[i])
			break;
	}
	if (i>=MAX_FLOORITEM) {
		printf("MAP_ERR: Item reached max item capasity.\n");
		return;
	}

	if (i>=object_n) {
		if (object_n) {
			RECREATE(object, char, i+256);
		} else {
			CREATE(object, char, 256);
		}
		object_n = i+256;
	}

	object[i] = 1;
	i += FLOORITEM_START_NUM;
	bl->id = i;
	bl->type = BL_ITEM;
	bl->prev = NULL;
	bl->next = NULL;
	map_addblock(bl);
	map_addiddb(bl);
}

int map_src_clear() {
	struct map_src_list *p=map_src_first;

	while(p) {
		struct map_src_list *p2=p;
		p=p->next;
		aFree(p2->mapfile);
		aFree(p2);
	}

	map_src_first=NULL;
	map_src_last=NULL;
	return 0;
}

int map_src_add(const char *r1) {
	int map_id,map_bgm;
	struct map_src_list *new;
	char map_title[1024], map_file[1024];
	if (sscanf(r1, "%d,%[^,],%d,%[^\r\n]", &map_id, map_title, &map_bgm, map_file) < 3) {
		return -1;
	}
	
	

	CREATE(new, struct map_src_list, 1);
	CREATE(new->mapfile, char, strlen(map_file) + 1);
	new->next = NULL;

	new->id = map_id;
	new->bgm = map_bgm;
	strncpy(new->title, map_title, 16);
	strncpy(new->mapfile, map_file, strlen(map_file));

	if (map_src_first == NULL)
		map_src_first = new;
	if (map_src_last)
		map_src_last->next = new;

	map_src_last = new;
	return 0;
}

struct map_sessiondata* map_id2sd(unsigned int id) {
	int i;
	struct map_sessiondata *sd;

	for(i=0;i<fd_max;i++){
		if(session[i] && (sd=session[i]->session_data)){
			if(id == sd->status.id)
				return sd;
		}
	}
	return NULL;
}

struct map_sessiondata* map_name2sd(char *name) {
	int i;
	struct map_sessiondata *sd;

	for(i=0;i<fd_max;i++){
		if(session[i] && (sd=session[i]->session_data)){
			if(strcmpi(name, sd->status.name) == 0)
				return sd;
		}
	}
	return NULL;
}

void map_termblock() {
	aFree(bl_head);
}

void map_initblock() {
	CREATE(bl_head, struct block_list, 1);
}

int map_freeblock(void *bl) {
	if (!bl_free_lock) {
		aFree(bl);
	} else {
		if (bl_free_count >= BL_FREE_MAX) {
			printf("BL_ERR: Too many block free list! Block free count: %d\n", bl_free_count);
		} else
			bl_free[bl_free_count++] = bl;
	}
	return bl_free_lock;
}

int map_freeblock_lock() {
	return ++bl_free_lock;
}

int map_freeblock_unlock() {
	if(!--bl_free_lock) {
		int i;
		for(i=0;i<bl_free_count;i++)
			aFree(bl_free[i]);
		bl_free_count=0;
	}
	return bl_free_lock;
}

int map_isarea(struct block_list *bl1, struct block_list *bl2) {
	int mapx = map[bl1->m].xs - 1;
	int tx1=0, tx2=0;
	int myx = bl1->x;

	int mapy = map[bl1->m].ys - 1;
	int ty1=0, ty2=0;
	int myy = bl1->y;

	if(myx <= 8){ myx = 8; }
	if(mapx - myx <=8){ myx = mapx - 8; }
	
	tx1 = myx - 8; 
	if(tx1 <= 0){ tx1 = 0; }
	
	tx2 = myx + 8;
	if(tx2 >= mapx){ tx2 = mapx; } 

	if(myy <= 7){ myy = 7; }
	if(mapy - myy <=7){ myy = mapy - 7; }

	ty1 = myy - 7; 
	if(ty1 <= 0){ ty1 = 0; }
	
	ty2 = myy + 7;
	if(ty2 >= mapy){ ty2 = mapy; } 


	if(!(tx1<=bl2->x && tx2>=bl2->x)) {
		return FALSE;
	}

	if(!(ty1<=bl2->y && ty2>=bl2->y)) {
		return FALSE;
	}
	
	return TRUE;
	/*  
	 ty1
  tx1--------------------tx2
	 |                  |
	 |                  |
	 |                  |
	 |                  |
	 |                  |
	 --------------------
	 ty2
  
	*/
	  
}

int map_addblock(struct block_list *bl) {
	int m, x, y;
	nullpo_ret(bl);
	
	x = bl->x;
	y = bl->y;
	m = bl->m;
	//printf("%d, %d, %d\n",bl->m,bl->x,bl->y);
	if (bl->prev != NULL) {
		return 0;
	}

	if (x >= map[m].xs || y >= map[m].ys)
		return 0;

	if(bl->type==BL_NPC){
		bl->next = map[m].block_npc[x/BLOCK_SIZE+(y/BLOCK_SIZE)*map[m].bxs];
		bl->prev = bl_head;
		if(bl->next) bl->next->prev = bl;
		map[m].block_npc[x/BLOCK_SIZE+(y/BLOCK_SIZE)*map[m].bxs] = bl;
		map[m].block_npc_count[x/BLOCK_SIZE+(y/BLOCK_SIZE)*map[m].bxs]++;
	}  else {
		bl->next = map[m].block[x/BLOCK_SIZE+(y/BLOCK_SIZE)*map[m].bxs];
		bl->prev = bl_head;
		if(bl->next) bl->next->prev = bl;
		map[m].block[x/BLOCK_SIZE+(y/BLOCK_SIZE)*map[m].bxs] = bl;
		map[m].block_count[x/BLOCK_SIZE+(y/BLOCK_SIZE)*map[m].bxs]++;
		if(bl->type==BL_PC)
			map[m].user++;
	}

	return 0;
}

int map_delblock(struct block_list *bl) {
	nullpo_ret(bl);
	//nullpo_ret(bl->prev);
	if(bl==bl_head || bl->prev==NULL) return;
	if(bl->type==BL_PC)
		map[bl->m].user--;
	if(bl->next) bl->next->prev = bl->prev;
	if(bl->prev==bl_head){
		if(bl->type==BL_NPC){
			map[bl->m].block_npc[bl->x/BLOCK_SIZE+(bl->y/BLOCK_SIZE)*map[bl->m].bxs] = bl->next;
			if((map[bl->m].block_npc_count[bl->x/BLOCK_SIZE+(bl->y/BLOCK_SIZE)*map[bl->m].bxs]--) < 0)
				map[bl->m].block_npc_count[bl->x/BLOCK_SIZE+(bl->y/BLOCK_SIZE)*map[bl->m].bxs] = 0;
		} else {
			map[bl->m].block[bl->x/BLOCK_SIZE+(bl->y/BLOCK_SIZE)*map[bl->m].bxs] = bl->next;
			if((map[bl->m].block_count[bl->x/BLOCK_SIZE+(bl->y/BLOCK_SIZE)*map[bl->m].bxs]--) < 0)
				map[bl->m].block_count[bl->x/BLOCK_SIZE+(bl->y/BLOCK_SIZE)*map[bl->m].bxs] = 0;
		}
	} else {
		bl->prev->next = bl->next;
	}
	bl->next = NULL;
	bl->prev = NULL;
	return 0;
}

int map_foreachinarea(int (*func)(struct block_list*,va_list),int m,int x,int y,int area,int type,...) {
	va_list ap;
	va_start(ap,type);

	switch(area) {
	case AREA:
        
		map_foreachinblockva(func,m,x-(AREAX_SIZE+1),y-(AREAY_SIZE+1),x+(AREAX_SIZE+1),y+(AREAY_SIZE+1),type,ap);
	case CORNER:
		if (map[m].xs > (AREAX_SIZE*2+1) && map[m].ys > (AREAY_SIZE*2+1)) {
			if (x < (AREAX_SIZE*2+2) && x > AREAX_SIZE)
				map_foreachinblockva(func,m,0,y-(AREAY_SIZE+1),x-(AREAX_SIZE+2),y+(AREAY_SIZE+1),type,ap);
			if (y < (AREAY_SIZE*2+2) && y > AREAY_SIZE) {
				map_foreachinblockva(func,m,x-(AREAX_SIZE+1),0,x+(AREAX_SIZE+1),y-(AREAY_SIZE+2),type,ap);
				if (x < (AREAX_SIZE*2+2) && x > AREAX_SIZE)
					map_foreachinblockva(func,m,0,0,x-(AREAX_SIZE+2),y-(AREAY_SIZE+2),type,ap);
				else if (x > map[m].xs-(AREAX_SIZE*2+3) && x < map[m].xs-(AREAX_SIZE+1))
					map_foreachinblockva(func,m,x+(AREAX_SIZE+2),0,map[m].xs-1,y+(AREAY_SIZE+2),type,ap);
			}
			if (x > map[m].xs-(AREAX_SIZE*2+3) && x < map[m].xs-(AREAX_SIZE+1))
				map_foreachinblockva(func,m,x+(AREAX_SIZE+2),y-(AREAY_SIZE+1),map[m].xs-1,y+(AREAY_SIZE+1),type,ap);
			if (y > map[m].ys-(AREAY_SIZE*2+3) && y < map[m].ys-(AREAY_SIZE+1)) {
				map_foreachinblockva(func,m,x-(AREAX_SIZE+1),y+(AREAY_SIZE+2),x+(AREAX_SIZE+1),map[m].ys-1,type,ap);
				if (x < (AREAX_SIZE*2+2) && x > AREAX_SIZE)
					map_foreachinblockva(func,m,0,y+(AREAY_SIZE+2),x-(AREAX_SIZE+2),map[m].ys-1,type,ap);
				else if (x > map[m].xs-(AREAX_SIZE*2+3) && x < map[m].xs-(AREAX_SIZE+1))
					map_foreachinblockva(func,m,x+(AREAX_SIZE+2),y+(AREAY_SIZE+2),map[m].xs-1,map[m].ys-1,type,ap);
			}
		}
		break;
	case SAMEAREA:
		if (x <= AREAX_SIZE && y <= AREAY_SIZE)
			map_foreachinblockva(func,m,0,0,(AREAX_SIZE*2+2),(AREAY_SIZE*2+2),type,ap);
		else if (x <= AREAX_SIZE && y >= map[m].ys-(AREAY_SIZE+1))
			map_foreachinblockva(func,m,0,map[m].ys-(AREAY_SIZE*2+3),(AREAX_SIZE*2+2),map[m].ys-1,type,ap);
		else if (x >= map[m].xs-(AREAX_SIZE+1) && y <= AREAY_SIZE)
			map_foreachinblockva(func,m,map[m].xs-(AREAX_SIZE*2+3),0,map[m].xs-1,(AREAY_SIZE*2+2),type,ap);
		else if (x >= map[m].xs-(AREAX_SIZE+1) && y >= map[m].ys-(AREAY_SIZE+1))
			map_foreachinblockva(func,m,map[m].xs-(AREAX_SIZE*2+3),map[m].ys-(AREAY_SIZE*2+3),map[m].xs-1,map[m].ys-1,type,ap);
		else
			map_foreachinblockva(func,m,x-(AREAX_SIZE+1),y-(AREAY_SIZE+1),x+(AREAX_SIZE+1),y+(AREAY_SIZE+1),type,ap);
		break;
	case SAMEMAP:
		map_foreachinblockva(func,m,0,0,map[x].xs-1, map[x].ys-1,type,ap);
		break;
		}
	
	va_end(ap);
	return 0;
}

int map_foreachinblockva(int (*func)(struct block_list*,va_list),int m,int x0,int y0,int x1,int y1,int type,va_list ap) {
	int bx,by;
	struct block_list *bl;
	int blockcount=bl_list_count,i,c;

	if(m < 0)
		return 0;

	if (x0 < 0) x0 = 0;
	if (y0 < 0) y0 = 0;
	if (x1 >= map[m].xs) x1 = map[m].xs-1;
	if (y1 >= map[m].ys) y1 = map[m].ys-1;
	if (type != BL_NPC)
		for (by = y0 / BLOCK_SIZE; by <= y1 / BLOCK_SIZE; by++) {
			for(bx=x0/BLOCK_SIZE;bx<=x1/BLOCK_SIZE;bx++){
				bl = map[m].block[bx+by*map[m].bxs];
				c = map[m].block_count[bx+by*map[m].bxs];
				for(i=0;i<c && bl;i++,bl=bl->next){
					if(bl && type && bl->type!=type)
						continue;
					if(bl && bl->x>=x0 && bl->x<=x1 && bl->y>=y0 && bl->y<=y1 && bl_list_count<BL_LIST_MAX)
						bl_list[bl_list_count++]=bl;
				}
			}
		}
	if(type==0 || type==BL_NPC)
       
		for(by=y0/BLOCK_SIZE;by<=y1/BLOCK_SIZE;by++){
			for(bx=x0/BLOCK_SIZE;bx<=x1/BLOCK_SIZE;bx++){
				bl = map[m].block_npc[bx+by*map[m].bxs];
				c = map[m].block_npc_count[bx+by*map[m].bxs];
				if (c>0)
				   //printf("%d",c);
				for(i=0;i<c && bl;i++,bl=bl->next){
					if(bl && bl->x>=x0 && bl->x<=x1 && bl->y>=y0 && bl->y<=y1 && bl_list_count<BL_LIST_MAX)
	                    //printf("blocklist: %d\n",bl->graphic_id);
                        bl_list[bl_list_count++]=bl;
				}
			}
		}

	if(bl_list_count>=BL_LIST_MAX) {
		printf("BL_ERR: Too many block list! Block list count: %d\n", bl_list_count);
	}

	map_freeblock_lock();
    //printf("BLOCKLISTCOUNT: %d,%d,%d,%d\n",x0,y0,x1,y1);
	for(i=blockcount;i<bl_list_count;i++) {
		if(bl_list[i]->prev) {
			if (func(bl_list[i],ap)) break;
		}
	}

	map_freeblock_unlock();

	bl_list_count = blockcount;
	if(type==BL_MOB) {
		
	}
	return 0;
}

int map_foreachinblock(int (*func)(struct block_list*,va_list),int m,int x0,int y0,int x1,int y1,int type,...) {
	va_list ap;
	va_start(ap,type);

	map_foreachinblockva(func,m,x0,y0,x1,y1,type,ap);

	va_end(ap);
	return 0;
}

int map_foreachincell(int (*func)(struct block_list*,va_list),int m,int x,int y,int type,...) {
	int bx,by;
	struct block_list *bl;
	va_list ap;
	int blockcount=bl_list_count,i,c;

	if(m < 0)
		return 0;
	va_start(ap,type);
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x >= map[m].xs) x = map[m].xs-1;
	if (y >= map[m].ys) y = map[m].ys-1;

	by=y/BLOCK_SIZE;
	bx=x/BLOCK_SIZE;

	if(type!=BL_NPC)
	{
		bl = map[m].block[bx+by*map[m].bxs];
		c = map[m].block_count[bx+by*map[m].bxs];
		for(i=0;i<c && bl;i++,bl=bl->next)
		{
			if(type && bl && bl->type!=type)
				continue;
			if(bl && bl->x==x && bl->y==y && bl_list_count<BL_LIST_MAX)
				bl_list[bl_list_count++]=bl;
		}
	}

	if(type==0 || type==BL_NPC)
	{
		bl = map[m].block_npc[bx+by*map[m].bxs];
		c = map[m].block_npc_count[bx+by*map[m].bxs];
		for(i=0;i<c && bl;i++,bl=bl->next)
		{
			if(bl && bl->x==x && bl->y==y && bl_list_count<BL_LIST_MAX)
				bl_list[bl_list_count++]=bl;
		}
	}

	if(bl_list_count>=BL_LIST_MAX) {
		printf("BL_ERR: Too many block list! Block list count: %d\n", bl_list_count);
	}

	map_freeblock_lock();

	for(i=blockcount;i<bl_list_count;i++) {
		if(bl_list[i]->prev) {
			if (func(bl_list[i],ap)) break;
		}
	}

	map_freeblock_unlock();

	va_end(ap);
	bl_list_count = blockcount;
	return 0;
}

struct block_list* map_firstincell(int m,int x,int y,int type) {
	int bx,by;
	int i,c;
	struct block_list *bl;
	bl = NULL;

	if(m < 0)
		return 0;
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x >= map[m].xs) x = map[m].xs-1;
	if (y >= map[m].ys) y = map[m].ys-1;

	by=y/BLOCK_SIZE;
	bx=x/BLOCK_SIZE;

	if(type!=BL_NPC)
	{
		bl = map[m].block[bx+by*map[m].bxs];
		c = map[m].block_count[bx+by*map[m].bxs];
		for(i=0;i<c && bl;i++,bl=bl->next)
		{
			if(type && bl && bl->type!=type)
				continue;
			if(bl && bl->x==x && bl->y==y)
				return bl;
		}
	}

	if(type==0 || type==BL_NPC)
	{
		bl = map[m].block_npc[bx+by*map[m].bxs];
		c = map[m].block_npc_count[bx+by*map[m].bxs];
		for(i=0;i<c && bl;i++,bl=bl->next)
		{
			if(bl && bl->x==x && bl->y==y)
				return bl;
		}
	}

	return bl; // no item is here
}

struct block_list* map_firstincell2(int m,int x,int y,int type) {
	int bx,by;
	int i,c;
	struct block_list *bl;
	bl = NULL;

	if(m < 0)
		return 0;
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x >= map[m].xs) x = map[m].xs-1;
	if (y >= map[m].ys) y = map[m].ys-1;

	by=y/BLOCK_SIZE;
	bx=x/BLOCK_SIZE;

	if(type!=BL_NPC)
	{
		bl = map[m].block[bx+by*map[m].bxs];
		c = map[m].block_count[bx+by*map[m].bxs];
		for(i=0;i<c && bl;i++,bl=bl->next)
		{
			if(bl->type==BL_ITEM)
				continue;
			if(bl && bl->x==x && bl->y==y)
				return bl;
		}
	}

	if(type==0 || type==BL_NPC)
	{
		bl = map[m].block_npc[bx+by*map[m].bxs];
		c = map[m].block_npc_count[bx+by*map[m].bxs];
		for(i=0;i<c && bl;i++,bl=bl->next)
		{
			if(bl && bl->x==x && bl->y==y)
				return bl;
		}
	}

	return bl; // no item is here
}

struct block_list* map_firstincell_Reverse(int m,int x,int y,int type) {
	int bx,by;
	int i,c;
	struct block_list *bl;
	bl = NULL;

	if(m < 0)
		return 0;
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x >= map[m].xs) x = map[m].xs-1;
	if (y >= map[m].ys) y = map[m].ys-1;

	by=y/BLOCK_SIZE;
	bx=x/BLOCK_SIZE;

	
	if(type!=BL_NPC)
	{
		bl = map[m].block[bx+by*map[m].bxs];
		bl = map_getlastblock(bl);
		c = map[m].block_count[bx+by*map[m].bxs];
		for(i=0;i<c && bl;i++,bl=bl->prev) {
	// ignore type, whatever
	//		if(type && bl && bl->type!=type)
	//			continue;
			if(bl->type==BL_ITEM)
				continue;
			
			if(bl && bl->x==x && bl->y==y)
				return bl;
		}
	}

	// it seems like we don't really have to process it...
	if(type==0 || type==BL_NPC)
	{
		bl = map[m].block_npc[bx+by*map[m].bxs];
		bl = map_getlastblock(bl);
		c = map[m].block_npc_count[bx+by*map[m].bxs];
		for(i=0;i<c && bl;i++,bl=bl->prev)
		{
			if(bl && bl->x==x && bl->y==y)
				return bl;
		}
	}

	return bl; // no item is here
}

//map data (title,id,and sound) is readed from configuration file
int map_read(int id, const char *title, int bgm, const char *map_file) {
	unsigned short buff;
	unsigned int pos = 0;
	FILE *fp;
	fp = fopen(map_file, "rb");
	

	if(fp == NULL) {
		printf("MAP_ERR: Map file not found (%s).\n", map_file);
		return -1;
	}

	
	strncpy(map[id].title, title, 32);
	map[id].wl_last=NULL;
	map[id].wl_head=NULL;
	map[id].id=id;
	map[id].title[32] = '\0';
	map[id].bgm = bgm;
	fread(&buff, 2, 1, fp);
	map[id].xs = SWAP16(buff);
	fread(&buff, 2, 1, fp);
	map[id].ys = SWAP16(buff);
//	printf("%s: %d %d\n",map[id].title,map[id].xs,map[id].ys);
	CREATE(map[id].tile, unsigned short, map[id].xs*map[id].ys);
	CREATE(map[id].obj, unsigned short, map[id].xs*map[id].ys);
	CREATE(map[id].map, unsigned char, map[id].xs*map[id].ys);
	CREATE(map[id].pass, unsigned short, map[id].xs*map[id].ys);
	
	map[id].bxs = (map[id].xs+(BLOCK_SIZE-1))/BLOCK_SIZE;
	map[id].bys = (map[id].ys+(BLOCK_SIZE-1))/BLOCK_SIZE;
	CREATE(map[id].block, struct block_list*, map[id].bxs*map[id].bys);
	CREATE(map[id].block_npc, struct block_list*, map[id].bxs*map[id].bys);
	CREATE(map[id].block_count, int, map[id].bxs*map[id].bys);
	CREATE(map[id].block_npc_count, int, map[id].bxs*map[id].bys);
	CREATE(map[id].block_mob, struct mobspawn_data*,10000);
	
	
	while(!feof(fp)) {
		fread(&buff, 2, 1, fp);
		map[id].tile[pos] = SWAP16(buff);
		
		fread(&buff, 2, 1, fp);
		map[id].pass[pos] = SWAP16(buff);
		fread(&buff,2,1,fp);
		map[id].obj[pos] = SWAP16(buff);
		//map[id].pass[pos]=0;
		//all map section is walkable
		//map[id].map[pos] = 0;
		pos++;
		if (pos>=map[id].xs*map[id].ys)
			break;
	}
	fclose(fp);
	ShowStatus(" Load %s: %d %d\n",map[id].title,map[id].xs,map[id].ys);
	map_n++;
	return 0;
}

int map_readall() {
	struct map_src_list *i;
	for(i=map_src_first;i;i=i->next)
		map_read(i->id, i->title, i->bgm, i->mapfile);
	map_src_clear();
	printf("Map data file reading finished. %d map loaded!\n", map_n);
	return 0;
}

int map_setmapip(int id, unsigned int ip, unsigned short port) {
	if(id > MAX_MAP_PER_SERVER)
		return 1;

	map[id].ip = ip;
	map[id].port = port;
	return 0;
}

int lang_read(const char *cfg_file) {
	char line[1024], r1[1024], r2[1024];
	int line_num =0,i;
	FILE *fp;

	struct {
		char *name;
		int number;
	} map_msg_db[] = {
		{"MAP_WHISPFAIL", MAP_WHISPFAIL},
		{"MAP_ERRGHOST", MAP_ERRGHOST},
		{"MAP_ERRHORSE", MAP_ERRHORSE},
		{"MAP_ERRINVIS", MAP_ERRINVIS},
		{"MAP_ERRITMLEVEL", MAP_ERRITMLEVEL},
		{"MAP_ERRITMMIGHT", MAP_ERRITMMIGHT},
		{"MAP_ERRITMGRACE", MAP_ERRITMGRACE},
		{"MAP_ERRITMWILL", MAP_ERRITMWILL},
		{"MAP_ERRITMSEX", MAP_ERRITMSEX},
		{"MAP_ERRITMFULL", MAP_ERRITMFULL},
		{"MAP_EQHELM", MAP_EQHELM},
		{"MAP_EQWEAP", MAP_EQWEAP},
		{"MAP_EQARMOR", MAP_EQARMOR},
		{"MAP_EQSHIELD", MAP_EQSHIELD},
		{"MAP_EQRIGHT", MAP_EQRIGHT},
		{"MAP_EQLEFT", MAP_EQLEFT},
		{"MAP_EQSCRIPT1",MAP_EQSCRIPT1},
		{"MAP_EQSCRIPT2",MAP_EQSCRIPT2},
		{"WP_LVMIN",WP_LVMIN},
		{"WP_LVMIN2",WP_LVMIN2},
		{"WP_LVMIN3",WP_LVMIN3},
		{"WP_LVMAX",WP_LVMAX},
		{"MAP_ERRITMCLASS", MAP_ERRITMCLASS},
		{"MAP_ERRITMSUBCLASS", MAP_ERRITMSUBCLASS},
		{"MAP_ERRITM2H",MAP_ERRITM2H},
		{"MAP_ERRITM_WITH2H",MAP_ERRITM_WITH2H},
		{"MAP_ERREXPMAX",MAP_ERREXPMAX},
		{"MAP_GETEXP",MAP_GETEXP},
		{"MAP_CASH_GOLD",MAP_CASH_GOLD},
		{"MAP_CASH_SILVER",MAP_CASH_SILVER},
		{"MAP_CASH_SILVERS",MAP_CASH_SILVERS},
		{"MAP_CASH_DONG",MAP_CASH_DONG},
		{"MAP_CASH_DONGS",MAP_CASH_DONGS},
		{"MAP_ERRMONEYMAX",MAP_ERRMONEYMAX},
		{"MAP_EQUIP_ITEM",MAP_EQUIP_ITEM},
		{"MAP_HASTO_CLASS",MAP_HASTO_CLASS},
		{"MAP_ERRITMFULL2",MAP_ERRITMFULL2},
		{"MAP_ERRITMCFULL",MAP_ERRITMCFULL},
		{"MAP_ERR_EX_NOTALLOWED",MAP_ERR_EX_NOTALLOWED},
		{"MAP_ERR_EX_NOTALLOWED_P",MAP_ERR_EX_NOTALLOWED_P},
		{"MAP_ERR_EX_ALREADY_DOING",MAP_ERR_EX_ALREADY_DOING},
		{"MAP_ERR_EX_NOTOSELF",MAP_ERR_EX_NOTOSELF},
		{"MAP_EX_CANCLE",MAP_EX_CANCLE},
		{"MAP_EX_CANCLED",MAP_EX_CANCLED},
		{"MAP_EX_OK",MAP_EX_OK},
		{"MAP_ERR_EX_TERMINATED",MAP_ERR_EX_TERMINATED},
		{"MAP_WHISPER_ON",MAP_WHISPER_ON},
		{"MAP_WHISPER_OFF",MAP_WHISPER_OFF},
		{"MAP_EX_ON",MAP_EX_ON},
		{"MAP_EX_OFF",MAP_EX_OFF},
		{"MAP_GROUP_ON",MAP_GROUP_ON},
		{"MAP_GROUP_OFF",MAP_GROUP_OFF},
		{"MAP_PVP_ON",MAP_PVP_ON},
		{"MAP_PVP_OFF",MAP_PVP_OFF},
		{"MAP_GROUP_OUT",MAP_GROUP_OUT},
		{"MAP_GROUP_NOTOSELF",MAP_GROUP_NOTOSELF},
		{"MAP_GROUP_NOTALLOWED",MAP_GROUP_NOTALLOWED},
		{"MAP_GROUP_NOTALLOWED_P",MAP_GROUP_NOTALLOWED_P},
		{"MAP_GROUP_JOIN",MAP_GROUP_JOIN},
		{"MAP_GROUP_ALREADYJOIN",MAP_GROUP_ALREADYJOIN},
		{"MAP_GROUP_NOTJOIND",MAP_GROUP_NOTJOIND},
		{"MAP_GROUP_NOTEXIST_P",MAP_GROUP_NOTEXIST_P},
		{"MAP_WELCOME",MAP_WELCOME}
	};
	fp = fopen(cfg_file, "r");
	if (fp == NULL) {
		printf("CFG_ERR: Language file (%s) not found.\n", cfg_file);
		return 1;
	}

	while(fgets(line, sizeof(line), fp)) {
		line_num++;
		if (line[0] == '/' && line[1] == '/')
			continue;

		if (sscanf(line, "%[^:]: %[^\r\n]", r1, r2) == 2) {
			for(i=0;i<MSG_MAX;i++) {
				if (strcmpi(map_msg_db[i].name, r1)==0) {
					strncpy(map_msg[map_msg_db[i].number].message, r2, 256);
					map_msg[map_msg_db[i].number].message[255] = '\0';
					map_msg[map_msg_db[i].number].len = strlen(r2);
					if (map_msg[map_msg_db[i].number].len > 255)
						map_msg[map_msg_db[i].number].len=255;
					break;
				}
			}
		}
	}
	fclose(fp);
	sprintf(item_infos," %s \n %s \n %s \n %s \n %s\n %s \n %s \n %s ",
		map_msg[MAP_EQWEAP].message,
		map_msg[MAP_EQARMOR].message, 
		map_msg[MAP_EQSHIELD].message, 
		map_msg[MAP_EQHELM].message,
		map_msg[MAP_EQLEFT].message,
		map_msg[MAP_EQRIGHT].message, 
		map_msg[MAP_EQSCRIPT1].message, 
		map_msg[MAP_EQSCRIPT2].message);
	printf("Language File (%s) reading finished!\n", cfg_file);
	return 0;
}

int config_read(const char *cfg_file) {
	char line[1024], r1[1024], r2[1024];
	int line_num = 0;
	FILE *fp;

	fp = fopen(cfg_file, "r");
	if (fp == NULL) {
		printf("CFG_ERR: Configuration file (%s) not found.\n", cfg_file);
		return 1;
	}

	while(fgets(line, sizeof(line), fp)) {
		line_num++;
		if (line[0] == '/' && line[1] == '/')
			continue;

		if (sscanf(line, "%[^:]: %[^\r\n]", r1, r2) == 2) {
			//MAP
			if(strcmpi(r1, "map_ip") == 0) {
				strncpy(map_ip_s, r2, 16);
				map_ip_s[15] = '\0';
				map_ip = inet_addr(map_ip_s);
			} else if(strcmpi(r1, "map_port") == 0) {
				map_port = atoi(r2);
			//CHAR
			} else if(strcmpi(r1, "char_ip") == 0) {
				strncpy(char_ip_s, r2, 16);
				char_ip_s[15] = '\0';
				char_ip = inet_addr(char_ip_s);
			} else if(strcmpi(r1, "char_port") == 0) {
				
				char_port = atoi(r2);
			//	printf("char_port read: %d\n",char_port);
			} else if(strcmpi(r1, "char_id") == 0) {
				strncpy(char_id, r2, 32);
				char_id[31] = '\0';
			} else if(strcmpi(r1, "char_pw") == 0) {
				strncpy(char_pw, r2, 32);
				char_pw[31] = '\0';
			//SAVE TIME
			} else if(strcmpi(r1, "save_time") == 0) {
				save_time = atoi(r2)*1000;
			//LOG & DUMP
			} else if(strcmpi(r1, "map_log") == 0) {
				set_logfile(r2);
			} else if(strcmpi(r1, "dump_log") == 0) {
				set_dmpfile(r2);
			} else if(strcmpi(r1, "dump_save") == 0) {
				dump_save = atoi(r2);
			//MAP & NPC SCRIPT
			} else if(strcmpi(r1, "map") == 0) {
				if (map_src_add(r2)) {
					printf("CFG_ERR: Map Script Configuration parse error!\n");
					printf(" line %d: %s\n", line_num, line);
				}
			} else if(strcmpi(r1, "npc") == 0) {
				npc_src_add(r2);
            } else if(strcmpi(r1, "import") == 0) {
                script_config_read_sub(r2);   
			} else if(strcmpi(r1, "xprate") == 0) {
				xp_rate=atoi(r2);
			} else if(strcmpi(r1,"droprate") == 0) {
				d_rate=atoi(r2);
			}
		}
	}
	fclose(fp);
	printf("Configuration File (%s) reading finished!\n", cfg_file);
	return 0;
}

void do_term(void) {
	int i;
	do_final_script();
	warp_final();
	map_clritem();
	map_termiddb();
	mobdb_final();
	npc_final();
	for(i=0;i<MAX_MAP_PER_SERVER;i++){
		aFree(map[i].tile);
		aFree(map[i].obj);
		aFree(map[i].map);
		aFree(map[i].pass);
		aFree(map[i].block);
		aFree(map[i].block_npc);
		aFree(map[i].block_count);
		aFree(map[i].block_npc_count);
		aFree(map[i].block_mob);
	}
	
	map_termblock();
	itemdb_term();
	magicdb_term();
	classdb_term();
	printf("Gaya Map Server Shutdown.\n");
	add_log("Shutdown.\n");
	db_final();
}

void help_screen() {
	printf("HELP LIST\n");
	printf("---------\n");
	printf(" --conf [FILENAME]  : set config file\n");
	printf(" --inter [FILENAME] : set inter file\n");
	printf(" --lang [FILENAME]  : set lang file\n");
	exit(0);
}

void SET_SERVERNAME()
{
	strcpy(SERVER_NAME,"Gaya Map-server");
}

int do_init(int argc, char **argv) {
	int i;
	char *CONF_FILE = "conf/map.conf";
	char *LANG_FILE = "conf/lang.conf";
	char *INTER_FILE = "conf/inter.conf";
	char SCRIPT_FILE[256] = "conf/script.conf";
	srand(gettick());
	set_logfile("log/map.log");
	set_dmpfile("log/map_dump.log");
	db_init();
	for (i = 1; i < argc ; i++) {
		if (strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "--h") == 0 || strcmp(argv[i], "--?") == 0 || strcmp(argv[i], "/?") == 0)
			help_screen();
		else if (strcmp(argv[i], "--conf") == 0)
			CONF_FILE=argv[i+1];
		else if (strcmp(argv[i],"--inter") == 0)
			INTER_FILE = argv[i+1];
		else if (strcmp(argv[i],"--lang") == 0)
			LANG_FILE = argv[i+1];
	}

	config_read(CONF_FILE);
	config_read(INTER_FILE);
	lang_read(LANG_FILE);
	set_termfunc(do_term);
	//CREATE(userlist,struct userlist_data,1);
	add_log("");
	add_log("Gaya Map Server Started.\n");

	

	map_readall();
	map_initblock();
	map_initiddb();
	
	script_config_read(SCRIPT_FILE);
	do_init_script();

	itemdb_init();
	npc_init();
	mobdb_init();
	magicdb_init();
    classdb_init();
	intif_init();
	for(i=0;i<MAX_MAP_PER_SERVER;i++) {
		map[i].wl_head=NULL;
		map[i].wl_last=NULL;
		map[i].flag.warp=0;
	}
	warp_init();
	set_defaultaccept(clif_accept);
	set_defaultparse(clif_parse);
	map_fd = make_listen_port(map_port);
	
	add_timer_interval( gettick()+1000, check_connect_char, 0, 0,1000*10);
	for(i=0;i<MAX_MAP_PER_SERVER;i++) {
		if(map_isloaded(i)) {
			add_timer_interval(gettick()+250,mob_handle,i,i,250);
		}
	}	
	
	stall_time=86400;
	ShowStatus("Gaya Map Server is \033[1;32mready\033[0m! Listening at %d.\n", map_port);
	
	add_log("Server Ready! Listening at %d.\n", map_port);
	
	return 0;
}

int map_canmove(int m, int x, int y) {
//	struct block_list *bl;
//	int c;
	int obj;
//	int bx,by;
	int pass;
	obj=read_obj(m,x,y);
	pass=read_pass(m,x,y);

//	bx=x/BLOCK_SIZE;
//	by=y/BLOCK_SIZE;
	
//	bl=map[m].block[x+y*map[m].bxs];
//	c=map[m].block_count[bx+by*map[m].bxs];
	
	
	if(obj!=0 || pass!=0) {
		//printf("Object #: %d\n",obj);
		//printf("Cannot move: object %d\n",obj);
		
		return 1;
		
	}
	
	
	
	return 0;
}

int map_addspawn(struct mobspawn_data* mob) {
	int m;
	
	m=mob->bl.m;
	
	map[m].block_mob_count++;
	
	mob->num=map[m].block_mob_count;
	
	map[m].block_mob[mob->num]=mob;
	
	//printf("Spawn count: %d\n",mob->num);
	

	return 0;
}

int map_adduserlist(struct map_sessiondata* sd) {
	int max;
	int x;
	userlist.user_count++;
	max=userlist.user_count;
	for(x=0;x<max;x++) {
		if(userlist.user[x]==0) {
			userlist.user[x]=sd->bl.id;
			return 0;
		}
	}
	return 0;
}

int map_deluserlist(struct map_sessiondata* sd) {
	int max;
	int x;
	max=userlist.user_count;
	for(x=0;x<max;x++) {
		if(userlist.user[x]==sd->bl.id) {
			userlist.user[x]=0;
			userlist.user_count--;
			return 0;
		}
	}
	return 0;
}

struct block_list *map_getlastblock(struct block_list* root) {
	struct block_list *bl=root;
	if(root==NULL) return NULL;
	while(bl->next!=NULL) {
		bl=bl->next;
	}
	return bl;
};