// trade system


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "../common/malloc.h"
#include "../common/nullpo.h"
#include "../common/db.h"
#include "../common/timer.h"
#include "../common/ers.h"
#include "../common/malloc.h"
#include "../common/socket.h"
#include "../common/crypt.h"
#include "map.h"
#include "npc.h"
#include "script.h"
#include "magic.h"
#include "trade.h"
#include "pc.h"
#include "clif.h"
#include "class_db.h"


struct trade_data *trade_spacealloc(struct map_sessiondata *psd, struct map_sessiondata *tsd);
void trade_spacefree(struct trade_data *td);
int trade_open(struct map_sessiondata *psd, struct map_sessiondata *tsd);
void trade_cancel(struct map_sessiondata *sd);
void trade_ok(struct map_sessiondata *sd);
void trade_addcash(struct map_sessiondata *sd);
void trade_sendclose(int fd, const char *msg, int len);
void trade_sendcash(int fd, char type, int amount);
void trade_sendadditem(struct map_sessiondata *sd, int num);
int trade_errcheck(struct map_sessiondata *sd);

struct trade_data *trade_spacealloc(struct map_sessiondata *psd, struct map_sessiondata *tsd) {
	struct trade_data *ttd;
	CREATE(ttd, struct trade_data, 1);
	ttd->promoter=psd;
	ttd->partner=tsd;
	
	psd->trade_info=ttd;
	tsd->trade_info=ttd;

	return ttd;
};

void trade_spacefree(struct trade_data *td) {
	td->partner->trade_info=NULL;
	td->promoter->trade_info=NULL;
	aFree(td);
}

int trade_errcheck(struct map_sessiondata *sd) {
	return 1;
}

void trade_sendclose(int fd, const char *msg, int len) {

	WFIFOB(fd, 0) = 0xAA;
	WFIFOW(fd, 1) = SWAP16(6+len);
	WFIFOB(fd, 3) = 0x42;
	WFIFOB(fd, 4) = 0;
	WFIFOB(fd, 5) = 0x04;
	WFIFOB(fd, 6) = 0x00;
	WFIFOB(fd, 7) = len;
	memcpy(WFIFOP(fd, 8), msg, len);
	WFIFOSET(fd, crypt(WFIFOP(fd, 0)));

}

void trade_sendadditem(struct map_sessiondata *sd, int num) {
	struct trade_data* td=sd->trade_info;
	if(td==NULL) {
		ShowError("NULL trade_info\n");
		return 0;
	}
	num--;
	
}

void trade_terminate(struct map_sessiondata *sd) {
	trade_sendclose(sd->trade_info->partner->fd, map_msg[MAP_ERR_EX_TERMINATED].message, map_msg[MAP_ERR_EX_TERMINATED].len);
	trade_spacefree(sd->trade_info);
}

// send exchange window opening packet to partner
// and allocate space to exchange things
int trade_open(struct map_sessiondata *psd, struct map_sessiondata *tsd) {
	int fd;
	int len;
	char name[256];
	if(psd->status.opt.al_exchg == 0) 
		return MAP_ERR_EX_NOTALLOWED;
	if(tsd->status.opt.al_exchg == 0) 
		return MAP_ERR_EX_NOTALLOWED_P;
	if(tsd->trade_info) 
		return MAP_ERR_EX_ALREADY_DOING;


	// send packet to open trade window
	fd=psd->fd;
	len=sprintf(name, "%s(%s)", tsd->status.name, class[tsd->status.class][tsd->status.subclass].class_name);

	WFIFOB(fd, 0) = 0xAA;
	WFIFOW(fd, 1) = SWAP16(len+9);
	WFIFOB(fd, 3) = 0x42;
	WFIFOB(fd, 4) = 0;
	WFIFOB(fd, 5) = 0;
	WFIFOL(fd, 6) = SWAP32(tsd->status.id);
	WFIFOB(fd, 10) = len&0xFF;
	memcpy(WFIFOP(fd, 11), name, len);
	WFIFOW(fd, len+11) = SWAP16(100); //ǰ
//	WFIFOB(fd, 13) = len;
//	memcpy(WFIFOP(fd, 14), tsd->status.name, len);
	WFIFOSET(fd, crypt(WFIFOP(fd, 0)) );

	fd=tsd->fd;
	len=sprintf(name, "%s(%s)", psd->status.name, class[psd->status.class][psd->status.subclass].class_name);
	WFIFOB(fd, 0) = 0xAA;
	WFIFOW(fd, 1) = SWAP16(len+9);
	WFIFOB(fd, 3) = 0x42;
	WFIFOB(fd, 4) = 0;
	WFIFOB(fd, 5) = 0;
	WFIFOL(fd, 6) = SWAP32(psd->status.id);
	WFIFOB(fd, 10) = len&0xFF;
	memcpy(WFIFOP(fd, 11), psd->status.name, len);
	WFIFOW(fd, 11+len) = SWAP16(100); //ǰ
//	WFIFOB(fd, 13) = len;
//	memcpy(WFIFOP(fd, 14), psd->status.name, len);

//	memcpy(WFIFOP(fd, 10), psd->status.name, len);
	WFIFOSET(fd, crypt(WFIFOP(fd, 0)) );

	trade_spacealloc(psd,tsd);
	return 0;
}

//trade cancel
void trade_cancel(struct map_sessiondata *sd) {
// AA 00 18 42 A0 04 00 13 45 78 63 68 61 6E 67 65
// 20 63 61 6E 63 65 6C 6C 65 64 2E
	struct trade_data *td;
	td=sd->trade_info;

	trade_sendclose(td->partner->fd, map_msg[MAP_EX_CANCLED].message, map_msg[MAP_EX_CANCLED].len);
	trade_sendclose(td->promoter->fd, map_msg[MAP_EX_CANCLE].message, map_msg[MAP_EX_CANCLE].len);
	trade_spacefree(td);
}

//trade commit
void trade_ok(struct map_sessiondata *sd) {
	struct trade_data *td=sd->trade_info;
	trade_sendclose(td->partner->fd, map_msg[MAP_EX_OK].message, map_msg[MAP_EX_OK].len);
	trade_sendclose(td->promoter->fd, map_msg[MAP_EX_OK].message, map_msg[MAP_EX_OK].len);

	trade_spacefree(td);
}

//add item
void trade_additem(struct map_sessiondata *sd) {
	return;
}

//add cash
void trade_addcash(struct map_sessiondata *sd) {
	struct trade_data *td=sd->trade_info;
	int money_amount=SWAP32(RFIFOL(sd->fd, 10));
	
	if(money_amount> sd->status.money) {
		money_amount= sd->status.money;
	}
	if(sd==td->promoter) {
		trade_sendcash(td->promoter->fd, 0, money_amount);
		trade_sendcash(td->partner->fd, 1, money_amount);
		td->p_money=money_amount;
	} else if(sd==td->partner) {
		trade_sendcash(td->promoter->fd, 1, money_amount);
		trade_sendcash(td->partner->fd, 0, money_amount);
		td->t_money=money_amount;
	} else {
		add_log("EXCHANGE ERROR: %s & %s occur some errors in exchange\n",td->promoter->status.name, td->partner->status.name);
		ShowError("EXCHANGE ERROR! see the map log file\n");
	}
	return;
}

void trade_sendcash(int fd, char type, int amount) {
	WFIFOB(fd, 0) = 0xAA;
	WFIFOW(fd, 1) = SWAP16(9);
	WFIFOB(fd, 3) = 0x42;
	WFIFOB(fd, 4) = 0;
	WFIFOB(fd, 5) = 0x03;
	WFIFOB(fd, 6) = type&0xFF;
	WFIFOL(fd, 7) = SWAP32(amount);
}


int clif_parsetrade(struct map_sessiondata *sd) {
	int fd=sd->fd;
	int ret;
	{ int i; for(i=0;i<20;i++) { printf("%x ",RFIFOB(fd, i));} printf("\n"); }

	switch(RFIFOB(fd, 5)) {
	case 0x00: {// Request to joint trade [OK]
			int id= SWAP32(RFIFOL(fd, 6));
			struct map_sessiondata *tsd;
			tsd=map_id2sd(id);
			if(tsd==NULL) {
				ShowError("tsd does not found! id:%d\n",id);
				return 0; // target doesn't found
			}
			if(ret=trade_open(sd,tsd)) {
				clif_sendminitext(sd,map_msg[ret].message,map_msg[ret].len);
				return 0;
			} 
		}
		break;
	case 0x01: // Request adding item into the list
		trade_additem(sd);
		break;
	case 0x03: // Request adding cash
		trade_addcash(sd);
		break;
	case 0x04: // Request to close trade [OK]
		trade_cancel(sd);
		break;
	case 0x05: // Request to close trade successful [OK]
		trade_ok(sd);
		break;
	}

	return 1;
}

	
